--[[
Name: LibBabble-Boss-3.0
Revision: $Rev: 304 $
Maintainers: ckknight, nevcairiel, Ackis
Website: http://www.wowace.com/projects/libbabble-boss-3-0/
Dependencies: None
License: MIT
]]

local MAJOR_VERSION = "DPSBabble-Boss-3.1"
local MINOR_VERSION = 90000 + tonumber(("$Rev: 305 $"):match("%d+"))

-- LibBabble-3.0 is hereby placed in the Public Domain
-- Credits: ckknight
local LIBBABBLE_MAJOR, LIBBABBLE_MINOR = "DPSBabble-3.1", 2

local LibBabble = LibStub:NewLibrary(LIBBABBLE_MAJOR, LIBBABBLE_MINOR)
if not LibBabble then
	return
end

local data = LibBabble.data or {}
for k,v in pairs(LibBabble) do
	LibBabble[k] = nil
end
LibBabble.data = data

local tablesToDB = {}
for namespace, db in pairs(data) do
	for k,v in pairs(db) do
		tablesToDB[v] = db
	end
end

local function warn(message)
	--local _, ret = pcall(error, message, 3)
	--geterrorhandler()(ret)
end

local lookup_mt = { __index = function(self, key)
	local db = tablesToDB[self]
	local current_key = db.current[key]
	if current_key then
		self[key] = current_key
		return current_key
	end
	local base_key = db.base[key]
	local real_MAJOR_VERSION
	for k,v in pairs(data) do
		if v == db then
			real_MAJOR_VERSION = k
			break
		end
	end
	if not real_MAJOR_VERSION then
		real_MAJOR_VERSION = LIBBABBLE_MAJOR
	end
	if base_key then
		--warn(("%s: Translation %q not found for locale %q"):format(real_MAJOR_VERSION, key, GetLocale()))
		rawset(self, key, base_key)
		return base_key
	end
	--warn(("%s: Translation %q not found."):format(real_MAJOR_VERSION, key))
	rawset(self, key, key)
	return key
end }

local function initLookup(module, lookup)
	local db = tablesToDB[module]
	for k in pairs(lookup) do
		lookup[k] = nil
	end
	setmetatable(lookup, lookup_mt)
	tablesToDB[lookup] = db
	db.lookup = lookup
	return lookup
end

local function initReverse(module, reverse)
	local db = tablesToDB[module]
	for k in pairs(reverse) do
		reverse[k] = nil
	end
	for k,v in pairs(db.current) do
		reverse[v] = k
	end
	tablesToDB[reverse] = db
	db.reverse = reverse
	db.reverseIterators = nil
	return reverse
end

local prototype = {}
local prototype_mt = {__index = prototype}
function prototype:GetLookupTable()
	local db = tablesToDB[self]

	local lookup = db.lookup
	if lookup then
		return lookup
	end
	return initLookup(self, {})
end
function prototype:GetUnstrictLookupTable()
	local db = tablesToDB[self]

	return db.current
end
function prototype:GetBaseLookupTable()
	local db = tablesToDB[self]

	return db.base
end
function prototype:GetReverseLookupTable()
	local db = tablesToDB[self]

	local reverse = db.reverse
	if reverse then
		return reverse
	end
	return initReverse(self, {})
end
local blank = {}
local weakVal = {__mode='v'}
function prototype:GetReverseIterator(key)
	local db = tablesToDB[self]
	local reverseIterators = db.reverseIterators
	if not reverseIterators then
		reverseIterators = setmetatable({}, weakVal)
		db.reverseIterators = reverseIterators
	elseif reverseIterators[key] then
		return pairs(reverseIterators[key])
	end
	local t
	for k,v in pairs(db.current) do
		if v == key then
			if not t then
				t = {}
			end
			t[k] = true
		end
	end
	reverseIterators[key] = t or blank
	return pairs(reverseIterators[key])
end
function prototype:Iterate()
	local db = tablesToDB[self]

	return pairs(db.current)
end

-- #NODOC
-- modules need to call this to set the base table
function prototype:SetBaseTranslations(base)
	local db = tablesToDB[self]
	local oldBase = db.base
	if oldBase then
		for k in pairs(oldBase) do
			oldBase[k] = nil
		end
		for k, v in pairs(base) do
			oldBase[k] = v
		end
		base = oldBase
	else
		db.base = base
	end
	for k,v in pairs(base) do
		if v == true then
			base[k] = k
		end
	end
end

local function init(module)
	local db = tablesToDB[module]
	if db.lookup then
		initLookup(module, db.lookup)
	end
	if db.reverse then
		initReverse(module, db.reverse)
	end
	db.reverseIterators = nil
end

-- #NODOC
-- modules need to call this to set the current table. if current is true, use the base table.
function prototype:SetCurrentTranslations(current)
	local db = tablesToDB[self]
	if current == true then
		db.current = db.base
	else
		local oldCurrent = db.current
		if oldCurrent then
			for k in pairs(oldCurrent) do
				oldCurrent[k] = nil
			end
			for k, v in pairs(current) do
				oldCurrent[k] = v
			end
			current = oldCurrent
		else
			db.current = current
		end
	end
	init(self)
end

for namespace, db in pairs(data) do
	setmetatable(db.module, prototype_mt)
	init(db.module)
end

-- #NODOC
-- modules need to call this to create a new namespace.
function LibBabble:New(namespace, minor)
	local module, oldminor = LibStub:NewLibrary(namespace, minor)
	if not module then
		return
	end

	if not oldminor then
		local db = {
			module = module,
		}
		data[namespace] = db
		tablesToDB[module] = db
	else
		for k,v in pairs(module) do
			module[k] = nil
		end
	end

	setmetatable(module, prototype_mt)

	return module
end


if not LibStub then error(MAJOR_VERSION .. " requires LibStub.") end
local lib = LibStub("DPSBabble-3.1"):New(MAJOR_VERSION, MINOR_VERSION)
if not lib then return end

local GAME_LOCALE = GetLocale()

lib:SetBaseTranslations {
	Acidmaw = "Acidmaw",
	Aeonus = "Aeonus",
	["Aerial Command Unit"] = "Aerial Command Unit",
	["Agathelos the Raging"] = "Agathelos the Raging",
	["Aggem Thorncurse"] = "Aggem Thorncurse",
	Ahune = "Ahune",
	["Akil'zon"] = "Akil'zon",
	["Aku'mai"] = "Aku'mai",
	["Al'ar"] = "Al'ar",
	["Algalon the Observer"] = "Algalon the Observer",
	Altarius = "Altarius",
	["Alzzin the Wildshaper"] = "Alzzin the Wildshaper",
	Amanitar = "Amanitar",
	["Ambassador Flamelash"] = "Ambassador Flamelash",
	["Ambassador Hellmaw"] = "Ambassador Hellmaw",
	Ammunae = "Ammunae",
	["Amnennar the Coldbringer"] = "Amnennar the Coldbringer",
	["Ancient Stone Keeper"] = "Ancient Stone Keeper",
	Anetheron = "Anetheron",
	["Anger'rel"] = "Anger'rel",
	Anomalus = "Anomalus",
	Anraphet = "Anraphet",
	["Antu'sul"] = "Antu'sul",
	["Anub'Rekhan"] = "Anub'Rekhan",
	["Anub'arak"] = "Anub'arak",
	["Anub'shiah"] = "Anub'shiah",
	["Anubisath Defender"] = "Anubisath Defender",
	["Anubisath Guardian"] = "Anubisath Guardian",
	Anzu = "Anzu",
	["Arcane Watchman"] = "Arcane Watchman",
	["Arcanist Doan"] = "Arcanist Doan",
	Archaedas = "Archaedas",
	["Archavon the Stone Watcher"] = "Archavon the Stone Watcher",
	Archimonde = "Archimonde",
	["Archivist Galford"] = "Archivist Galford",
	["Archmage Arugal"] = "Archmage Arugal",
	["Argent Confessor Paletress"] = "Argent Confessor Paletress",
	["Arugal's Voidwalker"] = "Arugal's Voidwalker",
	["Ascendant Lord Obsidius"] = "Ascendant Lord Obsidius",
	["Asim al Akir"] = "Asim al Akir",
	["Assault Bot"] = "Assault Bot",
	["Assembly of Iron"] = "Assembly of Iron",
	["Atal'alarion"] = "Atal'alarion",
	Atramedes = "Atramedes",
	["Attumen the Huntsman"] = "Attumen the Huntsman",
	Auriaya = "Auriaya",
	Avalanchion = "Avalanchion",
	["Avatar of Hakkar"] = "Avatar of Hakkar",
	["Ayamiss the Hunter"] = "Ayamiss the Hunter",
	Azgalor = "Azgalor",
	["Azshir the Sleepless"] = "Azshir the Sleepless",
	Azuregos = "Azuregos",
	["Bael'Gar"] = "Bael'Gar",
	Baelog = "Baelog",
	Balnazzar = "Balnazzar",
	["Baltharus the Warborn"] = "Baltharus the Warborn",
	["Bannok Grimaxe"] = "Bannok Grimaxe",
	["Baron Aquanis"] = "Baron Aquanis",
	["Baron Charr"] = "Baron Charr",
	["Baron Geddon"] = "Baron Geddon",
	["Baron Kazum"] = "Baron Kazum",
	["Baron Rivendare"] = "Baron Rivendare",
	["Baron Silverlaine"] = "Baron Silverlaine",
	["Baroness Anastari"] = "Baroness Anastari",
	["Battleguard Sartura"] = "Battleguard Sartura",
	["Bazil Thredd"] = "Bazil Thredd",
	Bazzalan = "Bazzalan",
	Beauty = "Beauty",
	["Black Guard Swordsmith"] = "Black Guard Swordsmith",
	["Blackheart the Inciter"] = "Blackheart the Inciter",
	["Blind Hunter"] = "Blind Hunter",
	["Blindeye the Seer"] = "Blindeye the Seer",
	["Blood Guard Porung"] = "Blood Guard Porung",
	["Blood Prince Council"] = "Blood Prince Council",
	["Blood Princes"] = "Blood Princes",
	["Blood Steward of Kirtonos"] = "Blood Steward of Kirtonos",
	["Blood-Queen Lana'thel"] = "Blood-Queen Lana'thel",
	["Bloodlord Mandokir"] = "Bloodlord Mandokir",
	["Bloodmage Thalnos"] = "Bloodmage Thalnos",
	Boahn = "Boahn",
	["Bomb Bot"] = "Bomb Bot",
	["Brain of Yogg-Saron"] = "Brain of Yogg-Saron",
	["Brainwashed Noble"] = "Brainwashed Noble",
	Broggok = "Broggok",
	Brokentoe = "Brokentoe",
	Bronjahm = "Bronjahm",
	["Broodlord Lashlayer"] = "Broodlord Lashlayer",
	["Bruegal Ironknuckle"] = "Bruegal Ironknuckle",
	Brutallus = "Brutallus",
	["Burning Felguard"] = "Burning Felguard",
	["Buru the Gorger"] = "Buru the Gorger",
	["C'Thun"] = "C'Thun",
	["Cache of the Firelord"] = "Cache of the Firelord",
	["Cache of the Legion"] = "Cache of the Legion",
	["Cannon Master Willey"] = "Cannon Master Willey",
	["Captain Greenskin"] = "Captain Greenskin",
	["Captain Kromcrush"] = "Captain Kromcrush",
	["Captain Skarloc"] = "Captain Skarloc",
	["Celebras the Cursed"] = "Celebras the Cursed",
	["Charlga Razorflank"] = "Charlga Razorflank",
	["Chess Event"] = "Chess Event",
	["Chest of The Seven"] = "Chest of The Seven",
	["Chief Ukorz Sandscalp"] = "Chief Ukorz Sandscalp",
	Chimaeron = "Chimaeron",
	["Cho'Rush the Observer"] = "Cho'Rush the Observer",
	Chromaggus = "Chromaggus",
	["Chrono Lord Deja"] = "Chrono Lord Deja",
	["Chrono-Lord Epoch"] = "Chrono-Lord Epoch",
	Claw = "Claw",
	["Coilfang Elite"] = "Coilfang Elite",
	["Coilfang Strider"] = "Coilfang Strider",
	["Commander Kolurg"] = "Commander Kolurg",
	["Commander Sarannis"] = "Commander Sarannis",
	["Commander Springvale"] = "Commander Springvale",
	["Commander Stoutbeard"] = "Commander Stoutbeard",
	["Commander Ulthok"] = "Commander Ulthok",
	["Constructor & Controller"] = "Constructor & Controller",
	Cookie = "Cookie",
	Corborus = "Corborus",
	["Coren Direbrew"] = "Coren Direbrew",
	["Corla, Herald of Twilight"] = "Corla, Herald of Twilight",
	["Cosmic Infuser"] = "Cosmic Infuser",
	["Crimson Hammersmith"] = "Crimson Hammersmith",
	["Crowd Pummeler 9-60"] = "Crowd Pummeler 9-60",
	["Crystal Fang"] = "Crystal Fang",
	Cyanigosa = "Cyanigosa",
	["Dalliah the Doomsayer"] = "Dalliah the Doomsayer",
	["Dalronn the Controller"] = "Dalronn the Controller",
	["Dark Coffer"] = "Dark Coffer",
	["Dark Iron Ambassador"] = "Dark Iron Ambassador",
	["Darkmaster Gandling"] = "Darkmaster Gandling",
	["Darkweaver Syth"] = "Darkweaver Syth",
	["Death Knight Darkreaver"] = "Death Knight Darkreaver",
	["Death Knight Understudy"] = "Death Knight Understudy",
	["Death Speaker Jargba"] = "Death Speaker Jargba",
	["Deathbound Ward"] = "Deathbound Ward",
	["Deathbringer Saurfang"] = "Deathbringer Saurfang",
	["Deathspeaker High Priest"] = "Deathspeaker High Priest",
	["Deathstalker Visceri"] = "Deathstalker Visceri",
	["Deathsworn Captain"] = "Deathsworn Captain",
	Devastation = "Devastation",
	["Deviate Faerie Dragon"] = "Deviate Faerie Dragon",
	["Devourer of Souls"] = "Devourer of Souls",
	["Dextren Ward"] = "Dextren Ward",
	["Digmaster Shovelphlange"] = "Digmaster Shovelphlange",
	["Doctor Theolen Krastinov"] = "Doctor Theolen Krastinov",
	["Doom Lord Kazzak"] = "Doom Lord Kazzak",
	["Doom'rel"] = "Doom'rel",
	Doomwalker = "Doomwalker",
	["Dope'rel"] = "Dope'rel",
	Dorothee = "Dorothee",
	["Drahga Shodowburner"] = "Drahga Shodowburner",
	["Drakkari Colossus"] = "Drakkari Colossus",
	["Drakos the Interrogator"] = "Drakos the Interrogator",
	Dreadscale = "Dreadscale",
	Dreamscythe = "Dreamscythe",
	["Dust Covered Chest"] = "Dust Covered Chest",
	Dustwraith = "Dustwraith",
	["Eadric the Pure"] = "Eadric the Pure",
	["Earthcaller Halmgar"] = "Earthcaller Halmgar",
	["Earthrager Ptah"] = "Earthrager Ptah",
	Ebonroc = "Ebonroc",
	["Eck the Ferocious"] = "Eck the Ferocious",
	["Edwin VanCleef"] = "Edwin VanCleef",
	["Elder Brightleaf"] = "Elder Brightleaf",
	["Elder Ironbranch"] = "Elder Ironbranch",
	["Elder Nadox"] = "Elder Nadox",
	["Elder Stonebark"] = "Elder Stonebark",
	["Electrocutioner 6000"] = "Electrocutioner 6000",
	["Emalon the Storm Watcher"] = "Emalon the Storm Watcher",
	Emeriss = "Emeriss",
	["Emperor Dagran Thaurissan"] = "Emperor Dagran Thaurissan",
	["Emperor Vek'lor"] = "Emperor Vek'lor",
	["Emperor Vek'nilash"] = "Emperor Vek'nilash",
	Entropius = "Entropius",
	["Eonar's Gift"] = "Eonar's Gift",
	["Epoch Hunter"] = "Epoch Hunter",
	Erekem = "Erekem",
	["Eressea Dawnsinger"] = "Eressea Dawnsinger",
	["Eric \"The Swift\""] = "Eric \"The Swift\"",
	Erudax = "Erudax",
	["Erunak Stonespeaker"] = "Erunak Stonespeaker",
	["Essence of Anger"] = "Essence of Anger",
	["Essence of Desire"] = "Essence of Desire",
	["Essence of Suffering"] = "Essence of Suffering",
	Eviscerator = "Eviscerator",
	["Exarch Maladaar"] = "Exarch Maladaar",
	["Expedition Commander"] = "Expedition Commander",
	["Eydis Darkbane"] = "Eydis Darkbane",
	["Eye of C'Thun"] = "Eye of C'Thun",
	["Faction Champions"] = "Faction Champions",
	["Fallen Champion"] = "Fallen Champion",
	Falric = "Falric",
	["Falric and Marwyn"] = "Falric and Marwyn",
	["Fankriss the Unyielding"] = "Fankriss the Unyielding",
	["Fathom-Lord Karathress"] = "Fathom-Lord Karathress",
	Felmyst = "Felmyst",
	["Fenrus the Devourer"] = "Fenrus the Devourer",
	["Feral Defender"] = "Feral Defender",
	Festergut = "Festergut",
	Feugen = "Feugen",
	["Fineous Darkvire"] = "Fineous Darkvire",
	Firemaw = "Firemaw",
	["Fjola Lightbane"] = "Fjola Lightbane",
	["Flame Leviathan"] = "Flame Leviathan",
	Flamegor = "Flamegor",
	["Foreman Thistlenettle"] = "Foreman Thistlenettle",
	["Forgemaster Garfrost"] = "Forgemaster Garfrost",
	["Forgemaster Throngus"] = "Forgemaster Throngus",
	["Four Horsemen Chest"] = "Four Horsemen Chest",
	["Fras Siabi"] = "Fras Siabi",
	Freya = "Freya",
	["Gahz'ranka"] = "Gahz'ranka",
	["Gahz'rilla"] = "Gahz'rilla",
	["Gal'darah"] = "Gal'darah",
	["Galgann Firehammer"] = "Galgann Firehammer",
	Garr = "Garr",
	["Garrosh Hellscream"] = "Garrosh Hellscream",
	Gasher = "Gasher",
	["Gatewatcher Gyro-Kill"] = "Gatewatcher Gyro-Kill",
	["Gatewatcher Iron-Hand"] = "Gatewatcher Iron-Hand",
	["Gathios the Shatterer"] = "Gathios the Shatterer",
	Gehennas = "Gehennas",
	Gelihast = "Gelihast",
	Gelk = "Gelk",
	["General Angerforge"] = "General Angerforge",
	["General Bjarngrim"] = "General Bjarngrim",
	["General Drakkisath"] = "General Drakkisath",
	["General Husam"] = "General Husam",
	["General Rajaxx"] = "General Rajaxx",
	["General Umbriss"] = "General Umbriss",
	["General Vezax"] = "General Vezax",
	["General Zarithrian"] = "General Zarithrian",
	["Ghamoo-ra"] = "Ghamoo-ra",
	["Ghaz'an"] = "Ghaz'an",
	["Ghok Bashguud"] = "Ghok Bashguud",
	Gilnid = "Gilnid",
	["Gizrul the Slavener"] = "Gizrul the Slavener",
	["Gloom'rel"] = "Gloom'rel",
	Gluth = "Gluth",
	Glutton = "Glutton",
	["Golem Lord Argelmach"] = "Golem Lord Argelmach",
	["Golemagg the Incinerator"] = "Golemagg the Incinerator",
	["Goraluk Anvilcrack"] = "Goraluk Anvilcrack",
	["Gormok the Impaler"] = "Gormok the Impaler",
	["Gorosh the Dervish"] = "Gorosh the Dervish",
	["Gortok Palehoof"] = "Gortok Palehoof",
	["Gothik the Harvester"] = "Gothik the Harvester",
	["Grand Astromancer Capernian"] = "Grand Astromancer Capernian",
	["Grand Champions"] = "Grand Champions",
	["Grand Crusader Dathrohan"] = "Grand Crusader Dathrohan",
	["Grand Magus Telestra"] = "Grand Magus Telestra",
	["Grand Vizier Ertan"] = "Grand Vizier Ertan",
	["Grand Warlock Alythess"] = "Grand Warlock Alythess",
	["Grand Warlock Nethekurse"] = "Grand Warlock Nethekurse",
	["Grand Widow Faerlina"] = "Grand Widow Faerlina",
	["Grandmaster Vorpil"] = "Grandmaster Vorpil",
	Grandmother = "Grandmother",
	["Grethok the Controller"] = "Grethok the Controller",
	["Gri'lek"] = "Gri'lek",
	Grimlok = "Grimlok",
	Grizzle = "Grizzle",
	Grobbulus = "Grobbulus",
	Grubbis = "Grubbis",
	["Gruul the Dragonkiller"] = "Gruul the Dragonkiller",
	["Guard Fengus"] = "Guard Fengus",
	["Guard Mol'dar"] = "Guard Mol'dar",
	["Guard Slip'kik"] = "Guard Slip'kik",
	["Guardian of Yogg-Saron"] = "Guardian of Yogg-Saron",
	["Gurtogg Bloodboil"] = "Gurtogg Bloodboil",
	Gyth = "Gyth",
	Hadronox = "Hadronox",
	Hakkar = "Hakkar",
	Halazzi = "Halazzi",
	Halion = "Halion",
	Halycon = "Halycon",
	Hamhock = "Hamhock",
	["Harbinger Skyriss"] = "Harbinger Skyriss",
	["Hate'rel"] = "Hate'rel",
	["Hazza'rah"] = "Hazza'rah",
	Hazzas = "Hazzas",
	["Headless Horseman"] = "Headless Horseman",
	["Hearthsinger Forresten"] = "Hearthsinger Forresten",
	["Hedrum the Creeper"] = "Hedrum the Creeper",
	["Heigan the Unclean"] = "Heigan the Unclean",
	["Hellfire Channeler"] = "Hellfire Channeler",
	["Henry Stern"] = "Henry Stern",
	["Herald Volazj"] = "Herald Volazj",
	Herod = "Herod",
	["Hex Lord Malacrass"] = "Hex Lord Malacrass",
	["High Astromancer Solarian"] = "High Astromancer Solarian",
	["High Botanist Freywinn"] = "High Botanist Freywinn",
	["High Inquisitor Fairbanks"] = "High Inquisitor Fairbanks",
	["High Inquisitor Whitemane"] = "High Inquisitor Whitemane",
	["High Interrogator Gerstahn"] = "High Interrogator Gerstahn",
	["High King Maulgar"] = "High King Maulgar",
	["High Marshal Whirlaxis"] = "High Marshal Whirlaxis",
	["High Nethermancer Zerevor"] = "High Nethermancer Zerevor",
	["High Overlord Saurfang"] = "High Overlord Saurfang",
	["High Priest Thekal"] = "High Priest Thekal",
	["High Priest Venoxis"] = "High Priest Venoxis",
	["High Priestess Arlokk"] = "High Priestess Arlokk",
	["High Priestess Azil"] = "High Priestess Azil",
	["High Priestess Jeklik"] = "High Priestess Jeklik",
	["High Priestess Mar'li"] = "High Priestess Mar'li",
	["High Priestess of Thaurissan"] = "High Priestess of Thaurissan",
	["High Prophet Barim"] = "High Prophet Barim",
	["High Warlord Naj'entus"] = "High Warlord Naj'entus",
	["Highlord Mograine"] = "Highlord Mograine",
	["Highlord Omokk"] = "Highlord Omokk",
	Hodir = "Hodir",
	["Houndmaster Grebmar"] = "Houndmaster Grebmar",
	["Houndmaster Loksey"] = "Houndmaster Loksey",
	Hukku = "Hukku",
	Hungarfen = "Hungarfen",
	["Hurley Blackbreath"] = "Hurley Blackbreath",
	["Hyakiss the Lurker"] = "Hyakiss the Lurker",
	["Hydromancer Thespia"] = "Hydromancer Thespia",
	["Hydromancer Velratha"] = "Hydromancer Velratha",
	Hydrospawn = "Hydrospawn",
	["Hydross the Unstable"] = "Hydross the Unstable",
	["Ice Sphere"] = "Ice Sphere",
	["Icecrown Gunship Battle"] = "Icecrown Gunship Battle",
	Icehowl = "Icehowl",
	Ichoron = "Ichoron",
	Ick = "Ick",
	["Ignis the Furnace Master"] = "Ignis the Furnace Master",
	["Illidan Stormrage"] = "Illidan Stormrage",
	["Illidari Council"] = "Illidari Council",
	["Illyanna Ravenoak"] = "Illyanna Ravenoak",
	["Immol'thar"] = "Immol'thar",
	["Infinite Corruptor"] = "Infinite Corruptor",
	["Infinity Blades"] = "Infinity Blades",
	["Ingvar the Plunderer"] = "Ingvar the Plunderer",
	["Instructor Malicia"] = "Instructor Malicia",
	["Instructor Razuvious"] = "Instructor Razuvious",
	["Interrogator Vishas"] = "Interrogator Vishas",
	Ionar = "Ionar",
	Ironaya = "Ironaya",
	Ironspine = "Ironspine",
	Isalien = "Isalien",
	Isiset = "Isiset",
	Jade = "Jade",
	["Jammal'an the Prophet"] = "Jammal'an the Prophet",
	["Jan'alai"] = "Jan'alai",
	["Jandice Barov"] = "Jandice Barov",
	["Jed Runewatcher"] = "Jed Runewatcher",
	["Jedoga Shadowseeker"] = "Jedoga Shadowseeker",
	["Jergosh the Invoker"] = "Jergosh the Invoker",
	["Jin'do the Hexxer"] = "Jin'do the Hexxer",
	["Jormungar Behemoth"] = "Jormungar Behemoth",
	Jormungars = "Jormungars",
	Julianne = "Julianne",
	["Junk Bot"] = "Junk Bot",
	["Kael'thas Sunstrider"] = "Kael'thas Sunstrider",
	Kalecgos = "Kalecgos",
	["Kam Deepfury"] = "Kam Deepfury",
	["Karsh Steelbender"] = "Karsh Steelbender",
	["Kaz'rogal"] = "Kaz'rogal",
	["Kazkaz the Unholy"] = "Kazkaz the Unholy",
	["Kel'Thuzad"] = "Kel'Thuzad",
	["Keli'dan the Breaker"] = "Keli'dan the Breaker",
	Keristrasza = "Keristrasza",
	["Kiggler the Crazed"] = "Kiggler the Crazed",
	["Kil'jaeden"] = "Kil'jaeden",
	["Kil'rek"] = "Kil'rek",
	["King Dred"] = "King Dred",
	["King Gordok"] = "King Gordok",
	["King Llane Piece"] = "King Llane Piece",
	["King Ymiron"] = "King Ymiron",
	["Kirtonos the Herald"] = "Kirtonos the Herald",
	["Knot Thimblejack's Cache"] = "Knot Thimblejack's Cache",
	Kolk = "Kolk",
	Kologarn = "Kologarn",
	["Koralon the Flame Watcher"] = "Koralon the Flame Watcher",
	Kormok = "Kormok",
	Kresh = "Kresh",
	Krick = "Krick",
	["Krick and Ick"] = "Krick and Ick",
	["Krik'thir the Gatewatcher"] = "Krik'thir the Gatewatcher",
	["Krosh Firehand"] = "Krosh Firehand",
	Krystallus = "Krystallus",
	Kurinnaxx = "Kurinnaxx",
	["Lady Anacondra"] = "Lady Anacondra",
	["Lady Blaumeux"] = "Lady Blaumeux",
	["Lady Deathwhisper"] = "Lady Deathwhisper",
	["Lady Illucia Barov"] = "Lady Illucia Barov",
	["Lady Malande"] = "Lady Malande",
	["Lady Naz'jar"] = "Lady Naz'jar",
	["Lady Sacrolash"] = "Lady Sacrolash",
	["Lady Sarevess"] = "Lady Sarevess",
	["Lady Vashj"] = "Lady Vashj",
	Laj = "Laj",
	Landslide = "Landslide",
	Lavanthor = "Lavanthor",
	["Left Arm"] = "Left Arm",
	["Leotheras the Blind"] = "Leotheras the Blind",
	Lethon = "Lethon",
	Lethtendris = "Lethtendris",
	["Leviathan Mk II"] = "Leviathan Mk II",
	["Ley-Guardian Eregos"] = "Ley-Guardian Eregos",
	["Lieutenant Drake"] = "Lieutenant Drake",
	["Lieutenant General Andorov"] = "Lieutenant General Andorov",
	Loatheb = "Loatheb",
	Lockmaw = "Lockmaw",
	Loken = "Loken",
	["Lord Alexei Barov"] = "Lord Alexei Barov",
	["Lord Cobrahn"] = "Lord Cobrahn",
	["Lord Hel'nurath"] = "Lord Hel'nurath",
	["Lord Incendius"] = "Lord Incendius",
	["Lord Jaraxxus"] = "Lord Jaraxxus",
	["Lord Kazzak"] = "Lord Kazzak",
	["Lord Kri"] = "Lord Kri",
	["Lord Marrowgar"] = "Lord Marrowgar",
	["Lord Pythas"] = "Lord Pythas",
	["Lord Roccor"] = "Lord Roccor",
	["Lord Sanguinar"] = "Lord Sanguinar",
	["Lord Serpentis"] = "Lord Serpentis",
	["Lord Skwol"] = "Lord Skwol",
	["Lord Valthalak"] = "Lord Valthalak",
	["Lord Victor Nefarius"] = "Lord Victor Nefarius",
	["Lord Vyletongue"] = "Lord Vyletongue",
	["Lorekeeper Polkelt"] = "Lorekeeper Polkelt",
	["Lorgus Jett"] = "Lorgus Jett",
	Loro = "Loro",
	Lucifron = "Lucifron",
	["M'uru"] = "M'uru",
	["Mad Magglish"] = "Mad Magglish",
	Maexxna = "Maexxna",
	["Mage-Lord Urom"] = "Mage-Lord Urom",
	["Magister Kalendris"] = "Magister Kalendris",
	["Magistrate Barthilas"] = "Magistrate Barthilas",
	Magmadar = "Magmadar",
	Magmaw = "Magmaw",
	Magmus = "Magmus",
	Magra = "Magra",
	Magtheridon = "Magtheridon",
	["Maiden of Grief"] = "Maiden of Grief",
	["Maiden of Virtue"] = "Maiden of Virtue",
	["Majordomo Executus"] = "Majordomo Executus",
	["Mal'Ganis"] = "Mal'Ganis",
	Malacrass = "Malacrass",
	["Maleki the Pallid"] = "Maleki the Pallid",
	Maloriak = "Maloriak",
	Malygos = "Malygos",
	Maraudos = "Maraudos",
	["Marduk Blackpool"] = "Marduk Blackpool",
	["Marisa du'Paige"] = "Marisa du'Paige",
	Marwyn = "Marwyn",
	["Master Engineer Telonicus"] = "Master Engineer Telonicus",
	["Maur Grimtotem"] = "Maur Grimtotem",
	Meathook = "Meathook",
	["Mechano-Lord Capacitus"] = "Mechano-Lord Capacitus",
	Medivh = "Medivh",
	["Mekgineer Steamrigger"] = "Mekgineer Steamrigger",
	["Mekgineer Thermaplugg"] = "Mekgineer Thermaplugg",
	["Mennu the Betrayer"] = "Mennu the Betrayer",
	["Meshlok the Harvester"] = "Meshlok the Harvester",
	Midnight = "Midnight",
	Mijan = "Mijan",
	Mimiron = "Mimiron",
	["Mindbender Ghur'sha"] = "Mindbender Ghur'sha",
	["Miner Johnson"] = "Miner Johnson",
	["Mistress of Pain"] = "Mistress of Pain",
	Moam = "Moam",
	Mogor = "Mogor",
	["Mokra the Skullcrusher"] = "Mokra the Skullcrusher",
	Moorabi = "Moorabi",
	["Mor Grayhoof"] = "Mor Grayhoof",
	Moragg = "Moragg",
	["Mordresh Fire Eye"] = "Mordresh Fire Eye",
	Moroes = "Moroes",
	["Morogrim Tidewalker"] = "Morogrim Tidewalker",
	Morphaz = "Morphaz",
	["Mother Shahraz"] = "Mother Shahraz",
	["Mother Smolderweb"] = "Mother Smolderweb",
	["Mr. Smite"] = "Mr. Smite",
	["Muradin Bronzebeard"] = "Muradin Bronzebeard",
	["Murkblood Twin"] = "Murkblood Twin",
	["Murkblood Twins"] = "Murkblood Twins",
	Murmur = "Murmur",
	["Murta Grimgut"] = "Murta Grimgut",
	Mushgog = "Mushgog",
	["Mutanus the Devourer"] = "Mutanus the Devourer",
	Nalorakk = "Nalorakk",
	Nazan = "Nazan",
	Nefarian = "Nefarian",
	["Nekrum Gutchewer"] = "Nekrum Gutchewer",
	["Nerub'enkan"] = "Nerub'enkan",
	["Nethermancer Sepethrea"] = "Nethermancer Sepethrea",
	Netherspite = "Netherspite",
	["Netherstrand Longbow"] = "Netherstrand Longbow",
	["Nexus-Prince Shaffar"] = "Nexus-Prince Shaffar",
	Nightbane = "Nightbane",
	["Noth the Plaguebringer"] = "Noth the Plaguebringer",
	["Novos the Summoner"] = "Novos the Summoner",
	Noxxion = "Noxxion",
	["Obsidian Sentinel"] = "Obsidian Sentinel",
	["Odo the Blindwatcher"] = "Odo the Blindwatcher",
	Oggleflint = "Oggleflint",
	["Ogom the Wretched"] = "Ogom the Wretched",
	Ohgan = "Ohgan",
	["Ok'thor the Breaker"] = "Ok'thor the Breaker",
	Olaf = "Olaf",
	["Old Serra'kis"] = "Old Serra'kis",
	["Olm the Summoner"] = "Olm the Summoner",
	["Omor the Unscarred"] = "Omor the Unscarred",
	Onyxia = "Onyxia",
	["Opera Event"] = "Opera Event",
	["Orgrim's Hammer"] = "Orgrim's Hammer",
	["Ormorok the Tree-Shaper"] = "Ormorok the Tree-Shaper",
	["Oro Eyegouge"] = "Oro Eyegouge",
	["Ossirian the Unscarred"] = "Ossirian the Unscarred",
	Ouro = "Ouro",
	["Overlord Ramtusk"] = "Overlord Ramtusk",
	["Overlord Wyrmthalak"] = "Overlord Wyrmthalak",
	["Overmaster Pyron"] = "Overmaster Pyron",
	["Overseer Tidewrath"] = "Overseer Tidewrath",
	Ozruk = "Ozruk",
	Ozumat = "Ozumat",
	Pandemonius = "Pandemonius",
	["Panzor the Invincible"] = "Panzor the Invincible",
	Patchwerk = "Patchwerk",
	["Pathaleon the Calculator"] = "Pathaleon the Calculator",
	Phalanx = "Phalanx",
	["Phaseshift Bulwark"] = "Phaseshift Bulwark",
	Pimgib = "Pimgib",
	["Plaguemaw the Rotting"] = "Plaguemaw the Rotting",
	["Plugger Spazzring"] = "Plugger Spazzring",
	["Postmaster Malown"] = "Postmaster Malown",
	["Priestess Delrissa"] = "Priestess Delrissa",
	["Prince Keleseth"] = "Prince Keleseth",
	["Prince Malchezaar"] = "Prince Malchezaar",
	["Prince Skaldrenox"] = "Prince Skaldrenox",
	["Prince Taldaram"] = "Prince Taldaram",
	["Prince Tenris Mirkblood"] = "Prince Tenris Mirkblood",
	["Prince Tortheldrin"] = "Prince Tortheldrin",
	["Prince Valanar"] = "Prince Valanar",
	["Princess Huhuran"] = "Princess Huhuran",
	["Princess Moira Bronzebeard"] = "Princess Moira Bronzebeard",
	["Princess Tempestria"] = "Princess Tempestria",
	["Princess Theradras"] = "Princess Theradras",
	["Princess Yauj"] = "Princess Yauj",
	["Professor Putricide"] = "Professor Putricide",
	["Pure Spawn of Hydross"] = "Pure Spawn of Hydross",
	Pusillin = "Pusillin",
	["Pyroguard Emberseer"] = "Pyroguard Emberseer",
	["Pyromancer Loregrain"] = "Pyromancer Loregrain",
	Quagmirran = "Quagmirran",
	["Quartermaster Zigris"] = "Quartermaster Zigris",
	["Rage Winterchill"] = "Rage Winterchill",
	Ragglesnout = "Ragglesnout",
	["Raging Spirit"] = "Raging Spirit",
	Ragnaros = "Ragnaros",
	Rajh = "Rajh",
	["Ramstein the Gorger"] = "Ramstein the Gorger",
	["Ras Frostwhisper"] = "Ras Frostwhisper",
	Rattlegore = "Rattlegore",
	["Razorclaw the Butcher"] = "Razorclaw the Butcher",
	["Razorgore the Untamed"] = "Razorgore the Untamed",
	Razorlash = "Razorlash",
	Razorscale = "Razorscale",
	["Reliquary of Souls"] = "Reliquary of Souls",
	Renataki = "Renataki",
	["Restless Skeleton"] = "Restless Skeleton",
	Rethilgore = "Rethilgore",
	Revelosh = "Revelosh",
	["Rhahk'Zor"] = "Rhahk'Zor",
	["Ribbly Screwspigot"] = "Ribbly Screwspigot",
	["Right Arm"] = "Right Arm",
	["Ring of Law"] = "Ring of Law",
	Roar = "Roar",
	["Rokad the Ravager"] = "Rokad the Ravager",
	["Rokdar the Sundered Lord"] = "Rokdar the Sundered Lord",
	["Rokmar the Crackler"] = "Rokmar the Crackler",
	["Rom'ogg Bonecrusher"] = "Rom'ogg Bonecrusher",
	Romulo = "Romulo",
	["Romulo & Julianne"] = "Romulo & Julianne",
	Roogug = "Roogug",
	Rotface = "Rotface",
	Rotgrip = "Rotgrip",
	["Runemaster Molgeim"] = "Runemaster Molgeim",
	["Runok Wildmane"] = "Runok Wildmane",
	Ruuzlu = "Ruuzlu",
	["Salramm the Fleshcrafter"] = "Salramm the Fleshcrafter",
	["Sanctum Sentry"] = "Sanctum Sentry",
	["Sandarr Dunereaver"] = "Sandarr Dunereaver",
	["Sandfury Executioner"] = "Sandfury Executioner",
	Sapphiron = "Sapphiron",
	Sara = "Sara",
	["Saronite Animus"] = "Saronite Animus",
	Sartharion = "Sartharion",
	["Sathrovarr the Corruptor"] = "Sathrovarr the Corruptor",
	["Saviana Ragefire"] = "Saviana Ragefire",
	["Scarlet Commander Mograine"] = "Scarlet Commander Mograine",
	["Scourgelord Tyrannus"] = "Scourgelord Tyrannus",
	["Seeth'rel"] = "Seeth'rel",
	["Selin Fireheart"] = "Selin Fireheart",
	["Sergeant Bly"] = "Sergeant Bly",
	Setesh = "Setesh",
	["Shade of Akama"] = "Shade of Akama",
	["Shade of Aran"] = "Shade of Aran",
	["Shade of Eranikus"] = "Shade of Eranikus",
	["Shadikith the Glider"] = "Shadikith the Glider",
	["Shadow Hunter Vosh'gajin"] = "Shadow Hunter Vosh'gajin",
	["Shadow of Leotheras"] = "Shadow of Leotheras",
	["Shadowpriest Sezz'ziz"] = "Shadowpriest Sezz'ziz",
	Shadron = "Shadron",
	Shazzrah = "Shazzrah",
	["Shirrak the Dead Watcher"] = "Shirrak the Dead Watcher",
	["Siamat, Lord of South Wind"] = "Siamat, Lord of South Wind",
	Sindragosa = "Sindragosa",
	["Sir Zeliek"] = "Sir Zeliek",
	["Sjonnir The Ironshaper"] = "Sjonnir The Ironshaper",
	["Skadi the Ruthless"] = "Skadi the Ruthless",
	["Skarr the Unbreakable"] = "Skarr the Unbreakable",
	["Skarvald the Constructor"] = "Skarvald the Constructor",
	["Skra'gath"] = "Skra'gath",
	Skul = "Skul",
	Skum = "Skum",
	Slabhide = "Slabhide",
	["Slad'ran"] = "Slad'ran",
	Sneed = "Sneed",
	["Sneed's Shredder"] = "Sneed's Shredder",
	["Solakar Flamewreath"] = "Solakar Flamewreath",
	["Solarium Agent"] = "Solarium Agent",
	["Solarium Priest"] = "Solarium Priest",
	["Spirestone Battle Lord"] = "Spirestone Battle Lord",
	["Spirestone Butcher"] = "Spirestone Butcher",
	["Spirestone Lord Magus"] = "Spirestone Lord Magus",
	["Staff of Disintegration"] = "Staff of Disintegration",
	Stalagg = "Stalagg",
	Steelbreaker = "Steelbreaker",
	["Stomper Kreeg"] = "Stomper Kreeg",
	Stonespine = "Stonespine",
	["Stormcaller Brundir"] = "Stormcaller Brundir",
	Strawman = "Strawman",
	["Sulfuron Harbinger"] = "Sulfuron Harbinger",
	Supremus = "Supremus",
	["Svala Sorrowgrave"] = "Svala Sorrowgrave",
	["Swamplord Musel'ek"] = "Swamplord Musel'ek",
	Taerar = "Taerar",
	["Tainted Spawn of Hydross"] = "Tainted Spawn of Hydross",
	["Talon King Ikiss"] = "Talon King Ikiss",
	["Taragaman the Hungerer"] = "Taragaman the Hungerer",
	["Targorr the Dread"] = "Targorr the Dread",
	Tavarok = "Tavarok",
	Techbot = "Techbot",
	["Temple Guardian Anhuur"] = "Temple Guardian Anhuur",
	Temporus = "Temporus",
	["Tendris Warpwood"] = "Tendris Warpwood",
	Tenebron = "Tenebron",
	["Terestian Illhoof"] = "Terestian Illhoof",
	["Teron Gorefiend"] = "Teron Gorefiend",
	Thaddius = "Thaddius",
	["Thaladred the Darkener"] = "Thaladred the Darkener",
	["Thane Korth'azz"] = "Thane Korth'azz",
	["The Beast"] = "The Beast",
	["The Beasts of Northrend"] = "The Beasts of Northrend",
	["The Big Bad Wolf"] = "The Big Bad Wolf",
	["The Black Knight"] = "The Black Knight",
	["The Black Stalker"] = "The Black Stalker",
	["The Blue Brothers"] = "The Blue Brothers",
	["The Bug Family"] = "The Bug Family",
	["The Crone"] = "The Crone",
	["The Curator"] = "The Curator",
	["The Eredar Twins"] = "The Eredar Twins",
	["The Four Horsemen"] = "The Four Horsemen",
	["The Illidari Council"] = "The Illidari Council",
	["The Iron Council"] = "The Iron Council",
	["The Lich King"] = "The Lich King",
	["The Lost Dwarves"] = "The Lost Dwarves",
	["The Lurker Below"] = "The Lurker Below",
	["The Maker"] = "The Maker",
	["The Prophet Skeram"] = "The Prophet Skeram",
	["The Prophet Tharon'ja"] = "The Prophet Tharon'ja",
	["The Ravenian"] = "The Ravenian",
	["The Razza"] = "The Razza",
	["The Seven Dwarves"] = "The Seven Dwarves",
	["The Skybreaker"] = "The Skybreaker",
	["The Tribunal of Ages"] = "The Tribunal of Ages",
	["The Twin Emperors"] = "The Twin Emperors",
	["The Twin Val'kyr"] = "The Twin Val'kyr",
	["The Unforgiven"] = "The Unforgiven",
	["The Windreaver"] = "The Windreaver",
	["Theka the Martyr"] = "Theka the Martyr",
	Thorim = "Thorim",
	["Thorngrin the Tender"] = "Thorngrin the Tender",
	["Tidewalker Lurker"] = "Tidewalker Lurker",
	["Timmy the Cruel"] = "Timmy the Cruel",
	Tinhead = "Tinhead",
	["Tinkerer Gizlock"] = "Tinkerer Gizlock",
	["Tirion Fordring"] = "Tirion Fordring",
	Tito = "Tito",
	["Toravon the Ice Watcher"] = "Toravon the Ice Watcher",
	["Trigore the Lasher"] = "Trigore the Lasher",
	Trollgore = "Trollgore",
	["Tsu'zee"] = "Tsu'zee",
	["Tuten'kash"] = "Tuten'kash",
	["Twilight Lord Kelris"] = "Twilight Lord Kelris",
	["Urok Doomhowl"] = "Urok Doomhowl",
	["VX-001"] = "VX-001",
	["Vaelastrasz the Corrupt"] = "Vaelastrasz the Corrupt",
	["Val'kyr Shadowguard"] = "Val'kyr Shadowguard",
	["Valithria Dreamwalker"] = "Valithria Dreamwalker",
	["Varian Wrynn"] = "Varian Wrynn",
	["Varos Cloudstrider"] = "Varos Cloudstrider",
	Vazruden = "Vazruden",
	["Vazruden the Herald"] = "Vazruden the Herald",
	Vectus = "Vectus",
	Vem = "Vem",
	Veng = "Veng",
	["Veras Darkshadow"] = "Veras Darkshadow",
	["Verdan the Everliving"] = "Verdan the Everliving",
	Verek = "Verek",
	Vesperon = "Vesperon",
	Vexallus = "Vexallus",
	["Veyzhak the Cannibal"] = "Veyzhak the Cannibal",
	["Vile'rel"] = "Vile'rel",
	Viscidus = "Viscidus",
	["Viscous Fallout"] = "Viscous Fallout",
	["Void Reaver"] = "Void Reaver",
	Volkhan = "Volkhan",
	["War Master Voone"] = "War Master Voone",
	["Warbringer O'mrogg"] = "Warbringer O'mrogg",
	["Warchief Blackhand Piece"] = "Warchief Blackhand Piece",
	["Warchief Kargath Bladefist"] = "Warchief Kargath Bladefist",
	["Warchief Rend Blackhand"] = "Warchief Rend Blackhand",
	["Warden Mellichar"] = "Warden Mellichar",
	["Warder Stilgiss"] = "Warder Stilgiss",
	["Warlord Kalithresh"] = "Warlord Kalithresh",
	["Warmaul Champion"] = "Warmaul Champion",
	["Warp Slicer"] = "Warp Slicer",
	["Warp Splinter"] = "Warp Splinter",
	["Watchkeeper Gargolmar"] = "Watchkeeper Gargolmar",
	Weaver = "Weaver",
	["Witch Doctor Zum'rah"] = "Witch Doctor Zum'rah",
	["Wolf Master Nandos"] = "Wolf Master Nandos",
	["Wrath-Scryer Soccothrates"] = "Wrath-Scryer Soccothrates",
	Wushoolay = "Wushoolay",
	["XT-002 Deconstructor"] = "XT-002 Deconstructor",
	Xevozz = "Xevozz",
	["Yogg-Saron"] = "Yogg-Saron",
	Yor = "Yor",
	Ysondre = "Ysondre",
	Zekkis = "Zekkis",
	["Zelemar the Wrathful"] = "Zelemar the Wrathful",
	["Zereketh the Unbound"] = "Zereketh the Unbound",
	Zerillis = "Zerillis",
	["Zevrim Thornhoof"] = "Zevrim Thornhoof",
	Zolo = "Zolo",
	["Zul'Farrak Dead Hero"] = "Zul'Farrak Dead Hero",
	["Zul'Lor"] = "Zul'Lor",
	["Zul'jin"] = "Zul'jin",
	["Zul'tore"] = "Zul'tore",
	["Zuramat the Obliterator"] = "Zuramat the Obliterator",
}


if GAME_LOCALE == "deDE" then
	lib:SetCurrentTranslations {
	Acidmaw = "Ätzschlund",
	Aeonus = "Aeonus",
	["Aerial Command Unit"] = "Luftkommandoeinheit",
	["Agathelos the Raging"] = "Agathelos der Tobende",
	["Aggem Thorncurse"] = "Aggem Dornfluch",
	Ahune = "Ahune",
	["Akil'zon"] = "Akil'zon",
	["Aku'mai"] = "Aku'mai",
	["Al'ar"] = "Al'ar",
	["Algalon the Observer"] = "Algalon der Beobachter",
	Altarius = "Altarius",
	["Alzzin the Wildshaper"] = "Alzzin der Wildformer",
	Amanitar = "Amanitar",
	["Ambassador Flamelash"] = "Botschafter Flammenschlag",
	["Ambassador Hellmaw"] = "Botschafter Höllenschlund",
	Ammunae = "Ammunae",
	["Amnennar the Coldbringer"] = "Amnennar der Kältebringer",
	["Ancient Stone Keeper"] = "Uralter Steinbewahrer",
	Anetheron = "Anetheron",
	["Anger'rel"] = "Anger'rel",
	Anomalus = "Anomalus",
	Anraphet = "Anraphet",
	["Antu'sul"] = "Antu'sul",
	["Anub'Rekhan"] = "Anub'Rekhan",
	["Anub'arak"] = "Anub'arak",
	["Anub'shiah"] = "Anub'shiah",
	["Anubisath Defender"] = "Verteidiger des Anubisath",
	["Anubisath Guardian"] = "Beschützer des Anubisath",
	Anzu = "Anzu",
	["Arcane Watchman"] = "Arkanwachmann",
	["Arcanist Doan"] = "Arkanist Doan",
	Archaedas = "Archaedas",
	["Archavon the Stone Watcher"] = "Archavon der Steinwächter",
	Archimonde = "Archimonde",
	["Archivist Galford"] = "Archivar Galford",
	["Archmage Arugal"] = "Erzmagier Arugal",
	["Argent Confessor Paletress"] = "Argentumbeichtpatin Blondlocke",
	["Arugal's Voidwalker"] = "Arugals Leerwandler",
	["Ascendant Lord Obsidius"] = "Aszendentenfürst Obsidius",
	["Asim al Akir"] = "Asaad",
	["Assault Bot"] = "Angriffsbot",
	["Assembly of Iron"] = "Versammlung des Eisens",
	["Atal'alarion"] = "Atal'alarion",
	-- Atramedes = "",
	["Attumen the Huntsman"] = "Attumen der Jäger",
	Auriaya = "Auriaya",
	Avalanchion = "Avalanchion",
	["Avatar of Hakkar"] = "Avatar von Hakkar",
	["Ayamiss the Hunter"] = "Ayamiss der Jäger",
	Azgalor = "Azgalor",
	["Azshir the Sleepless"] = "Azshir der Schlaflose",
	Azuregos = "Azuregos",
	["Bael'Gar"] = "Bael'Gar",
	Baelog = "Baelog",
	Balnazzar = "Balnazzar",
	["Baltharus the Warborn"] = "Baltharus der Kriegsjünger",
	["Bannok Grimaxe"] = "Bannok Grimmaxt",
	["Baron Aquanis"] = "Baron Aquanis",
	["Baron Charr"] = "Baron Glutarr",
	["Baron Geddon"] = "Baron Geddon",
	["Baron Kazum"] = "Baron Kazum",
	["Baron Rivendare"] = "Baron Totenschwur",
	["Baron Silverlaine"] = "Baron Silberlein",
	["Baroness Anastari"] = "Baroness Anastari",
	["Battleguard Sartura"] = "Schlachtwache Sartura",
	["Bazil Thredd"] = "Bazil Thredd",
	Bazzalan = "Bazzalan",
	Beauty = "Stutzer",
	["Black Guard Swordsmith"] = "Schwertschmied der schwarzen Wache",
	["Blackheart the Inciter"] = "Schwarzherz der Hetzer",
	["Blind Hunter"] = "Blinder Jäger",
	["Blindeye the Seer"] = "Blindauge der Seher",
	["Blood Guard Porung"] = "Blutwache Porung",
	["Blood Prince Council"] = "Rat der Blutprinzen",
	["Blood Princes"] = "Rat der Blutprinzen",
	["Blood Steward of Kirtonos"] = "Blutdiener von Kirtonos",
	["Blood-Queen Lana'thel"] = "Blutkönigin Lana'thel",
	["Bloodlord Mandokir"] = "Blutfürst Mandokir",
	["Bloodmage Thalnos"] = "Blutmagier Thalnos",
	Boahn = "Boahn",
	["Bomb Bot"] = "Bombenbot",
	["Brain of Yogg-Saron"] = "Yogg-Sarons Gehirn",
	["Brainwashed Noble"] = "Manipulierter Adliger",
	Broggok = "Broggok",
	Brokentoe = "Schmetterzehe",
	Bronjahm = "Bronjahm",
	["Broodlord Lashlayer"] = "Brutwächter Dreschbringer",
	["Bruegal Ironknuckle"] = "Bruegal Eisenfaust",
	Brutallus = "Brutallus",
	["Burning Felguard"] = "Brennende Teufelswache",
	["Buru the Gorger"] = "Buru der Verschlinger",
	["C'Thun"] = "C'Thun",
	["Cache of the Firelord"] = "Truhe des Feuerlords",
	["Cache of the Legion"] = "Behälter der Legion",
	["Cannon Master Willey"] = "Kanonenmeister Willey",
	["Captain Greenskin"] = "Kapitän Grünhaut",
	["Captain Kromcrush"] = "Hauptmann Krombruch",
	["Captain Skarloc"] = "Kapitän Skarloc",
	["Celebras the Cursed"] = "Celebras der Verfluchte",
	["Charlga Razorflank"] = "Charlga Klingenflanke",
	["Chess Event"] = "Chess Event",
	["Chest of The Seven"] = "Truhe der Sieben",
	["Chief Ukorz Sandscalp"] = "Häuptling Ukorz Sandwüter",
	-- Chimaeron = "",
	["Cho'Rush the Observer"] = "Cho'Rush der Beobachter",
	Chromaggus = "Chromaggus",
	["Chrono Lord Deja"] = "Chronolord Deja",
	["Chrono-Lord Epoch"] = "Chronolord Epoch",
	Claw = "Klaue",
	["Coilfang Elite"] = "Elitesoldat des Echsenkessels",
	["Coilfang Strider"] = "Schreiter des Echsenkessels",
	["Commander Kolurg"] = "Kommandant Kolurg",
	["Commander Sarannis"] = "Kommandant Sarannis",
	["Commander Springvale"] = "Kommandant Springvale",
	["Commander Stoutbeard"] = "Kommandant Starkbart",
	["Commander Ulthok"] = "Kommandant Ulthok",
	["Constructor & Controller"] = "Konstrukteur & Kontrolleur",
	Cookie = "Krümel",
	Corborus = "Corborus",
	["Coren Direbrew"] = "Coren Düsterbräu",
	["Corla, Herald of Twilight"] = "Corla, Botin des Zwielichts",
	["Cosmic Infuser"] = "Kosmische Macht",
	["Crimson Hammersmith"] = "Purpurroter Hammerschmied",
	["Crowd Pummeler 9-60"] = "Meuteverprügler 9-60",
	["Crystal Fang"] = "Kristallfangzahn",
	Cyanigosa = "Cyanigosa",
	["Dalliah the Doomsayer"] = "Dalliah die Verdammnisverkünderin",
	["Dalronn the Controller"] = "Dalronn der Aufseher",
	["Dark Coffer"] = "Dunkler Kasten",
	["Dark Iron Ambassador"] = "Botschafter der Dunkeleisenzwerge",
	["Darkmaster Gandling"] = "Dunkelmeister Gandling",
	["Darkweaver Syth"] = "Dunkelwirker Syth",
	["Death Knight Darkreaver"] = "Todesritter Schattensichel",
	["Death Knight Understudy"] = "Reservist der Todesritter",
	["Death Speaker Jargba"] = "Todessprecher Jargba",
	["Deathbound Ward"] = "Todesgeweihter Wächter",
	["Deathbringer Saurfang"] = "Todesbringer Saurfang",
	["Deathspeaker High Priest"] = "Hohepriester der Todessprecher",
	["Deathstalker Visceri"] = "Todespirscher Visceri",
	["Deathsworn Captain"] = "Todeshöriger Captain",
	Devastation = "Verwüstung",
	["Deviate Faerie Dragon"] = "Deviatfeendrache",
	["Devourer of Souls"] = "Verschlinger der Seelen",
	["Dextren Ward"] = "Dextren Ward",
	["Digmaster Shovelphlange"] = "Grubenmeister Schaufelphlansch",
	["Doctor Theolen Krastinov"] = "Doktor Theolen Krastinov",
	["Doom Lord Kazzak"] = "Verdammnislord Kazzak",
	["Doom'rel"] = "Un'rel",
	Doomwalker = "Verdammniswandler",
	["Dope'rel"] = "Trott'rel",
	Dorothee = "Dorothee",
	["Drahga Shodowburner"] = "Drahga Schattenbrenner",
	["Drakkari Colossus"] = "Koloss der Drakkari",
	["Drakos the Interrogator"] = "Drakos der Befrager",
	Dreadscale = "Schreckensmaul",
	Dreamscythe = "Traumsense",
	["Dust Covered Chest"] = "Staub Bedeckter Kasten",
	Dustwraith = "Karaburan",
	["Eadric the Pure"] = "Eadric der Reine",
	["Earthcaller Halmgar"] = "Erdenrufer Halmgar",
	["Earthrager Ptah"] = "Erdwüter Ptah",
	Ebonroc = "Schattenschwinge",
	["Eck the Ferocious"] = "Der wilde Eck",
	["Edwin VanCleef"] = "Edwin van Cleef",
	["Elder Brightleaf"] = "Ältester Hellblatt",
	["Elder Ironbranch"] = "Ältester Eisenast",
	["Elder Nadox"] = "Urahne Nadox",
	["Elder Stonebark"] = "Ältester Steinrinde",
	["Electrocutioner 6000"] = "Elektrokutor 6000",
	["Emalon the Storm Watcher"] = "Emalon der Sturmwächter",
	Emeriss = "Smariss",
	["Emperor Dagran Thaurissan"] = "Imperator Dagran Thaurissan",
	["Emperor Vek'lor"] = "Imperator Vek'lor",
	["Emperor Vek'nilash"] = "Imperator Vek'nilash",
	Entropius = "Entropius",
	["Eonar's Gift"] = "Eonars Geschenk",
	["Epoch Hunter"] = "Epochenjäger",
	Erekem = "Erekem",
	["Eressea Dawnsinger"] = "Eressea Morgensänger",
	["Eric \"The Swift\""] = "Eric \"Der Flinke\"",
	Erudax = "Erudax",
	["Erunak Stonespeaker"] = "Erunak Steinsprecher",
	["Essence of Anger"] = "Essenz des Zorns",
	["Essence of Desire"] = "Essenz der Begierde",
	["Essence of Suffering"] = "Essenz des Leidens",
	Eviscerator = "Ausweider",
	["Exarch Maladaar"] = "Exarch Maladaar",
	["Expedition Commander"] = "Kommandant der Expedition",
	["Eydis Darkbane"] = "Eydis Nachtbann",
	["Eye of C'Thun"] = "Auge von C'Thun",
	["Faction Champions"] = "Fraktionschampions",
	["Fallen Champion"] = "Gestürzter Held",
	Falric = "Falric",
	["Falric and Marwyn"] = "Falric und Marwyn",
	["Fankriss the Unyielding"] = "Fankriss der Unnachgiebige",
	["Fathom-Lord Karathress"] = "Tiefenlord Karathress",
	Felmyst = "Teufelsruch",
	["Fenrus the Devourer"] = "Fenrus der Verschlinger",
	["Feral Defender"] = "Wilder Verteidiger",
	Festergut = "Fauldarm",
	Feugen = "Feugen",
	["Fineous Darkvire"] = "Fineous Dunkelader",
	Firemaw = "Feuerschwinge",
	["Fjola Lightbane"] = "Fjola Lichtbann",
	["Flame Leviathan"] = "Flammenleviathan",
	Flamegor = "Flammenmaul",
	["Foreman Thistlenettle"] = "Großknecht Distelklette",
	["Forgemaster Garfrost"] = "Schmiedemeister Garfrost",
	-- ["Forgemaster Throngus"] = "",
	["Four Horsemen Chest"] = "Die Vier Reiter Kiste",
	["Fras Siabi"] = "Fras Siabi",
	Freya = "Freya",
	["Gahz'ranka"] = "Gahz'ranka",
	["Gahz'rilla"] = "Gahz'rilla",
	["Gal'darah"] = "Gal'darah",
	["Galgann Firehammer"] = "Galgann Feuerhammer",
	Garr = "Garr",
	["Garrosh Hellscream"] = "Garrosh Höllschrei",
	Gasher = "Schlitzer",
	["Gatewatcher Gyro-Kill"] = "Torwächter Gyrotot",
	["Gatewatcher Iron-Hand"] = "Torwächter Eisenhand",
	["Gathios the Shatterer"] = "Gathios der Zerschmetterer",
	Gehennas = "Gehennas",
	Gelihast = "Gelihast",
	Gelk = "Gelk",
	["General Angerforge"] = "General Zornesschmied",
	["General Bjarngrim"] = "General Bjarngrim",
	["General Drakkisath"] = "General Drakkisath",
	["General Husam"] = "General Husam",
	["General Rajaxx"] = "General Rajaxx",
	["General Umbriss"] = "General Umbriss",
	["General Vezax"] = "General Vezax",
	["General Zarithrian"] = "General Zarithrian",
	["Ghamoo-ra"] = "Ghamoo-ra",
	["Ghaz'an"] = "Ghaz'an",
	["Ghok Bashguud"] = "Ghok Haudrauf",
	Gilnid = "Gilnid",
	["Gizrul the Slavener"] = "Gizrul der Geifernde",
	["Gloom'rel"] = "Dunk'rel",
	Gluth = "Gluth",
	Glutton = "Nimmersatt",
	["Golem Lord Argelmach"] = "Golemlord Argelmach",
	["Golemagg the Incinerator"] = "Golemagg der Verbrenner",
	["Goraluk Anvilcrack"] = "Goraluk Hammerbruch",
	["Gormok the Impaler"] = "Gormok der Pfähler",
	["Gorosh the Dervish"] = "Gorosh der Derwisch",
	["Gortok Palehoof"] = "Gortok Bleichhuf",
	["Gothik the Harvester"] = "Gothik der Ernter",
	["Grand Astromancer Capernian"] = "Großastronom Capernian",
	["Grand Champions"] = "Großchampions",
	["Grand Crusader Dathrohan"] = "Oberster Kreuzfahrer Dathrohan",
	["Grand Magus Telestra"] = "Großmagistrix Telestra",
	["Grand Vizier Ertan"] = "Großwesir Ertan",
	["Grand Warlock Alythess"] = "Großhexenmeisterin Alythess",
	["Grand Warlock Nethekurse"] = "Großhexenmeister Nethekurse",
	["Grand Widow Faerlina"] = "Großwitwe Faerlina",
	["Grandmaster Vorpil"] = "Großmeister Vorpil",
	Grandmother = "Großmutter",
	["Grethok the Controller"] = "Grethok der Aufseher",
	["Gri'lek"] = "Gri'lek",
	Grimlok = "Grimlok",
	Grizzle = "Grizzle",
	Grobbulus = "Grobbulus",
	Grubbis = "Grubbis",
	["Gruul the Dragonkiller"] = "Gruul der Drachenschlächter",
	["Guard Fengus"] = "Wache Fengus",
	["Guard Mol'dar"] = "Wache Mol'dar",
	["Guard Slip'kik"] = "Wache Slip'kik",
	["Guardian of Yogg-Saron"] = "Wächter des Yogg-Saron",
	["Gurtogg Bloodboil"] = "Gurtogg Siedeblut",
	Gyth = "Gyth",
	Hadronox = "Hadronox",
	Hakkar = "Hakkar",
	Halazzi = "Halazzi",
	Halion = "Halion",
	Halycon = "Halycon",
	Hamhock = "Hamhock",
	["Harbinger Skyriss"] = "Herold Horizontiss",
	["Hate'rel"] = "Hass'rel",
	["Hazza'rah"] = "Hazza'rah",
	Hazzas = "Hazzas",
	["Headless Horseman"] = "Der kopflose Reiter",
	["Hearthsinger Forresten"] = "Herdsinger Forresten",
	["Hedrum the Creeper"] = "Hedrum der Krabbler",
	["Heigan the Unclean"] = "Heigan der Unreine",
	["Hellfire Channeler"] = "Kanalisierer des Höllenfeuers",
	["Henry Stern"] = "Henry Stern",
	["Herald Volazj"] = "Herold Volazj",
	Herod = "Herod",
	["Hex Lord Malacrass"] = "Hexlord Malacrass",
	["High Astromancer Solarian"] = "Hochastromantin Solarian",
	["High Botanist Freywinn"] = "Hochbotaniker Freywinn",
	["High Inquisitor Fairbanks"] = "Hochinquisitor Fairbanks",
	["High Inquisitor Whitemane"] = "Hochinquisitor Weißsträhne",
	["High Interrogator Gerstahn"] = "Verhörmeisterin Gerstahn",
	["High King Maulgar"] = "Hochkönig Maulgar",
	["High Marshal Whirlaxis"] = "Hochmarschall Whirlaxis",
	["High Nethermancer Zerevor"] = "Hochnethermant Zerevor",
	["High Overlord Saurfang"] = "Hochfürst Saurfang",
	["High Priest Thekal"] = "Hohepriester Thekal",
	["High Priest Venoxis"] = "Hohepriester Venoxis",
	["High Priestess Arlokk"] = "Hohepriesterin Arlokk",
	["High Priestess Azil"] = "Hohepriesterin Azil",
	["High Priestess Jeklik"] = "Hohepriesterin Jeklik",
	["High Priestess Mar'li"] = "Hohepriesterin Mar'li",
	["High Priestess of Thaurissan"] = "\009Hohepriesterin von Thaurissan",
	["High Prophet Barim"] = "Hochprophet Barim",
	["High Warlord Naj'entus"] = "Oberster Kriegsfürst Naj'entus",
	["Highlord Mograine"] = "Hochlord Mograine",
	["Highlord Omokk"] = "Hochlord Omokk",
	Hodir = "Hodir",
	["Houndmaster Grebmar"] = "Hundemeister Grebmar",
	["Houndmaster Loksey"] = "Hundemeister Loksey",
	Hukku = "Hukku",
	Hungarfen = "Hungarfenn",
	["Hurley Blackbreath"] = "Hurley Pestatem",
	["Hyakiss the Lurker"] = "Hyakiss der Lauerer",
	["Hydromancer Thespia"] = "Wasserbeschwörerin Thespia",
	["Hydromancer Velratha"] = "Wasserbeschwörerin Velratha",
	Hydrospawn = "Hydrobrut",
	["Hydross the Unstable"] = "Hydross der Unstete",
	["Ice Sphere"] = "Eissphäre",
	["Icecrown Gunship Battle"] = "Luftschlacht um die Eiskronenzitadelle",
	Icehowl = "Eisheuler",
	Ichoron = "Ichoron",
	Ick = "Ick",
	["Ignis the Furnace Master"] = "Ignis, Meister des Eisenwerks",
	["Illidan Stormrage"] = "Illidan Sturmgrimm",
	["Illidari Council"] = "Rat der Illidari",
	["Illyanna Ravenoak"] = "Illyanna Rabeneiche",
	["Immol'thar"] = "Immol'thar",
	["Infinite Corruptor"] = "Ewiger Verderber",
	["Infinity Blades"] = "Klinge der Unendlichkeit",
	["Ingvar the Plunderer"] = "Ingvar der Brandschatzer",
	["Instructor Malicia"] = "Instrukteurin Malicia",
	["Instructor Razuvious"] = "Instrukteur Razuvious",
	["Interrogator Vishas"] = "Befrager Vishas",
	Ionar = "Ionar",
	Ironaya = "Ironaya",
	Ironspine = "Eisenrücken",
	Isalien = "Isalien",
	Isiset = "Isiset",
	Jade = "Jade",
	["Jammal'an the Prophet"] = "Jammal'an der Prophet",
	["Jan'alai"] = "Jan'alai",
	["Jandice Barov"] = "Jandice Barov",
	["Jed Runewatcher"] = "Jed Runenblick",
	["Jedoga Shadowseeker"] = "Jedoga Schattensucher",
	["Jergosh the Invoker"] = "Jergosh der Herbeirufer",
	["Jin'do the Hexxer"] = "Jin'do der Verhexer",
	["Jormungar Behemoth"] = "Jormungarungetüm",
	Jormungars = "Jormungars",
	Julianne = "Julianne",
	["Junk Bot"] = "Schrottbot",
	["Kael'thas Sunstrider"] = "Kael'thas Sonnenwanderer",
	Kalecgos = "Kalecgos",
	["Kam Deepfury"] = "Kam Tiefenzorn",
	["Karsh Steelbender"] = "Karsh Stahlbieger",
	["Kaz'rogal"] = "Kaz'rogal",
	["Kazkaz the Unholy"] = "Kazkaz der Unheilige",
	["Kel'Thuzad"] = "Kel'Thuzad",
	["Keli'dan the Breaker"] = "Keli'dan der Zerstörer",
	Keristrasza = "Keristrasza",
	["Kiggler the Crazed"] = "Kiggler the Crazed",
	["Kil'jaeden"] = "Kil'jaeden",
	["Kil'rek"] = "Kil'rek",
	["King Dred"] = "König Dred",
	["King Gordok"] = "König Gordok",
	["King Llane Piece"] = "König Llane",
	["King Ymiron"] = "König Ymiron",
	["Kirtonos the Herald"] = "Kirtonos der Herold",
	["Knot Thimblejack's Cache"] = "Knot Thimblejacks Truhe",
	Kolk = "Kolk",
	Kologarn = "Kologarn",
	["Koralon the Flame Watcher"] = "Koralon der Flammenwächter",
	Kormok = "Kormok",
	Kresh = "Kresh",
	Krick = "Krick",
	["Krick and Ick"] = "Krick und Ick",
	["Krik'thir the Gatewatcher"] = "Krik'thir der Torwächter",
	["Krosh Firehand"] = "Krosh Feuerhand",
	Krystallus = "Krystallus",
	Kurinnaxx = "Kurinnaxx",
	["Lady Anacondra"] = "Lady Anacondra",
	["Lady Blaumeux"] = "Lady Blaumeux",
	["Lady Deathwhisper"] = "Lady Todeswisper",
	["Lady Illucia Barov"] = "Lady Illucia Barov",
	["Lady Malande"] = "Lady Malande",
	["Lady Naz'jar"] = "Lady Naz'jar",
	["Lady Sacrolash"] = "Lady Sacrolash",
	["Lady Sarevess"] = "Lady Sarevess",
	["Lady Vashj"] = "Lady Vashj",
	Laj = "Laj",
	Landslide = "Erdrutsch",
	Lavanthor = "Lavanthor",
	["Left Arm"] = "Linker Arm",
	["Leotheras the Blind"] = "Leotheras der Blinde",
	Lethon = "Lethon",
	Lethtendris = "Lethtendris",
	["Leviathan Mk II"] = "Leviathan Mk II",
	["Ley-Guardian Eregos"] = "Leywächter Eregos",
	["Lieutenant Drake"] = "Leutnant Drach",
	["Lieutenant General Andorov"] = "Generallieutenant Andorov",
	Loatheb = "Loatheb",
	Lockmaw = "Maulsperre",
	Loken = "Loken",
	["Lord Alexei Barov"] = "Lord Alexei Barov",
	["Lord Cobrahn"] = "Lord Kobrahn",
	["Lord Hel'nurath"] = "Lord Hel'nurath",
	["Lord Incendius"] = "Lord Incendius",
	["Lord Jaraxxus"] = "Lord Jaraxxus",
	["Lord Kazzak"] = "Lord Kazzak",
	["Lord Kri"] = "Lord Kri",
	["Lord Marrowgar"] = "Lord Mark'gar",
	["Lord Pythas"] = "Lord Pythas",
	["Lord Roccor"] = "Lord Roccor",
	["Lord Sanguinar"] = "Fürst Blutdurst",
	["Lord Serpentis"] = "Lord Serpentis",
	["Lord Skwol"] = "Lord Skwol",
	["Lord Valthalak"] = "Lord Valthalak",
	["Lord Victor Nefarius"] = "Lord Victor Nefarius",
	["Lord Vyletongue"] = "Lord Schlangenzunge",
	["Lorekeeper Polkelt"] = "Hüter des Wissens Polkelt",
	["Lorgus Jett"] = "Lorgus Jett",
	Loro = "Loro",
	Lucifron = "Lucifron",
	["M'uru"] = "M'uru",
	["Mad Magglish"] = "Zausel der Verrückte",
	Maexxna = "Maexxna",
	["Mage-Lord Urom"] = "Magierlord Urom",
	["Magister Kalendris"] = "Magister Kalendris",
	["Magistrate Barthilas"] = "Magistrat Barthilas",
	Magmadar = "Magmadar",
	-- Magmaw = "",
	Magmus = "Magmus",
	Magra = "Magra",
	Magtheridon = "Magtheridon",
	["Maiden of Grief"] = "Maid der Trauer",
	["Maiden of Virtue"] = "Tugendhafte Maid",
	["Majordomo Executus"] = "Majordomus Exekutus",
	["Mal'Ganis"] = "Mal'Ganis",
	Malacrass = "Malacrass",
	["Maleki the Pallid"] = "Maleki der Leichenblasse",
	-- Maloriak = "",
	Malygos = "Malygos",
	Maraudos = "Maraudos",
	["Marduk Blackpool"] = "Marduk Blackpool",
	["Marisa du'Paige"] = "Marisa du'Paige",
	Marwyn = "Marwyn",
	["Master Engineer Telonicus"] = "Meisteringenieur Telonicus",
	["Maur Grimtotem"] = "Maur Grimmtotem",
	Meathook = "Fleischhaken",
	["Mechano-Lord Capacitus"] = "Mechanolord Kapazitus",
	Medivh = "Medivh",
	["Mekgineer Steamrigger"] = "Robogenieur Dampfhammer",
	["Mekgineer Thermaplugg"] = "Robogenieur Thermadraht",
	["Mennu the Betrayer"] = "Mennu der Verräter",
	["Meshlok the Harvester"] = "Meshlok der Ernter",
	Midnight = "Mittnacht",
	Mijan = "Mijan",
	Mimiron = "Mimiron",
	["Mindbender Ghur'sha"] = "Geistbeuger Ghur'sha",
	["Miner Johnson"] = "Minenarbeiter Johnson",
	["Mistress of Pain"] = "Herrin der Schmerzen",
	Moam = "Moam",
	Mogor = "Mogor",
	["Mokra the Skullcrusher"] = "Mokra der Schädelberster",
	Moorabi = "Moorabi",
	["Mor Grayhoof"] = "Mor Grauhuf",
	Moragg = "Moragg",
	["Mordresh Fire Eye"] = "Mordresh Feuerauge",
	Moroes = "Moroes",
	["Morogrim Tidewalker"] = "Morogrim Gezeitenwandler",
	Morphaz = "Morphaz",
	["Mother Shahraz"] = "Mutter Shahraz",
	["Mother Smolderweb"] = "Mutter Glimmernetz",
	["Mr. Smite"] = "Handlanger Pein",
	["Muradin Bronzebeard"] = "Muradin Bronzebart",
	["Murkblood Twin"] = "Zwilling der Finsterblut",
	["Murkblood Twins"] = "Zwillinge der Finsterblut",
	Murmur = "Murmur",
	["Murta Grimgut"] = "Murta Bauchgrimm",
	Mushgog = "Mushgog",
	["Mutanus the Devourer"] = "Mutanus der Verschlinger",
	Nalorakk = "Nalorakk",
	Nazan = "Nazan",
	Nefarian = "Nefarian",
	["Nekrum Gutchewer"] = "Nekrum der Ausweider",
	["Nerub'enkan"] = "Nerub'enkan",
	["Nethermancer Sepethrea"] = "Nethermant Sepethrea",
	Netherspite = "Nethergroll",
	["Netherstrand Longbow"] = "Netherbespannter Langbogen",
	["Nexus-Prince Shaffar"] = "Nexusprinz Shaffar",
	Nightbane = "Schrecken der Nacht",
	["Noth the Plaguebringer"] = "Noth der Seuchenfürst",
	["Novos the Summoner"] = "Novos der Beschwörer",
	Noxxion = "Noxxion",
	["Obsidian Sentinel"] = "Obsidianschildwache",
	["Odo the Blindwatcher"] = "Odo der Blindseher",
	Oggleflint = "Flintauge",
	["Ogom the Wretched"] = "Ogom der Elende",
	Ohgan = "Ohgan",
	["Ok'thor the Breaker"] = "Ok'thor der Zerstörer",
	Olaf = "Olaf",
	["Old Serra'kis"] = "Old Serra'kis",
	["Olm the Summoner"] = "Olm der Beschwörer",
	["Omor the Unscarred"] = "Omor der Narbenlose",
	Onyxia = "Onyxia",
	["Opera Event"] = "Opernevent",
	["Orgrim's Hammer"] = "Orgrims Hammer",
	["Ormorok the Tree-Shaper"] = "Ormorok der Baumformer",
	["Oro Eyegouge"] = "Oro Hohlauge",
	["Ossirian the Unscarred"] = "Ossirian der Narbenlose",
	Ouro = "Ouro",
	["Overlord Ramtusk"] = "Oberanführer Rammhauer",
	["Overlord Wyrmthalak"] = "Oberanführer Wyrmthalak",
	["Overmaster Pyron"] = "Übermeister Pyron",
	["Overseer Tidewrath"] = "Overseer Tidewrath",
	Ozruk = "Ozruk",
	Ozumat = "Ozumat",
	Pandemonius = "Pandemonius",
	["Panzor the Invincible"] = "Panzor der Unbesiegbare",
	Patchwerk = "Flickwerk",
	["Pathaleon the Calculator"] = "Pathaleon der Kalkulator",
	Phalanx = "Phalanx",
	["Phaseshift Bulwark"] = "Phasenverschobenes Bollwerk",
	Pimgib = "Pimgib",
	["Plaguemaw the Rotting"] = "Seuchenschlund der Faulende",
	["Plugger Spazzring"] = "Stöpsel Zapfring",
	["Postmaster Malown"] = "Postmeister Malown",
	["Priestess Delrissa"] = "Priesterin Delrissa",
	["Prince Keleseth"] = "Prinz Keleseth",
	["Prince Malchezaar"] = "Prinz Malchezaar",
	["Prince Skaldrenox"] = "Prince Skaldrenox",
	["Prince Taldaram"] = "Prinz Taldaram",
	["Prince Tenris Mirkblood"] = "Prinz Tenris Mirkblut",
	["Prince Tortheldrin"] = "Prinz Tortheldrin",
	["Prince Valanar"] = "Prinz Valanar",
	["Princess Huhuran"] = "Prinzessin Huhuran",
	["Princess Moira Bronzebeard"] = "Prinzessin Moira Bronzebeard",
	["Princess Tempestria"] = "Prinzessin Tempestria",
	["Princess Theradras"] = "Prinzessin Theradras",
	["Princess Yauj"] = "Prinzessin Yauj",
	["Professor Putricide"] = "Professor Seuchenmord",
	["Pure Spawn of Hydross"] = "Gereinigter Nachkomme Hydross'",
	Pusillin = "Pusillin",
	["Pyroguard Emberseer"] = "Feuerwache Glutseher",
	["Pyromancer Loregrain"] = "Pyromant Weisenkorn",
	Quagmirran = "Quagmirran",
	["Quartermaster Zigris"] = "Rüstmeister Zigris",
	["Rage Winterchill"] = "Furor Winterfrost",
	Ragglesnout = "Struppmähne",
	["Raging Spirit"] = "Tobender Geist",
	Ragnaros = "Ragnaros",
	Rajh = "Rajh",
	["Ramstein the Gorger"] = "Ramstein der Verschlinger",
	["Ras Frostwhisper"] = "Ras Frostraunen",
	Rattlegore = "Blutrippe",
	["Razorclaw the Butcher"] = "Klingenklaue der Metzger",
	["Razorgore the Untamed"] = "Razorgore der Ungezähmte",
	Razorlash = "Schlingwurzler",
	Razorscale = "Klingenschuppe",
	["Reliquary of Souls"] = "Reliquium der Seelen",
	Renataki = "Renataki",
	["Restless Skeleton"] = "Ruheloses Skelett",
	Rethilgore = "Rotkralle",
	Revelosh = "Revelosh",
	["Rhahk'Zor"] = "Rhahk'Zor",
	["Ribbly Screwspigot"] = "Ribbly Schraubstutz",
	["Right Arm"] = "Rechter Arm",
	["Ring of Law"] = "Ring des Gesetzes",
	Roar = "Brüller",
	["Rokad the Ravager"] = "Rokad der Verheerer",
	["Rokdar the Sundered Lord"] = "Rokdar der Zerklüftete",
	["Rokmar the Crackler"] = "Rokmar der Zerquetscher",
	["Rom'ogg Bonecrusher"] = "Rom'ogg Knochenbrecher",
	Romulo = "Romulo",
	["Romulo & Julianne"] = "Romulo & Julianne",
	Roogug = "Roogug",
	Rotface = "Modermiene",
	Rotgrip = "Faulschnapper",
	["Runemaster Molgeim"] = "Runenmeister Molgeim",
	["Runok Wildmane"] = "Runok Wildmähne",
	Ruuzlu = "Ruuzlu",
	["Salramm the Fleshcrafter"] = "Salramm der Fleischformer",
	["Sanctum Sentry"] = "Späher des Sanktums",
	["Sandarr Dunereaver"] = "Sandarr der Wüstenräuber",
	["Sandfury Executioner"] = "Henker der Sandwüter",
	Sapphiron = "Saphiron",
	Sara = "Sara",
	["Saronite Animus"] = "Saronitanimus",
	Sartharion = "Sartharion",
	["Sathrovarr the Corruptor"] = "Sathrovarr der Verderber",
	["Saviana Ragefire"] = "Saviana Flammenschlund",
	["Scarlet Commander Mograine"] = "Scharlachroter Kommandant Mograine",
	["Scourgelord Tyrannus"] = "Geißelfürst Tyrannus",
	["Seeth'rel"] = "Wut'rel",
	["Selin Fireheart"] = "Selin Feuerherz",
	["Sergeant Bly"] = "Unteroffizier Bly",
	Setesh = "Setesh",
	["Shade of Akama"] = "Akamas Schemen",
	["Shade of Aran"] = "Arans Schemen",
	["Shade of Eranikus"] = "Eranikus' Schemen",
	["Shadikith the Glider"] = "Shadikith der Segler",
	["Shadow Hunter Vosh'gajin"] = "Schattenjägerin Vosh'gajin",
	["Shadow of Leotheras"] = "Schatten von Leotheras",
	["Shadowpriest Sezz'ziz"] = "Schattenpriester Sezz'ziz",
	Shadron = "Shadron",
	Shazzrah = "Shazzrah",
	["Shirrak the Dead Watcher"] = "Shirrak der Totenwächter",
	["Siamat, Lord of South Wind"] = "Siamat, Gebieter des Südwinds",
	Sindragosa = "Sindragosa",
	["Sir Zeliek"] = "Sir Zeliek",
	["Sjonnir The Ironshaper"] = "Sjonnir der Eisenformer",
	["Skadi the Ruthless"] = "Skadi der Skrupellose",
	["Skarr the Unbreakable"] = "Skarr der Unbezwingbare",
	["Skarvald the Constructor"] = "Skarvald der Konstrukteur",
	["Skra'gath"] = "Skra'gath",
	Skul = "Skul",
	Skum = "Skum",
	Slabhide = "Plattenhaut",
	["Slad'ran"] = "Slad'ran",
	Sneed = "Sneed",
	["Sneed's Shredder"] = "Sneeds Schredder",
	["Solakar Flamewreath"] = "Solakar Feuerkrone",
	["Solarium Agent"] = "Solarian Agent",
	["Solarium Priest"] = "Solarian Priester",
	["Spirestone Battle Lord"] = "Kampflord der Felsspitzoger",
	["Spirestone Butcher"] = "Metzger der Felsspitzoger",
	["Spirestone Lord Magus"] = "Maguslord der Felsspitzoger",
	["Staff of Disintegration"] = "Stab der Auflösung",
	Stalagg = "Stalagg",
	Steelbreaker = "Stahlbrecher",
	["Stomper Kreeg"] = "Stampfer Kreeg",
	Stonespine = "Steinbuckel",
	["Stormcaller Brundir"] = "Sturmrufer Brundir",
	Strawman = "Strohmann",
	["Sulfuron Harbinger"] = "Sulfuronherold",
	Supremus = "Supremus",
	["Svala Sorrowgrave"] = "Svala Grabesleid",
	["Swamplord Musel'ek"] = "Sumpffürst Musel'ek",
	Taerar = "Taerar",
	["Tainted Spawn of Hydross"] = "Besudelter Nachkomme Hydross'",
	["Talon King Ikiss"] = "Klauenkönig Ikiss",
	["Taragaman the Hungerer"] = "Taragaman der Hungerleider",
	["Targorr the Dread"] = "Targorr der Schreckliche",
	Tavarok = "Tavarok",
	Techbot = "Techbot",
	["Temple Guardian Anhuur"] = "Tempelwächter Anhuur",
	Temporus = "Temporus",
	["Tendris Warpwood"] = "Tendris Wucherborke",
	Tenebron = "Tenebron",
	["Terestian Illhoof"] = "Terestian Siechhuf",
	["Teron Gorefiend"] = "Teron Blutschatten",
	Thaddius = "Thaddius",
	["Thaladred the Darkener"] = "Thaladred der Verfinsterer",
	["Thane Korth'azz"] = "Than Korth'azz",
	["The Beast"] = "Die Bestie",
	["The Beasts of Northrend"] = "Bestien von Nordend",
	["The Big Bad Wolf"] = "Der große böse Wolf",
	["The Black Knight"] = "Der Schwarze Ritter",
	["The Black Stalker"] = "Die Schattenmutter",
	["The Blue Brothers"] = "Die Blaumänner",
	["The Bug Family"] = "Die Käferfamilie",
	["The Crone"] = "Die böse Hexe",
	["The Curator"] = "Der Kurator",
	["The Eredar Twins"] = "Die Eredar Zwillinge",
	["The Four Horsemen"] = "Die Vier Reiter",
	["The Illidari Council"] = "Rat der Illidari",
	["The Iron Council"] = "Versammlung des Eisens",
	["The Lich King"] = "Der Lichkönig",
	["The Lost Dwarves"] = "Die verschollenen Zwerge",
	["The Lurker Below"] = "Das Grauen aus der Tiefe",
	["The Maker"] = "Der Schöpfer",
	["The Prophet Skeram"] = "Der Prophet Skeram",
	["The Prophet Tharon'ja"] = "Der Prophet Tharon'ja",
	["The Ravenian"] = "Der Ravenier",
	["The Razza"] = "Der Razza",
	["The Seven Dwarves"] = "Die Sieben Zwerge",
	["The Skybreaker"] = "Die Himmelsbrecher",
	["The Tribunal of Ages"] = "Das Tribunal der Zeitalter",
	["The Twin Emperors"] = "Die Zwillings-Imperatoren",
	["The Twin Val'kyr"] = "Zwillingsval'kyr",
	["The Unforgiven"] = "Der Unverziehene",
	["The Windreaver"] = "Der Windhäscher",
	["Theka the Martyr"] = "Theka der Märtyrer",
	Thorim = "Thorim",
	["Thorngrin the Tender"] = "Dorngrin der Hüter",
	["Tidewalker Lurker"] = "Lauerer der Gezeitenwandler",
	["Timmy the Cruel"] = "Timmy der Grausame",
	Tinhead = "Blechkopf",
	["Tinkerer Gizlock"] = "Tüftler Gizlock",
	["Tirion Fordring"] = "Tirion Fordring",
	Tito = "Tito",
	["Toravon the Ice Watcher"] = "Toravon der Eiswächter",
	["Trigore the Lasher"] = "Trigore der Peitscher",
	Trollgore = "Trollgrind",
	["Tsu'zee"] = "Tsu'zee",
	["Tuten'kash"] = "Tuten'kash",
	["Twilight Lord Kelris"] = "Lord des Schattenhammers Kelris",
	["Urok Doomhowl"] = "Urok Schreckensbote",
	["VX-001"] = "VX-001",
	["Vaelastrasz the Corrupt"] = "Vaelastrasz der Verdorbene",
	["Val'kyr Shadowguard"] = "Schattenwächterin der Val'kyr",
	["Valithria Dreamwalker"] = "Valithria Traumwandler",
	["Varian Wrynn"] = "Varian Wrynn",
	["Varos Cloudstrider"] = "Varos Wolkenschreiter",
	Vazruden = "Vazruden",
	["Vazruden the Herald"] = "Vazruden der Herold",
	Vectus = "Vectus",
	Vem = "Vem",
	Veng = "Veng",
	["Veras Darkshadow"] = "Veras Schwarzschatten",
	["Verdan the Everliving"] = "Verdan der Ewiglebende",
	Verek = "Verek",
	Vesperon = "Vesperon",
	Vexallus = "Vexallus",
	["Veyzhak the Cannibal"] = "Veyzhack der Kannibale",
	["Vile'rel"] = "Bös'rel",
	Viscidus = "Viscidus",
	["Viscous Fallout"] = "Verflüssigte Ablagerung",
	["Void Reaver"] = "Leerhäscher",
	Volkhan = "Volkhan",
	["War Master Voone"] = "Kriegsmeister Voone",
	["Warbringer O'mrogg"] = "Kriegshetzer O'mrogg",
	["Warchief Blackhand Piece"] = "Kriegshäuptling Schwarzfaust",
	["Warchief Kargath Bladefist"] = "Kriegshäuptling Kargath Messerfaust",
	["Warchief Rend Blackhand"] = "Kriegshäuptling Rend Schwarzfaust",
	["Warden Mellichar"] = "Aufseher Mellichar",
	["Warder Stilgiss"] = "Wärter Stilgiss",
	["Warlord Kalithresh"] = "Kriegsherr Kalithresh",
	["Warmaul Champion"] = "Champion der Totschläger",
	["Warp Slicer"] = "Warpschnitter",
	["Warp Splinter"] = "Warpzweig",
	["Watchkeeper Gargolmar"] = "Wachhabender Gargolmar",
	Weaver = "Wirker",
	["Witch Doctor Zum'rah"] = "Hexendoktor Zum'rah",
	["Wolf Master Nandos"] = "Wolfmeister Nados",
	["Wrath-Scryer Soccothrates"] = "Zornseher Soccothrates",
	Wushoolay = "Wushoolay",
	["XT-002 Deconstructor"] = "XT-002 Dekonstruktor",
	Xevozz = "Xevozz",
	["Yogg-Saron"] = "Yogg-Saron",
	Yor = "Yor",
	Ysondre = "Ysondre",
	Zekkis = "Zekkis",
	["Zelemar the Wrathful"] = "Zelemar der Hasserfüllte",
	["Zereketh the Unbound"] = "Zereketh der Unabhängige",
	Zerillis = "Zerillis",
	["Zevrim Thornhoof"] = "Zevrim Dornhuf",
	Zolo = "Zolo",
	["Zul'Farrak Dead Hero"] = "Untoter Held aus Zul'Farrak",
	["Zul'Lor"] = "Zul'Lor",
	["Zul'jin"] = "Zul'jin",
	["Zul'tore"] = "Zul'tore",
	["Zuramat the Obliterator"] = "Zuramat der Vernichter",
}
elseif GAME_LOCALE == "frFR" then
	lib:SetCurrentTranslations {
	Acidmaw = "Gueule-d'acide",
	Aeonus = "Aeonus",
	["Aerial Command Unit"] = "Unité de commandement aérien",
	["Agathelos the Raging"] = "Agathelos le Déchaîné",
	["Aggem Thorncurse"] = "Aggem Malépine",
	Ahune = "Ahune",
	["Akil'zon"] = "Akil'zon",
	["Aku'mai"] = "Aku'mai",
	["Al'ar"] = "Al'ar",
	["Algalon the Observer"] = "Algalon l'Observateur",
	-- Altarius = "",
	["Alzzin the Wildshaper"] = "Alzzin le Modeleur",
	Amanitar = "Amanitar",
	["Ambassador Flamelash"] = "Ambassadeur Cinglefouet",
	["Ambassador Hellmaw"] = "Ambassadeur Gueule-d'enfer",
	-- Ammunae = "",
	["Amnennar the Coldbringer"] = "Amnennar le Porte-froid",
	["Ancient Stone Keeper"] = "Ancien Gardien des pierres",
	Anetheron = "Anetheron",
	["Anger'rel"] = "Colé'rel",
	Anomalus = "Anomalus",
	-- Anraphet = "",
	["Antu'sul"] = "Antu'sul",
	["Anub'Rekhan"] = "Anub'Rekhan",
	["Anub'arak"] = "Anub'arak",
	["Anub'shiah"] = "Anub'shiah",
	["Anubisath Defender"] = "Défenseur Anubisath",
	["Anubisath Guardian"] = "Gardien Anubisath",
	Anzu = "Anzu",
	["Arcane Watchman"] = "Veilleur arcanique",
	["Arcanist Doan"] = "Arcaniste Doan",
	Archaedas = "Archaedas",
	["Archavon the Stone Watcher"] = "Archavon le Gardien des pierres",
	Archimonde = "Archimonde",
	["Archivist Galford"] = "Archiviste Galford",
	["Archmage Arugal"] = "Archimage Arugal",
	["Argent Confessor Paletress"] = "Confesseur d'argent Paletress",
	["Arugal's Voidwalker"] = "Marcheur du Vide d'Arugal",
	["Ascendant Lord Obsidius"] = "Seigneur ascendant Obsidius",
	-- ["Asim al Akir"] = "",
	["Assault Bot"] = "Robot d'assaut",
	["Assembly of Iron"] = "Assemblée du Fer",
	["Atal'alarion"] = "Atal'alarion",
	-- Atramedes = "",
	["Attumen the Huntsman"] = "Attumen le Veneur",
	Auriaya = "Auriaya",
	Avalanchion = "Avalanchion",
	["Avatar of Hakkar"] = "Avatar d'Hakkar",
	["Ayamiss the Hunter"] = "Ayamiss le Chasseur",
	Azgalor = "Azgalor",
	["Azshir the Sleepless"] = "Azshir le Sans-sommeil",
	Azuregos = "Azuregos",
	["Bael'Gar"] = "Bael'Gar",
	Baelog = "Baelog",
	Balnazzar = "Balnazzar",
	["Baltharus the Warborn"] = "Baltharus l'Enfant de la guerre",
	["Bannok Grimaxe"] = "Bannok Hache-sinistre",
	["Baron Aquanis"] = "Baron Aquanis",
	["Baron Charr"] = "Baron Charr",
	["Baron Geddon"] = "Baron Geddon",
	["Baron Kazum"] = "Baron Kazum",
	["Baron Rivendare"] = "Baron Vaillefendre",
	["Baron Silverlaine"] = "Baron d'Argelaine",
	["Baroness Anastari"] = "Baronne Anastari",
	["Battleguard Sartura"] = "Garde de guerre Sartura",
	["Bazil Thredd"] = "Bazil Thredd",
	Bazzalan = "Bazzalan",
	Beauty = "La Belle",
	["Black Guard Swordsmith"] = "Fabricant d'épées de la Garde noire",
	["Blackheart the Inciter"] = "Coeur-noir le Séditieux",
	["Blind Hunter"] = "Chasseur aveugle",
	["Blindeye the Seer"] = "Oeillaveugle le Voyant",
	["Blood Guard Porung"] = "Garde de sang Porung",
	["Blood Prince Council"] = "Conseil des princes de sang",
	["Blood Princes"] = "Princes de sang",
	["Blood Steward of Kirtonos"] = "Régisseuse sanglante de Kirtonos",
	["Blood-Queen Lana'thel"] = "Reine de sang Lana'thel",
	["Bloodlord Mandokir"] = "Seigneur sanglant Mandokir",
	["Bloodmage Thalnos"] = "Mage de sang Thalnos",
	Boahn = "Boahn",
	["Bomb Bot"] = "Robo-bombe",
	["Brain of Yogg-Saron"] = "Cerveau de Yogg-Saron",
	["Brainwashed Noble"] = "Noble manipulé",
	Broggok = "Broggok",
	Brokentoe = "Brisorteil",
	Bronjahm = "Bronjahm",
	["Broodlord Lashlayer"] = "Seigneur des couvées Lanistaire",
	["Bruegal Ironknuckle"] = "Bruegal Poing-de-fer",
	Brutallus = "Brutallus",
	["Burning Felguard"] = "Gangregarde ardent",
	["Buru the Gorger"] = "Buru Grandgosier",
	["C'Thun"] = "C'Thun",
	["Cache of the Firelord"] = "Cachette du Seigneur du feu",
	["Cache of the Legion"] = "Cache de la Légion",
	["Cannon Master Willey"] = "Maître canonnier Willey",
	["Captain Greenskin"] = "Capitaine Vertepeau",
	["Captain Kromcrush"] = "Capitaine Kromcrush",
	["Captain Skarloc"] = "Capitaine Skarloc",
	["Celebras the Cursed"] = "Celebras le Maudit",
	["Charlga Razorflank"] = "Charlga Trancheflanc",
	["Chess Event"] = "Partie d'échec",
	["Chest of The Seven"] = "Coffre des sept",
	["Chief Ukorz Sandscalp"] = "Chef Ukorz Scalpessable",
	-- Chimaeron = "",
	["Cho'Rush the Observer"] = "Cho'Rush l'Observateur",
	Chromaggus = "Chromaggus",
	["Chrono Lord Deja"] = "Chronoseigneur Déjà",
	["Chrono-Lord Epoch"] = "Chronoseigneur Epoch",
	Claw = "Griffe",
	["Coilfang Elite"] = "Elite de Glissecroc",
	["Coilfang Strider"] = "Trotteur de Glissecroc",
	["Commander Kolurg"] = "Commandant Kolurg",
	["Commander Sarannis"] = "Commandant Sarannis",
	["Commander Springvale"] = "Commandant Springvale",
	["Commander Stoutbeard"] = "Commandant Rudebarbe",
	["Commander Ulthok"] = "Commandant Ulthok",
	["Constructor & Controller"] = "Constructeur & Contrôleur",
	Cookie = "Macaron",
	-- Corborus = "",
	["Coren Direbrew"] = "Coren Navrebière",
	["Corla, Herald of Twilight"] = "Corla, héraut du Crépuscule",
	["Cosmic Infuser"] = "Masse d'infusion cosmique",
	["Crimson Hammersmith"] = "Forgeur de marteaux cramoisi",
	["Crowd Pummeler 9-60"] = "Faucheur de foule 9-60",
	["Crystal Fang"] = "Croc cristallin",
	Cyanigosa = "Cyanigosa",
	["Dalliah the Doomsayer"] = "Dalliah l'Auspice-funeste",
	["Dalronn the Controller"] = "Dalronn le Contrôleur",
	["Dark Coffer"] = "Coffre sombre",
	["Dark Iron Ambassador"] = "Ambassadeur Sombrefer",
	["Darkmaster Gandling"] = "Sombre Maître Gandling",
	["Darkweaver Syth"] = "Tisseur d'ombre Syth",
	["Death Knight Darkreaver"] = "Chevalier de la mort Ravassombre",
	["Death Knight Understudy"] = "Doublure de chevalier de la mort",
	["Death Speaker Jargba"] = "Nécrorateur Jargba",
	["Deathbound Ward"] = "Gardien lié par la mort",
	["Deathbringer Saurfang"] = "Porte-mort Saurcroc",
	["Deathspeaker High Priest"] = "Grand prêtre nécrorateur",
	["Deathstalker Visceri"] = "Nécrotraqueur Viscéri",
	["Deathsworn Captain"] = "Capitaine Ligemort",
	Devastation = "Dévastation",
	["Deviate Faerie Dragon"] = "Dragon féérique déviant",
	["Devourer of Souls"] = "Dévoreur d’âmes",
	["Dextren Ward"] = "Dextren Ward",
	["Digmaster Shovelphlange"] = "Maître des fouilles Pellaphlange",
	["Doctor Theolen Krastinov"] = "Docteur Theolen Krastinov",
	["Doom Lord Kazzak"] = "Seigneur funeste Kazzak",
	["Doom'rel"] = "Tragi'rel",
	Doomwalker = "Marche-funeste",
	["Dope'rel"] = "Demeu'rel",
	Dorothee = "Dorothée",
	["Drahga Shodowburner"] = "Drahga Brûle-Ombre",
	["Drakkari Colossus"] = "Colosse drakkari",
	["Drakos the Interrogator"] = "Drakos l'Interrogateur",
	Dreadscale = "Écaille-d'effroi",
	Dreamscythe = "Fauche-rêve",
	["Dust Covered Chest"] = "Coffre couvert de poussière",
	Dustwraith = "Ame en peine poudreuse",
	["Eadric the Pure"] = "Eadric le Pur",
	["Earthcaller Halmgar"] = "Implorateur de la terre Halmgar",
	-- ["Earthrager Ptah"] = "",
	Ebonroc = "Rochébène",
	["Eck the Ferocious"] = "Eck le Féroce",
	["Edwin VanCleef"] = "Edwin VanCleef",
	["Elder Brightleaf"] = "Ancien Brillefeuille",
	["Elder Ironbranch"] = "Ancien Branchefer",
	["Elder Nadox"] = "Ancien Nadox",
	["Elder Stonebark"] = "Ancien Ecorcepierre",
	["Electrocutioner 6000"] = "Electrocuteur 6000",
	["Emalon the Storm Watcher"] = "Emalon le Guetteur d'orage",
	Emeriss = "Emeriss",
	["Emperor Dagran Thaurissan"] = "Empereur Dagran Thaurissan",
	["Emperor Vek'lor"] = "Empereur Vek'lor",
	["Emperor Vek'nilash"] = "Empereur Vek'nilash",
	Entropius = "Entropius",
	["Eonar's Gift"] = "Cadeau d'Eonar",
	["Epoch Hunter"] = "Chasseur d'époques",
	Erekem = "Erekem",
	["Eressea Dawnsinger"] = "Eressa Chantelaube",
	["Eric \"The Swift\""] = "Eric « l'Agile »",
	Erudax = " Erudax",
	["Erunak Stonespeaker"] = "Erunak Parlepierre",
	["Essence of Anger"] = "Essence de la colère",
	["Essence of Desire"] = "Essence du désir",
	["Essence of Suffering"] = "Essence de la souffrance",
	Eviscerator = "Eviscérateur",
	["Exarch Maladaar"] = "Exarque Maladaar",
	["Expedition Commander"] = "Commandant de l'expédition",
	["Eydis Darkbane"] = "Eydis Plaie-sombre",
	["Eye of C'Thun"] = "Œil de C'Thun",
	["Faction Champions"] = "Champions de faction",
	["Fallen Champion"] = "Champion mort",
	Falric = "Falric",
	["Falric and Marwyn"] = "Falric et Marwyn",
	["Fankriss the Unyielding"] = "Fankriss l'Inflexible",
	["Fathom-Lord Karathress"] = "Seigneur des fonds Karathress",
	Felmyst = "Gangrebrume",
	["Fenrus the Devourer"] = "Fenrus le Dévoreur",
	["Feral Defender"] = "Défenseur farouche",
	Festergut = "Pulentraille",
	Feugen = "Feugen",
	["Fineous Darkvire"] = "Fineous Sombrevire",
	Firemaw = "Gueule-de-feu",
	["Fjola Lightbane"] = "Fjola Plaie-lumineuse",
	["Flame Leviathan"] = "Léviathan des flammes",
	Flamegor = "Flamegor",
	["Foreman Thistlenettle"] = "Contremaître Crispechardon",
	["Forgemaster Garfrost"] = "Maître-forge Gargivre",
	-- ["Forgemaster Throngus"] = "",
	["Four Horsemen Chest"] = "Coffre des quatre cavaliers",
	["Fras Siabi"] = "Fras Siabi",
	Freya = "Freya",
	["Gahz'ranka"] = "Gahz'ranka",
	["Gahz'rilla"] = "Gahz'rilla",
	["Gal'darah"] = "Gal'darah",
	["Galgann Firehammer"] = "Galgann Martel-de-feu",
	Garr = "Garr",
	["Garrosh Hellscream"] = "Garrosh Hurlenfer",
	Gasher = "Gasher",
	["Gatewatcher Gyro-Kill"] = "Gardien de porte Gyro-Meurtre",
	["Gatewatcher Iron-Hand"] = "Gardien de porte Main-en-fer",
	["Gathios the Shatterer"] = "Gathios le Briseur",
	Gehennas = "Gehennas",
	Gelihast = "Gelihast",
	Gelk = "Gelk",
	["General Angerforge"] = "Général Forgehargne",
	["General Bjarngrim"] = "Général Bjarngrim",
	["General Drakkisath"] = "Général Drakkisath",
	-- ["General Husam"] = "",
	["General Rajaxx"] = "Général Rajaxx",
	["General Umbriss"] = "Général Umbriss",
	["General Vezax"] = "Général Vezax",
	["General Zarithrian"] = "Général Zarithrian",
	["Ghamoo-ra"] = "Ghamoo-ra",
	["Ghaz'an"] = "Ghaz'an",
	["Ghok Bashguud"] = "Ghok Bounnebaffe",
	Gilnid = "Gilnid",
	["Gizrul the Slavener"] = "Gizrul l'esclavagiste",
	["Gloom'rel"] = "Funéb'rel",
	Gluth = "Gluth",
	Glutton = "Glouton",
	["Golem Lord Argelmach"] = "Seigneur golem Argelmach",
	["Golemagg the Incinerator"] = "Golemagg l'Incinérateur",
	["Goraluk Anvilcrack"] = "Goraluk Brisenclume",
	["Gormok the Impaler"] = "Gormok l'Empaleur",
	["Gorosh the Dervish"] = "Gorosh le Derviche",
	["Gortok Palehoof"] = "Gortok Pâle-sabot",
	["Gothik the Harvester"] = "Gothik le Moissonneur",
	["Grand Astromancer Capernian"] = "Grande astromancienne Capernian",
	["Grand Champions"] = "Grand champions",
	["Grand Crusader Dathrohan"] = "Grand croisé Dathrohan",
	["Grand Magus Telestra"] = "Grand magus Telestra",
	["Grand Vizier Ertan"] = "Grand vizir Ertan",
	["Grand Warlock Alythess"] = "Grande démoniste Alythess",
	["Grand Warlock Nethekurse"] = "Grand démoniste Néanathème",
	["Grand Widow Faerlina"] = "Grande veuve Faerlina",
	["Grandmaster Vorpil"] = "Grand Maître Vorpil",
	Grandmother = "Mère-grand",
	["Grethok the Controller"] = "Grethok le Contrôleur",
	["Gri'lek"] = "Gri'lek",
	Grimlok = "Grimlok",
	Grizzle = "Grison",
	Grobbulus = "Grobbulus",
	Grubbis = "Grubbis",
	["Gruul the Dragonkiller"] = "Gruul le Tue-dragon",
	["Guard Fengus"] = "Garde Fengus",
	["Guard Mol'dar"] = "Garde Mol'dar",
	["Guard Slip'kik"] = "Garde Slip'kik",
	["Guardian of Yogg-Saron"] = "Gardien de Yogg-Saron",
	["Gurtogg Bloodboil"] = "Gurtogg Fièvresang",
	Gyth = "Gyth",
	Hadronox = "Hadronox",
	Hakkar = "Hakkar",
	Halazzi = "Halazzi",
	Halion = "Halion",
	Halycon = "Halycon",
	Hamhock = "Hamhock",
	["Harbinger Skyriss"] = "Messager Cieuriss",
	["Hate'rel"] = "Haine'rel",
	["Hazza'rah"] = "Hazza'rah",
	Hazzas = "Hazzas",
	["Headless Horseman"] = "Cavalier sans tête",
	["Hearthsinger Forresten"] = "Chanteloge Forrestin",
	["Hedrum the Creeper"] = "Hedrum le Rampant",
	["Heigan the Unclean"] = "Heigan l'Impur",
	["Hellfire Channeler"] = "Canaliste des Flammes infernales",
	["Henry Stern"] = "Henry Stern",
	["Herald Volazj"] = "Héraut Volazj",
	Herod = "Hérode",
	["Hex Lord Malacrass"] = "Seigneur des maléfices Malacrass",
	["High Astromancer Solarian"] = "Grande astromancienne Solarian",
	["High Botanist Freywinn"] = "Grand botaniste Freywinn",
	["High Inquisitor Fairbanks"] = "Grand Inquisiteur Fairbanks",
	["High Inquisitor Whitemane"] = "Grand Inquisiteur Blanchetête",
	["High Interrogator Gerstahn"] = "Grand Interrogateur Gerstahn",
	["High King Maulgar"] = "Haut Roi Maulgar",
	["High Marshal Whirlaxis"] = "Haut maréchal Trombe",
	["High Nethermancer Zerevor"] = "Grand néantomancien Zerevor",
	["High Overlord Saurfang"] = "Haut seigneur Saurcroc",
	["High Priest Thekal"] = "Grand prêtre Thekal",
	["High Priest Venoxis"] = "Grand prêtre Venoxis",
	["High Priestess Arlokk"] = "Grande prêtresse Arlokk",
	["High Priestess Azil"] = "Grande prêtresse Azil",
	["High Priestess Jeklik"] = "Grande prêtresse Jeklik",
	["High Priestess Mar'li"] = "Grande prêtresse Mar'li",
	["High Priestess of Thaurissan"] = "Grande prêtresse de Thaurissan",
	-- ["High Prophet Barim"] = "",
	["High Warlord Naj'entus"] = "Grand seigneur de guerre Naj'entus",
	["Highlord Mograine"] = "Généralissime Mograine",
	["Highlord Omokk"] = "Généralissime Omokk",
	Hodir = "Hodir",
	["Houndmaster Grebmar"] = "Maître-chien Grebmar",
	["Houndmaster Loksey"] = "Maître-chien Loksey",
	Hukku = "Hukku",
	Hungarfen = "Hungarfen",
	["Hurley Blackbreath"] = "Hurley Soufflenoir",
	["Hyakiss the Lurker"] = "Hyakiss le rôdeur",
	["Hydromancer Thespia"] = "Hydromancienne Thespia",
	["Hydromancer Velratha"] = "Hydromancienne Velratha",
	Hydrospawn = "Hydrogénos",
	["Hydross the Unstable"] = "Hydross l'Instable",
	["Ice Sphere"] = "Sphère de glace",
	["Icecrown Gunship Battle"] = "Bataille des canonnières",
	Icehowl = "Glace-hurlante",
	Ichoron = "Ichoron",
	Ick = "Ick",
	["Ignis the Furnace Master"] = "Ignis le maître de la Fournaise",
	["Illidan Stormrage"] = "Illidan Hurlorage",
	["Illidari Council"] = "Conseil illidari",
	["Illyanna Ravenoak"] = "Illyanna Corvichêne",
	["Immol'thar"] = "Immol'thar",
	["Infinite Corruptor"] = "Corrupteur infini",
	["Infinity Blades"] = "Lames d'infinité",
	["Ingvar the Plunderer"] = "Ingvar le Pilleur",
	["Instructor Malicia"] = "Instructeur Malicia",
	["Instructor Razuvious"] = "Instructeur Razuvious",
	["Interrogator Vishas"] = "Interrogateur Vishas",
	Ionar = "Ionar",
	Ironaya = "Ironaya",
	Ironspine = "Echine-de-fer",
	Isalien = "Isalien",
	-- Isiset = "",
	Jade = "Jade",
	["Jammal'an the Prophet"] = "Jammal'an le prophète",
	["Jan'alai"] = "Jan'alai",
	["Jandice Barov"] = "Jandice Barov",
	["Jed Runewatcher"] = "Jed Guette-runes",
	["Jedoga Shadowseeker"] = "Jedoga Cherchelombre",
	["Jergosh the Invoker"] = "Jergosh l'Invocateur",
	["Jin'do the Hexxer"] = "Jin'do le Maléficieur",
	["Jormungar Behemoth"] = "Béhémoth jormungar",
	Jormungars = "Jormungars",
	Julianne = "Julianne",
	["Junk Bot"] = "Robot camelote",
	["Kael'thas Sunstrider"] = "Kael'thas Haut-soleil",
	Kalecgos = "Kalecgos",
	["Kam Deepfury"] = "Kam Furie-du-fond",
	["Karsh Steelbender"] = "Karsh Plielacier",
	["Kaz'rogal"] = "Kaz'rogal",
	["Kazkaz the Unholy"] = "Kazkaz l'Impie",
	["Kel'Thuzad"] = "Kel'Thuzad",
	["Keli'dan the Breaker"] = "Keli'dan le Briseur",
	Keristrasza = "Keristrasza",
	["Kiggler the Crazed"] = "Kiggler le Cinglé",
	["Kil'jaeden"] = "Kil'jaeden",
	["Kil'rek"] = "Kil'rek",
	["King Dred"] = "Roi Dred",
	["King Gordok"] = "Roi Gordok",
	["King Llane Piece"] = "Pion du Roi Llane",
	["King Ymiron"] = "Roi Ymiron",
	["Kirtonos the Herald"] = "Kirtonos le Héraut",
	["Knot Thimblejack's Cache"] = "Réserve de Noué Dédodevie",
	Kolk = "Kolk",
	Kologarn = "Kologarn",
	["Koralon the Flame Watcher"] = "Koralon le Veilleur des flammes",
	Kormok = "Kormok",
	Kresh = "Kresh",
	Krick = "Krick",
	["Krick and Ick"] = "Krick et Ick",
	["Krik'thir the Gatewatcher"] = "Krik'thir le Gardien de porte",
	["Krosh Firehand"] = "Krosh Brasemain",
	Krystallus = "Krystallus",
	Kurinnaxx = "Kurinnaxx",
	["Lady Anacondra"] = "Dame Anacondra",
	["Lady Blaumeux"] = "Dame Blaumeux",
	["Lady Deathwhisper"] = "Dame Murmemort",
	["Lady Illucia Barov"] = "Dame Illucia Barov",
	["Lady Malande"] = "Dame Malande",
	["Lady Naz'jar"] = "Dame Naz'jar",
	["Lady Sacrolash"] = "Dame Sacrocingle",
	["Lady Sarevess"] = "Dame Sarevess",
	["Lady Vashj"] = "Dame Vashj",
	Laj = "Laj",
	Landslide = "Glissement de terrain",
	Lavanthor = "Lavanthor",
	["Left Arm"] = "Bras gauche",
	["Leotheras the Blind"] = "Leotheras l'Aveugle",
	Lethon = "Léthon",
	Lethtendris = "Lethtendris",
	["Leviathan Mk II"] = "Léviathan Mod. II",
	["Ley-Guardian Eregos"] = "Gardien-tellurique Eregos",
	["Lieutenant Drake"] = "Lieutenant Drake",
	["Lieutenant General Andorov"] = "Général de division Andorov",
	Loatheb = "Horreb",
	-- Lockmaw = "",
	Loken = "Loken",
	["Lord Alexei Barov"] = "Seigneur Alexei Barov",
	["Lord Cobrahn"] = "Seigneur Cobrahn",
	["Lord Hel'nurath"] = "Seigneur Hel'nurath",
	["Lord Incendius"] = "Seigneur Incendius",
	["Lord Jaraxxus"] = "Seigneur Jaraxxus",
	["Lord Kazzak"] = "Seigneur Kazzak",
	["Lord Kri"] = "Seigneur Kri",
	["Lord Marrowgar"] = "Seigneur Gargamoelle",
	["Lord Pythas"] = "Seigneur Pythas",
	["Lord Roccor"] = "Seigneur Roccor",
	["Lord Sanguinar"] = "Seigneur Sanguinar",
	["Lord Serpentis"] = "Seigneur Serpentis",
	["Lord Skwol"] = "Seigneur Skwol",
	["Lord Valthalak"] = "Seigneur Valthalak",
	["Lord Victor Nefarius"] = "Seigneur Victor Nefarius",
	["Lord Vyletongue"] = "Seigneur Vylelangue",
	["Lorekeeper Polkelt"] = "Gardien du savoir Polkelt",
	["Lorgus Jett"] = "Lorgus Jett",
	Loro = "Loro",
	Lucifron = "Lucifron",
	["M'uru"] = "M'uru",
	["Mad Magglish"] = "Magglish le Dingue",
	Maexxna = "Maexxna",
	["Mage-Lord Urom"] = "Seigneur-mage Urom",
	["Magister Kalendris"] = "Magistère Kalendris",
	["Magistrate Barthilas"] = "Magistrat Barthilas",
	Magmadar = "Magmadar",
	-- Magmaw = "",
	Magmus = "Magmus",
	Magra = "Magra",
	Magtheridon = "Magtheridon",
	["Maiden of Grief"] = "Damoiselle de peine",
	["Maiden of Virtue"] = "Damoiselle de vertu",
	["Majordomo Executus"] = "Chambellan Executus",
	["Mal'Ganis"] = "Mal'Ganis",
	Malacrass = "Malacrass",
	["Maleki the Pallid"] = "Maleki le Blafard",
	-- Maloriak = "",
	Malygos = "Malygos",
	Maraudos = "Maraudos",
	["Marduk Blackpool"] = "Marduk Noirétang",
	["Marisa du'Paige"] = "Marisa du'Paige",
	Marwyn = "Marwyn",
	["Master Engineer Telonicus"] = "Maître ingénieur Telonicus",
	["Maur Grimtotem"] = "Maur Totem-sinistre",
	Meathook = "Grancrochet",
	["Mechano-Lord Capacitus"] = "Mécano-seigneur Capacitus",
	Medivh = "Medivh",
	["Mekgineer Steamrigger"] = "Mékgénieur Montevapeur",
	["Mekgineer Thermaplugg"] = "Mekgénieur Thermojoncteur",
	["Mennu the Betrayer"] = "Mennu le Traître",
	["Meshlok the Harvester"] = "Meshlok le Moissonneur",
	Midnight = "Minuit",
	Mijan = "Mijan",
	Mimiron = "Mimiron",
	-- ["Mindbender Ghur'sha"] = "",
	["Miner Johnson"] = "Mineur Johnson",
	["Mistress of Pain"] = "Maîtresse de Douleur",
	Moam = "Moam",
	Mogor = "Mogor",
	["Mokra the Skullcrusher"] = "Mokra le Brise-tête",
	Moorabi = "Moorabi",
	["Mor Grayhoof"] = "Mor Sabot-gris",
	Moragg = "Moragg",
	["Mordresh Fire Eye"] = "Mordresh Oeil-de-feu",
	Moroes = "Moroes",
	["Morogrim Tidewalker"] = "Morogrim Marcheur-des-flots",
	Morphaz = "Morphaz",
	["Mother Shahraz"] = "Mère Shahraz",
	["Mother Smolderweb"] = "Matriarche Couveuse",
	["Mr. Smite"] = "M. Châtiment",
	["Muradin Bronzebeard"] = "Muradin Barbe-de-bronze",
	["Murkblood Twin"] = "Jumeau bourbesang",
	["Murkblood Twins"] = "Jumeaux bourbesang",
	Murmur = "Marmon",
	["Murta Grimgut"] = "Murta Mornentraille",
	Mushgog = "Mushgog",
	["Mutanus the Devourer"] = "Mutanus le Dévoreur",
	Nalorakk = "Nalorakk",
	Nazan = "Nazan",
	Nefarian = "Nefarian",
	["Nekrum Gutchewer"] = "Nekrum Mâchetripes",
	["Nerub'enkan"] = "Nerub'enkan",
	["Nethermancer Sepethrea"] = "Néantomancien Sepethrea",
	Netherspite = "Dédain-du-Néant",
	["Netherstrand Longbow"] = "Arc long brins-de-Néant",
	["Nexus-Prince Shaffar"] = "Prince-nexus Shaffar",
	Nightbane = "Plaie-de-nuit",
	["Noth the Plaguebringer"] = "Noth le Porte-peste",
	["Novos the Summoner"] = "Novos l'Invocateur",
	Noxxion = "Noxcion",
	["Obsidian Sentinel"] = "Sentinelle d'obsidienne",
	["Odo the Blindwatcher"] = "Odo l'Aveugle",
	Oggleflint = "Lorgnesilex",
	["Ogom the Wretched"] = "Ogom le Misérable",
	Ohgan = "Ohgan",
	["Ok'thor the Breaker"] = "Ok'thor le Briseur",
	Olaf = "Olaf",
	["Old Serra'kis"] = "Vieux Serra'kis",
	["Olm the Summoner"] = "Olm l'Invocateur",
	["Omor the Unscarred"] = "Omor l'Intouché",
	Onyxia = "Onyxia",
	["Opera Event"] = "Evénement de l'opéra",
	["Orgrim's Hammer"] = "Marteau d'Orgrim",
	["Ormorok the Tree-Shaper"] = "Ormorok le Sculpte-arbre",
	["Oro Eyegouge"] = "Oro Crève-oeil ",
	["Ossirian the Unscarred"] = "Ossirian l'Intouché",
	Ouro = "Ouro",
	["Overlord Ramtusk"] = "Seigneur Brusquebroche",
	["Overlord Wyrmthalak"] = "Seigneur Wyrmthalak",
	["Overmaster Pyron"] = "Grand seigneur Pyron",
	["Overseer Tidewrath"] = "Surveillant Tidewrath",
	-- Ozruk = "",
	Ozumat = "Ozumat",
	Pandemonius = "Pandemonius",
	["Panzor the Invincible"] = "Panzor l'Invincible",
	Patchwerk = "Le Recousu",
	["Pathaleon the Calculator"] = "Pathaleon le Calculateur",
	Phalanx = "Phalange",
	["Phaseshift Bulwark"] = "Rempart de déphasage",
	Pimgib = "Pimgib",
	["Plaguemaw the Rotting"] = "Pestegueule le Pourrissant",
	["Plugger Spazzring"] = "Lanfiche Brouillecircuit",
	["Postmaster Malown"] = "Postier Malown",
	["Priestess Delrissa"] = "Prêtresse Delrissa",
	["Prince Keleseth"] = "Prince Keleseth",
	["Prince Malchezaar"] = "Prince Malchezaar",
	["Prince Skaldrenox"] = "Prince Skaldrenox ",
	["Prince Taldaram"] = "Prince Taldaram",
	["Prince Tenris Mirkblood"] = "Prince Tenris Bourbassang",
	["Prince Tortheldrin"] = "Prince Tortheldrin",
	["Prince Valanar"] = "Prince Valanar",
	["Princess Huhuran"] = "Princesse Huhuran",
	["Princess Moira Bronzebeard"] = "Princesse Moira Barbe-de-bronze",
	["Princess Tempestria"] = "Princesse Tempestria",
	["Princess Theradras"] = "Princesse Theradras",
	["Princess Yauj"] = "Princesse Yauj",
	["Professor Putricide"] = "Professeur Putricide",
	["Pure Spawn of Hydross"] = "Pur rejeton d'Hydross",
	Pusillin = "Pusillin",
	["Pyroguard Emberseer"] = "Pyrogarde Prophète ardent",
	["Pyromancer Loregrain"] = "Pyromancien Blé-du-savoir",
	Quagmirran = "Bourbierreux",
	["Quartermaster Zigris"] = "Intendant Zigris",
	["Rage Winterchill"] = "Rage Froidhiver",
	Ragglesnout = "Groinfendu",
	["Raging Spirit"] = "Esprit déchaîné",
	Ragnaros = "Ragnaros",
	Rajh = "Rajh",
	["Ramstein the Gorger"] = "Ramstein Grandgosier",
	["Ras Frostwhisper"] = "Ras Murmegivre",
	Rattlegore = "Cliquettripes",
	["Razorclaw the Butcher"] = "Tranchegriffe le Boucher",
	["Razorgore the Untamed"] = "Tranchetripe l'Indompté",
	Razorlash = "Tranchefouet",
	Razorscale = "Tranchécaille",
	["Reliquary of Souls"] = "Le reliquaire des âmes",
	Renataki = "Renataki",
	["Restless Skeleton"] = "Squelette sans repos",
	Rethilgore = "Rethiltripe",
	Revelosh = "Revelosh",
	["Rhahk'Zor"] = "Rhahk'Zor",
	["Ribbly Screwspigot"] = "Ribbly Fermevanne",
	["Right Arm"] = "Bras droit",
	["Ring of Law"] = "Cercle de la loi",
	Roar = "Graou",
	["Rokad the Ravager"] = "Rodak le ravageur",
	["Rokdar the Sundered Lord"] = "Rokdar le Seigneur scindé",
	["Rokmar the Crackler"] = "Rokmar le Crépitant",
	["Rom'ogg Bonecrusher"] = "Broie-les-os Rom'ogg",
	Romulo = "Romulo",
	["Romulo & Julianne"] = "Romulo & Julianne",
	Roogug = "Roogug",
	Rotface = "Trognepus",
	Rotgrip = "Grippe-charogne",
	["Runemaster Molgeim"] = "Maître des runes Molgeim",
	["Runok Wildmane"] = "Runok Crin-sauvage",
	Ruuzlu = "Ruuzlu",
	["Salramm the Fleshcrafter"] = "Salramm le Façonneur de chair",
	["Sanctum Sentry"] = "Factionnaire du sanctum",
	["Sandarr Dunereaver"] = "Sandarr Ravadune",
	["Sandfury Executioner"] = "Bourreau Furie-des-sables",
	Sapphiron = "Saphiron",
	Sara = "Sara",
	["Saronite Animus"] = "Animus de saronite",
	Sartharion = "Sartharion",
	["Sathrovarr the Corruptor"] = "Sathrovarr le Corrupteur",
	["Saviana Ragefire"] = "Saviana Ragefeu",
	["Scarlet Commander Mograine"] = "Commandant écarlate Mograine",
	["Scourgelord Tyrannus"] = "Seigneur du Fléau Tyrannus",
	["Seeth'rel"] = "Fulmi'rel",
	["Selin Fireheart"] = "Selin Coeur-de-feu",
	["Sergeant Bly"] = "Sergent Bly",
	-- Setesh = "",
	["Shade of Akama"] = "Ombre d'Akama",
	["Shade of Aran"] = "Ombre d'Aran",
	["Shade of Eranikus"] = "Ombre d'Eranikus",
	["Shadikith the Glider"] = "Shadikith le glisseur",
	["Shadow Hunter Vosh'gajin"] = "Chasseresse des ombres Vosh'gajin",
	["Shadow of Leotheras"] = "Ombre de Leotheras",
	["Shadowpriest Sezz'ziz"] = "Prêtre des ombres Sezz'ziz",
	Shadron = "Obscuron",
	Shazzrah = "Shazzrah",
	["Shirrak the Dead Watcher"] = "Shirrak le Veillemort",
	["Siamat, Lord of South Wind"] = "Siamat, seigneur du vent du Sud",
	Sindragosa = "Sindragosa",
	["Sir Zeliek"] = "Sire Zeliek",
	["Sjonnir The Ironshaper"] = "Sjonnir le Sculptefer",
	["Skadi the Ruthless"] = "Skadi le Brutal",
	["Skarr the Unbreakable"] = "Bâlhafr l'Invaincu",
	["Skarvald the Constructor"] = "Skarvald le Constructeur",
	["Skra'gath"] = "Skra'gath",
	Skul = "Krân",
	Skum = "Skum",
	-- Slabhide = "",
	["Slad'ran"] = "Slad'ran",
	Sneed = "Sneed",
	["Sneed's Shredder"] = "Déchiqueteur de Sneed",
	["Solakar Flamewreath"] = "Solakar Voluteflamme",
	["Solarium Agent"] = "Agent du Solarium",
	["Solarium Priest"] = "Prêtre du Solarium",
	["Spirestone Battle Lord"] = "Seigneur de bataille Pierre-du-pic",
	["Spirestone Butcher"] = "Boucher Pierre-du-pic",
	["Spirestone Lord Magus"] = "Seigneur magus Pierre-du-pic",
	["Staff of Disintegration"] = "Bâton de désintégration",
	Stalagg = "Stalagg",
	Steelbreaker = "Brise-acier",
	["Stomper Kreeg"] = "Kreeg le Marteleur",
	Stonespine = "Echine-de-pierre",
	["Stormcaller Brundir"] = "Mande-foudre Brundir",
	Strawman = "Homme de paille",
	["Sulfuron Harbinger"] = "Messager de Sulfuron",
	Supremus = "Supremus",
	["Svala Sorrowgrave"] = "Svala Tristetombe",
	["Swamplord Musel'ek"] = "Seigneur des marais Musel'ek",
	Taerar = "Taerar",
	["Tainted Spawn of Hydross"] = "Rejeton d'Hydross souillé",
	["Talon King Ikiss"] = "Roi-serre Ikiss",
	["Taragaman the Hungerer"] = "Taragaman l'Affameur",
	["Targorr the Dread"] = "Targorr le Terrifiant",
	Tavarok = "Tavarok",
	Techbot = "Techbot",
	-- ["Temple Guardian Anhuur"] = "",
	Temporus = "Temporus",
	["Tendris Warpwood"] = "Tendris Crochebois",
	Tenebron = "Ténébron",
	["Terestian Illhoof"] = "Terestian Malsabot",
	["Teron Gorefiend"] = "Teron Fielsang",
	Thaddius = "Thaddius",
	["Thaladred the Darkener"] = "Thaladred l'Assombrisseur",
	["Thane Korth'azz"] = "Thane Korth'azz",
	["The Beast"] = "La Bête",
	["The Beasts of Northrend"] = "Les bêtes du Norfendre",
	["The Big Bad Wolf"] = "Le Grand Méchant Loup",
	["The Black Knight"] = "Le Chevalier noir",
	["The Black Stalker"] = "La Traqueuse noire",
	["The Blue Brothers"] = "Les Grands Bleus",
	["The Bug Family"] = "La famille insecte",
	["The Crone"] = "La Mégère",
	["The Curator"] = "Le conservateur",
	["The Eredar Twins"] = "Les jumelles érédars",
	["The Four Horsemen"] = "Les quatre cavaliers",
	["The Illidari Council"] = "Le conseil illidari",
	["The Iron Council"] = "Assemblée du fer",
	["The Lich King"] = "Le roi-liche",
	["The Lost Dwarves"] = "Les nains perdus",
	["The Lurker Below"] = "Le Rôdeur d'En bas",
	["The Maker"] = "Le Faiseur",
	["The Prophet Skeram"] = "Le Prophète Skeram",
	["The Prophet Tharon'ja"] = "Le prophète Tharon'ja",
	["The Ravenian"] = "Le Voracien",
	["The Razza"] = "La Razza",
	["The Seven Dwarves"] = "Les sept nains",
	["The Skybreaker"] = "Le Brise-ciel",
	["The Tribunal of Ages"] = "Le tribunal des Âges",
	["The Twin Emperors"] = "Les Empereurs jumeaux",
	["The Twin Val'kyr"] = "Les jumelles val'kyr",
	["The Unforgiven"] = "Le Condamné",
	["The Windreaver"] = "Ouraganien",
	["Theka the Martyr"] = "Theka le Martyr",
	Thorim = "Thorim",
	["Thorngrin the Tender"] = "Rirépine le Tendre",
	["Tidewalker Lurker"] = "Rôdeur marcheur-des-flots",
	["Timmy the Cruel"] = "Timmy le Cruel",
	Tinhead = "Tête de fer-blanc",
	["Tinkerer Gizlock"] = "Bricoleur Kadenaz",
	["Tirion Fordring"] = "Tirion Fordring",
	Tito = "Tito",
	["Toravon the Ice Watcher"] = "Toravon la Sentinelle de glace",
	["Trigore the Lasher"] = "Trigore le Flagelleur",
	Trollgore = "Trollétripe",
	["Tsu'zee"] = "Tsu'zee",
	["Tuten'kash"] = "Tuten'kash",
	["Twilight Lord Kelris"] = "Seigneur du crépuscule Kelris",
	["Urok Doomhowl"] = "Urok Hurleruine",
	["VX-001"] = "VX-001",
	["Vaelastrasz the Corrupt"] = "Vaelastrasz le Corrompu",
	["Val'kyr Shadowguard"] = "Garde de l'ombre val'kyr",
	["Valithria Dreamwalker"] = "Valithria Marcherêve",
	["Varian Wrynn"] = "Varian Wrynn",
	["Varos Cloudstrider"] = "Varos Arpentenuée",
	Vazruden = "Vazruden",
	["Vazruden the Herald"] = "Vazruden le Héraut",
	Vectus = "Vectus",
	Vem = "Vem",
	Veng = "Veng",
	["Veras Darkshadow"] = "Veras Ombrenoir",
	["Verdan the Everliving"] = "Verdan l'Immortel",
	Verek = "Verek",
	Vesperon = "Vespéron",
	Vexallus = "Vexallus",
	["Veyzhak the Cannibal"] = "Veyzhak le Cannibale",
	["Vile'rel"] = "Ignobl'rel",
	Viscidus = "Viscidus",
	["Viscous Fallout"] = "Retombée visqueuse",
	["Void Reaver"] = "Saccageur du Vide",
	Volkhan = "Volkhan",
	["War Master Voone"] = "Maître de guerre Voone",
	["Warbringer O'mrogg"] = "Porteguerre O'mrogg",
	["Warchief Blackhand Piece"] = "Pion du Chef de guerre Main-noire",
	["Warchief Kargath Bladefist"] = "Chef de guerre Kargath Lamepoing",
	["Warchief Rend Blackhand"] = "Chef de guerre Rend Main-noire",
	["Warden Mellichar"] = "Gardien Mellichar",
	["Warder Stilgiss"] = "Gardien Stilgiss",
	["Warlord Kalithresh"] = "Seigneur de guerre Kalithresh",
	["Warmaul Champion"] = "Champion Cogneguerre",
	["Warp Slicer"] = "Tranchoir dimensionnel",
	["Warp Splinter"] = "Brise-dimension",
	["Watchkeeper Gargolmar"] = "Gardien des guetteurs Gargolmar",
	Weaver = "Tisserand",
	["Witch Doctor Zum'rah"] = "Sorcier-docteur Zum'rah",
	["Wolf Master Nandos"] = "Maître-loup Nandos",
	["Wrath-Scryer Soccothrates"] = "Scrute-courroux Soccothrates",
	Wushoolay = "Wushoolay",
	["XT-002 Deconstructor"] = "Déconstructeur XT-002",
	Xevozz = "Xevozz",
	["Yogg-Saron"] = "Yogg-Saron",
	Yor = "Yor",
	Ysondre = "Ysondre",
	Zekkis = "Zekkis",
	["Zelemar the Wrathful"] = "Zelemar le Courroucé",
	["Zereketh the Unbound"] = "Zereketh le Délié",
	Zerillis = "Zerillis",
	["Zevrim Thornhoof"] = "Zevrim Sabot-de-ronce",
	Zolo = "Zolo",
	["Zul'Farrak Dead Hero"] = "Héros mort de Zul'Farrak",
	["Zul'Lor"] = "Zul'Lor",
	["Zul'jin"] = "Zul'jin",
	["Zul'tore"] = "Zul'tore",
	["Zuramat the Obliterator"] = "Zuramat l'Oblitérateur",
}
elseif GAME_LOCALE == "ruRU" then
	lib:SetCurrentTranslations {
	Acidmaw = "Кислотная Утроба",
	Aeonus = "Эонус",
	["Aerial Command Unit"] = "Воздушное судно",
	["Agathelos the Raging"] = "Агателос Свирепый",
	["Aggem Thorncurse"] = "Аггем Терновое Проклятие",
	Ahune = "Ахун",
	["Akil'zon"] = "Акил'зон",
	["Aku'mai"] = "Аку'май",
	["Al'ar"] = "Ал'ар",
	["Algalon the Observer"] = "Алгалон Наблюдатель",
	-- Altarius = "",
	["Alzzin the Wildshaper"] = "Алззин Перевертень",
	Amanitar = "Аманитар",
	["Ambassador Flamelash"] = "Посол Огнехлыст",
	["Ambassador Hellmaw"] = "Посол Гиблочрев",
	-- Ammunae = "",
	["Amnennar the Coldbringer"] = "Амненнар Хладовей",
	["Ancient Stone Keeper"] = "Древний Хранитель Камня",
	Anetheron = "Анетерон",
	["Anger'rel"] = "Гнев'рел",
	Anomalus = "Аномалус",
	-- Anraphet = "",
	["Antu'sul"] = "Анту'сул",
	["Anub'Rekhan"] = "Ануб'Рекан",
	["Anub'arak"] = "Ануб'арак",
	["Anub'shiah"] = "Ануб'шиа",
	["Anubisath Defender"] = "Защитник-анубисат",
	["Anubisath Guardian"] = "Анубисат-страж",
	Anzu = "Анзу",
	["Arcane Watchman"] = "Чародейский стражник",
	["Arcanist Doan"] = "Чародей Доан",
	Archaedas = "Аркедас",
	["Archavon the Stone Watcher"] = "Аркавон Страж Камня",
	Archimonde = "Архимонд",
	["Archivist Galford"] = "Архивариус Галфорд",
	["Archmage Arugal"] = "Верховный маг Аругал",
	["Argent Confessor Paletress"] = "Исповедница Серебряного Авангарда Пейлтресс",
	["Arugal's Voidwalker"] = "Демон Бездны Аругала",
	-- ["Ascendant Lord Obsidius"] = "",
	-- ["Asim al Akir"] = "",
	["Assault Bot"] = "Штурмовой робот",
	-- ["Assembly of Iron"] = "",
	["Atal'alarion"] = "Атал'аларион",
	-- Atramedes = "",
	["Attumen the Huntsman"] = "Ловчий Аттумен",
	Auriaya = "Ауриайя",
	Avalanchion = "Лавинион",
	["Avatar of Hakkar"] = "Аватара Хаккара",
	["Ayamiss the Hunter"] = "Айамисса Охотница",
	Azgalor = "Азгалор",
	["Azshir the Sleepless"] = "Азшир Неспящий",
	Azuregos = "Азурегос",
	["Bael'Gar"] = "Бейл'Гор",
	Baelog = "Бейлог",
	Balnazzar = "Балназзар",
	["Baltharus the Warborn"] = "Балтар Рожденный в Битве",
	["Bannok Grimaxe"] = "Баннок Люторез",
	["Baron Aquanis"] = "Барон Акванис",
	["Baron Charr"] = "Барон Огнерр",
	["Baron Geddon"] = "Барон Геддон",
	["Baron Kazum"] = "Барон Казум",
	["Baron Rivendare"] = "Барон Ривендер",
	["Baron Silverlaine"] = "Барон Сребролен",
	["Baroness Anastari"] = "Баронесса Анастари",
	["Battleguard Sartura"] = "Боевой страж Сартура",
	["Bazil Thredd"] = "Базиль Тредд",
	Bazzalan = "Баззалан",
	-- Beauty = "",
	["Black Guard Swordsmith"] = "Оружейник Черной Стражи",
	["Blackheart the Inciter"] = "Черносерд Проповедник",
	["Blind Hunter"] = "Слепой охотник",
	["Blindeye the Seer"] = "Слепоглаз Ясновидец",
	["Blood Guard Porung"] = "Кровавый страж Порунг",
	["Blood Prince Council"] = "Совет Принцев Крови",
	["Blood Princes"] = "Совет Принцев Крови",
	["Blood Steward of Kirtonos"] = "Кровавая прислужница Киртоноса",
	["Blood-Queen Lana'thel"] = "Кровавая королева Лана'тель",
	["Bloodlord Mandokir"] = "Мандокир Повелитель Крови",
	["Bloodmage Thalnos"] = "Волшебник Крови Талнос",
	Boahn = "Боан",
	["Bomb Bot"] = "Бомбот",
	["Brain of Yogg-Saron"] = "Мозг Йогг-Сарона",
	["Brainwashed Noble"] = "Зомбированный дворянин",
	Broggok = "Броггок",
	Brokentoe = "Хромоног",
	Bronjahm = "Броньям",
	["Broodlord Lashlayer"] = "Предводитель драконов Разящий Бич",
	["Bruegal Ironknuckle"] = "Бругал Железный Кулак",
	Brutallus = "Бруталл",
	["Burning Felguard"] = "Пылающий страж Скверны",
	["Buru the Gorger"] = "Буру Ненасытный",
	["C'Thun"] = "К'Тун",
	["Cache of the Firelord"] = "Тайник повелителя огня",
	["Cache of the Legion"] = "Тайник Легиона",
	["Cannon Master Willey"] = "Мастер-канонир Вилли",
	["Captain Greenskin"] = "Капитан Зеленямс",
	["Captain Kromcrush"] = "Капитан Давигром",
	["Captain Skarloc"] = "Капитан Скарлок",
	["Celebras the Cursed"] = "Селебрас Проклятый",
	["Charlga Razorflank"] = "Чарлга Остробок",
	["Chess Event"] = "Шахматы",
	["Chest of The Seven"] = "Сундук Семерых",
	["Chief Ukorz Sandscalp"] = "Вождь Укорз Песчаная Плешь",
	-- Chimaeron = "",
	["Cho'Rush the Observer"] = "Чо'Раш Наблюдатель",
	Chromaggus = "Хроммагус",
	["Chrono Lord Deja"] = "Повелитель времени Дежа",
	["Chrono-Lord Epoch"] = "Хронолорд Эпох",
	Claw = "Коготь",
	["Coilfang Elite"] = "Гвардеец резервуара Кривого Клыка",
	["Coilfang Strider"] = "Странник резервуара Кривого Клыка",
	["Commander Kolurg"] = "Командир Колург",
	["Commander Sarannis"] = "Командир Сараннис",
	["Commander Springvale"] = "Командир Ручьедол",
	["Commander Stoutbeard"] = "Командир Пивобород",
	-- ["Commander Ulthok"] = "",
	["Constructor & Controller"] = "Строитель и Контролирующий",
	Cookie = "Пирожок",
	-- Corborus = "",
	["Coren Direbrew"] = "Корен Худовар",
	-- ["Corla, Herald of Twilight"] = "",
	["Cosmic Infuser"] = "Вселенский вдохновитель",
	["Crimson Hammersmith"] = "Молотобоец из Багрового Легиона",
	["Crowd Pummeler 9-60"] = "Толпогон 9-60",
	["Crystal Fang"] = "Хрустальный Клык",
	Cyanigosa = "Синигоса",
	["Dalliah the Doomsayer"] = "Далия Глашатай Судьбы",
	["Dalronn the Controller"] = "Далронн Контролирующий",
	["Dark Coffer"] = "Черный ящик",
	["Dark Iron Ambassador"] = "Посол из клана Черного Железа",
	["Darkmaster Gandling"] = "Темный магистр Гандлинг",
	["Darkweaver Syth"] = "Темнопряд Сит",
	["Death Knight Darkreaver"] = "Рыцарь смерти Темный Терзатель",
	["Death Knight Understudy"] = "Ученик рыцаря смерти",
	["Death Speaker Jargba"] = "Вестник смерти Джаргба",
	["Deathbound Ward"] = "Заклятый страж",
	["Deathbringer Saurfang"] = "Саурфанг Смертоносный",
	["Deathspeaker High Priest"] = "Вестник смерти - верховный жрец",
	["Deathstalker Visceri"] = "Страж смерти Визери",
	["Deathsworn Captain"] = "Капитан служителей Смерти",
	Devastation = "Сокрушитель",
	["Deviate Faerie Dragon"] = "Загадочный волшебный дракон",
	["Devourer of Souls"] = "Пожиратель Душ",
	["Dextren Ward"] = "Декстрен Вард",
	["Digmaster Shovelphlange"] = "Мастер Лопаторук",
	["Doctor Theolen Krastinov"] = "Доктор Теолен Крастинов",
	["Doom Lord Kazzak"] = "Владыка Судеб Каззак",
	["Doom'rel"] = "Рок'рел",
	Doomwalker = "Судьболом",
	["Dope'rel"] = "Праздн'рел",
	Dorothee = "Дороти",
	-- ["Drahga Shodowburner"] = "",
	["Drakkari Colossus"] = "Колосс Драккари",
	["Drakos the Interrogator"] = "Дракос Дознаватель",
	Dreadscale = "Жуткая Чешуя",
	Dreamscythe = "Жнец Снов",
	["Dust Covered Chest"] = "Пыльный сундук",
	Dustwraith = "Пыльный Призрак",
	["Eadric the Pure"] = "Эдрик Чистый",
	["Earthcaller Halmgar"] = "Заклинательница земли Халмгар",
	-- ["Earthrager Ptah"] = "",
	Ebonroc = "Черноскал",
	["Eck the Ferocious"] = "Эк Свирепый",
	["Edwin VanCleef"] = "Эдвин ван Клиф",
	["Elder Brightleaf"] = "Древень Яркий Лист",
	["Elder Ironbranch"] = "Древень Железная Ветвь",
	["Elder Nadox"] = "Старейшина Надокс",
	["Elder Stonebark"] = "Древень Каменная Кора",
	["Electrocutioner 6000"] = "Электрошокер 6000",
	["Emalon the Storm Watcher"] = "Эмалон Страж Бури",
	Emeriss = "Эмерисс",
	["Emperor Dagran Thaurissan"] = "Император Дагран Таурисан",
	["Emperor Vek'lor"] = "Император Век'лор",
	["Emperor Vek'nilash"] = "Император Век'нилаш",
	Entropius = "Энтропий",
	["Eonar's Gift"] = "Дар Эонара",
	["Epoch Hunter"] = "Охотник Вечности",
	Erekem = "Эрекем",
	["Eressea Dawnsinger"] = "Эрессея Певица Рассвета",
	["Eric \"The Swift\""] = "Эрик \"Быстрый\"",
	-- Erudax = "",
	-- ["Erunak Stonespeaker"] = "",
	["Essence of Anger"] = "Воплощение гнева",
	["Essence of Desire"] = "Воплощение желания",
	["Essence of Suffering"] = "Воплощение страдания",
	Eviscerator = "Потрошитель",
	["Exarch Maladaar"] = "Экзарх Маладаар",
	["Expedition Commander"] = "Начальник экспедиции",
	["Eydis Darkbane"] = "Эйдис Погибель Тьмы",
	["Eye of C'Thun"] = "Око К'Туна",
	["Faction Champions"] = "Чемпионы фракций",
	["Fallen Champion"] = "Павший воитель",
	Falric = "Фалрик",
	["Falric and Marwyn"] = "Фалрик и Марвин",
	["Fankriss the Unyielding"] = "Фанкрисс Неприклонный",
	["Fathom-Lord Karathress"] = "Повелитель глубин Каратресс",
	Felmyst = "Пророк Скверны",
	["Fenrus the Devourer"] = "Фенрус Пожиратель",
	["Feral Defender"] = "Дикий защитник",
	Festergut = "Тухлопуз",
	Feugen = "Фойген",
	["Fineous Darkvire"] = "Точень Темнострой",
	Firemaw = "Огнечрев",
	["Fjola Lightbane"] = "Фьола Погибель Света",
	["Flame Leviathan"] = "Огненный Левиафан",
	Flamegor = "Пламегор",
	["Foreman Thistlenettle"] = "Штейгер Чертополох",
	["Forgemaster Garfrost"] = "Начальник кузни Гархлад",
	-- ["Forgemaster Throngus"] = "",
	["Four Horsemen Chest"] = "Сундук Четырех Всадников",
	["Fras Siabi"] = "Фрас Сиаби",
	Freya = "Фрейя",
	["Gahz'ranka"] = "Газ'ранка",
	["Gahz'rilla"] = "Газ'рилла",
	["Gal'darah"] = "Гал'дара",
	["Galgann Firehammer"] = "Галганн Огнемолот",
	Garr = "Гарр",
	["Garrosh Hellscream"] = "Гаррош Адский Крик",
	Gasher = "Ранокол",
	["Gatewatcher Gyro-Kill"] = "Страж ворот Точеный Нож",
	["Gatewatcher Iron-Hand"] = "Страж ворот Стальная Клешня",
	["Gathios the Shatterer"] = "Гатиос Изувер",
	Gehennas = "Гееннас",
	Gelihast = "Гелихаст",
	Gelk = "Джелк",
	["General Angerforge"] = "Генерал Кузня Гнева",
	["General Bjarngrim"] = "Генерал Бьярнгрин",
	["General Drakkisath"] = "Генерал Драккисат",
	-- ["General Husam"] = "",
	["General Rajaxx"] = "Генерал Раджакс",
	-- ["General Umbriss"] = "",
	["General Vezax"] = "Генерал Везакс",
	["General Zarithrian"] = "Генерал Заритриан",
	["Ghamoo-ra"] = "Гхаму-ра",
	["Ghaz'an"] = "Газ'ан",
	["Ghok Bashguud"] = "Гок Крепкобив",
	Gilnid = "Гилнид",
	["Gizrul the Slavener"] = "Гизрул Поработитель",
	["Gloom'rel"] = "Мрак'нел",
	Gluth = "Глут",
	Glutton = "Обжора",
	["Golem Lord Argelmach"] = "Повелитель големов Аргелмах",
	["Golemagg the Incinerator"] = "Маг-лорд из клана Гордок",
	["Goraluk Anvilcrack"] = "Горалук Треснувшая Наковальня",
	["Gormok the Impaler"] = "Гормок Пронзающий Бивень",
	["Gorosh the Dervish"] = "Горош Дервиш",
	["Gortok Palehoof"] = "Горток Бледное Копыто",
	["Gothik the Harvester"] = "Готик Жнец",
	["Grand Astromancer Capernian"] = "Верховный звездочет Каперниан",
	["Grand Champions"] = "Абсолютные чемпионы",
	["Grand Crusader Dathrohan"] = "Верховный рыцарь Датрохан",
	["Grand Magus Telestra"] = "Великая ведунья Телестра",
	-- ["Grand Vizier Ertan"] = "",
	["Grand Warlock Alythess"] = "Главная чернокнижница Алитесса",
	["Grand Warlock Nethekurse"] = "Главный чернокнижник Пустоклят",
	["Grand Widow Faerlina"] = "Великая вдова Фарлина",
	["Grandmaster Vorpil"] = "Великий мастер Ворпил",
	Grandmother = "Бабушка",
	["Grethok the Controller"] = "Греток Регулятор",
	["Gri'lek"] = "Гри'лек",
	Grimlok = "Гримлок",
	Grizzle = "Гриззл",
	Grobbulus = "Гроббулус",
	Grubbis = "Грязнюк",
	["Gruul the Dragonkiller"] = "Груул Драконобой",
	["Guard Fengus"] = "Стражник Фенгус",
	["Guard Mol'dar"] = "Стражник Мол'дар",
	["Guard Slip'kik"] = "Стражник Слип'кик",
	["Guardian of Yogg-Saron"] = "Страж Йогг-Сарона",
	["Gurtogg Bloodboil"] = "Гуртогг Кипящая Кровь",
	Gyth = "Гит",
	Hadronox = "Хадронокс",
	Hakkar = "Хаккар",
	Halazzi = "Халаззи",
	Halion = "Халион",
	Halycon = "Халикон",
	Hamhock = "Хрупконог",
	["Harbinger Skyriss"] = "Предвестник Скайрисс",
	["Hate'rel"] = "Злоб'рел",
	["Hazza'rah"] = "Хазза'рах",
	Hazzas = "Хаззас",
	["Headless Horseman"] = "Всадник без головы",
	["Hearthsinger Forresten"] = "Певчий Форрестен",
	["Hedrum the Creeper"] = "Хедрум Ползун",
	["Heigan the Unclean"] = "Хейган Нечестивый",
	["Hellfire Channeler"] = "Чаротворец из цитадели Адского Пламени",
	["Henry Stern"] = "Генри Штерн",
	["Herald Volazj"] = "Глашатай Волаж",
	Herod = "Герод",
	["Hex Lord Malacrass"] = "Повелитель проклятий Малакрасс",
	["High Astromancer Solarian"] = "Верховный звездочет Солариан",
	["High Botanist Freywinn"] = "Верховный ботаник Фрейвин",
	["High Inquisitor Fairbanks"] = "Верховный инквизитор Фэйрбанкс",
	["High Inquisitor Whitemane"] = "Верховный инквизитор Вайтмейн",
	["High Interrogator Gerstahn"] = "Верховный дознаватель Герштан",
	["High King Maulgar"] = "Король Молгар",
	["High Marshal Whirlaxis"] = "Маршал Кольцекрут",
	["High Nethermancer Zerevor"] = "Верховный пустомант Зеревор",
	["High Overlord Saurfang"] = "Верховный правитель Саурфанг",
	["High Priest Thekal"] = "Верховный жрец Текал",
	["High Priest Venoxis"] = "Верховный жрец Веноксис",
	["High Priestess Arlokk"] = "Верховная жрица Арлокк",
	-- ["High Priestess Azil"] = "",
	["High Priestess Jeklik"] = "Верховная жрица Джеклик",
	["High Priestess Mar'li"] = "Верховная жрица Мар'ли",
	["High Priestess of Thaurissan"] = "Верховная жрица Тауриссана",
	-- ["High Prophet Barim"] = "",
	["High Warlord Naj'entus"] = "Верховный Полководец Надж'ентус",
	["Highlord Mograine"] = "Верховный лорд Могрейн",
	["Highlord Omokk"] = "Вождь Омокк",
	Hodir = "Ходир",
	["Houndmaster Grebmar"] = "Псарь Гребмар",
	["Houndmaster Loksey"] = "Псарь Локси",
	Hukku = "Хукку",
	Hungarfen = "Топеглад",
	["Hurley Blackbreath"] = "Харли Чернопых",
	["Hyakiss the Lurker"] = "Хиакисс Скрытень",
	["Hydromancer Thespia"] = "Гидромант Теспия",
	["Hydromancer Velratha"] = "Гидромант Велрата",
	Hydrospawn = "Гидротварь",
	["Hydross the Unstable"] = "Гидросс Нестабильный",
	["Ice Sphere"] = "Ледяная сфера",
	["Icecrown Gunship Battle"] = "Бой на кораблях",
	Icehowl = "Ледяной Рев",
	Ichoron = "Гнойрон",
	Ick = "Ик",
	["Ignis the Furnace Master"] = "Повелитель Горнов Игнис",
	["Illidan Stormrage"] = "Иллидан Ярость Бури",
	["Illidari Council"] = "Совет Иллидари",
	["Illyanna Ravenoak"] = "Иллиана Воронья Ольха",
	["Immol'thar"] = "Бессмер'тер",
	["Infinite Corruptor"] = "Осквернитель из рода Бесконечности",
	["Infinity Blades"] = "Клинки Бесконечности",
	["Ingvar the Plunderer"] = "Ингвар Расхититель",
	["Instructor Malicia"] = "Инструктор Коварница",
	["Instructor Razuvious"] = "Инструктор Разувий",
	["Interrogator Vishas"] = "Дознаватель Вишас",
	Ionar = "Ионар",
	Ironaya = "Иронайа",
	Ironspine = "Железноспин",
	Isalien = "Изалиен",
	-- Isiset = "",
	Jade = "Нефрит",
	["Jammal'an the Prophet"] = "Джаммал'ан Пророк",
	["Jan'alai"] = "Джан'алай",
	["Jandice Barov"] = "Джандис Барова",
	["Jed Runewatcher"] = "Джед Руновед",
	["Jedoga Shadowseeker"] = "Джедога Искательница Теней",
	["Jergosh the Invoker"] = "Жергош Призыватель Духов",
	["Jin'do the Hexxer"] = "Джин'до Проклинатель",
	["Jormungar Behemoth"] = "Йормунгарское чудовище",
	Jormungars = "Йормунгары",
	Julianne = "Джулианна",
	["Junk Bot"] = "Барахлобот",
	["Kael'thas Sunstrider"] = "Кель'тас Солнечный Скиталец",
	Kalecgos = "Калесгос",
	["Kam Deepfury"] = "Кам Гневливый",
	-- ["Karsh Steelbender"] = "",
	["Kaz'rogal"] = "Каз'рогал",
	["Kazkaz the Unholy"] = "Казказ Нечистый",
	["Kel'Thuzad"] = "Кел'Тузад",
	["Keli'dan the Breaker"] = "Кели'дан Разрушитель",
	Keristrasza = "Керистраза",
	["Kiggler the Crazed"] = "Кигглер Безумный",
	["Kil'jaeden"] = "Кил'джеден",
	["Kil'rek"] = "Кил'рек",
	["King Dred"] = "Король Дред",
	["King Gordok"] = "Король Гордок",
	["King Llane Piece"] = "Король Ллейн",
	["King Ymiron"] = "Король Имирон",
	["Kirtonos the Herald"] = "Киртонос Глашатай",
	["Knot Thimblejack's Cache"] = "Тайник Уззла Наперстяка",
	Kolk = "Колк",
	Kologarn = "Кологарн",
	["Koralon the Flame Watcher"] = "Коралон Страж Огня",
	Kormok = "Кормок",
	Kresh = "Криг",
	Krick = "Крик",
	["Krick and Ick"] = "Крик и Ик",
	["Krik'thir the Gatewatcher"] = "Крик'Тир Хранитель Врат",
	["Krosh Firehand"] = "Крош Огненная Рука",
	Krystallus = "Кристаллус",
	Kurinnaxx = "Куриннакс",
	["Lady Anacondra"] = "Повелительница Анакондра",
	["Lady Blaumeux"] = "Леди Бломе",
	["Lady Deathwhisper"] = "Леди Смертный Шепот",
	["Lady Illucia Barov"] = "Леди Иллюсия Барова",
	["Lady Malande"] = "Леди Маланда",
	-- ["Lady Naz'jar"] = "",
	["Lady Sacrolash"] = "Леди Сакролаш",
	["Lady Sarevess"] = "Леди Саревесс",
	["Lady Vashj"] = "Леди Вайш",
	Laj = "Ладж",
	Landslide = "Сель",
	Lavanthor = "Лавантор",
	["Left Arm"] = "Левая рука",
	["Leotheras the Blind"] = "Леотерас Слепец",
	Lethon = "Летон",
	Lethtendris = "Лефтендрис",
	["Leviathan Mk II"] = "Левиафан II",
	["Ley-Guardian Eregos"] = "Хранитель энергии Эрегос",
	["Lieutenant Drake"] = "Лейтенант Дрейк",
	["Lieutenant General Andorov"] = "Генерал-лейтенант Андоров",
	Loatheb = "Лотхиб",
	-- Lockmaw = "",
	Loken = "Локен",
	["Lord Alexei Barov"] = "Лорд Алексей Баров",
	["Lord Cobrahn"] = "Лорд Кобран",
	["Lord Hel'nurath"] = "Лорд Хел'нурат",
	["Lord Incendius"] = "Лорд Опалитель",
	["Lord Jaraxxus"] = "Лорд Джараксус",
	["Lord Kazzak"] = "Лорд Каззак",
	["Lord Kri"] = "Лорд Кри",
	["Lord Marrowgar"] = "Лорд Ребрад",
	["Lord Pythas"] = "Лорд Питонас",
	["Lord Roccor"] = "Лорд Роккор",
	["Lord Sanguinar"] = "Лорд Сангвинар",
	["Lord Serpentis"] = "Лорд Серпентис",
	["Lord Skwol"] = "Лорд Сквол",
	["Lord Valthalak"] = "Лорд Вальтхалак",
	["Lord Victor Nefarius"] = "Лорд Виктор Нефариус",
	["Lord Vyletongue"] = "Лорд Злоязыкий",
	["Lorekeeper Polkelt"] = "Сказитель Полкелт",
	["Lorgus Jett"] = "Лоргус Джетт",
	Loro = "Лоро",
	Lucifron = "Люцифрон",
	["M'uru"] = "М'ару",
	["Mad Magglish"] = "Безумный Магглиш",
	Maexxna = "Мексна",
	["Mage-Lord Urom"] = "Маг-лорд Уром",
	["Magister Kalendris"] = "Магистр Календрис",
	["Magistrate Barthilas"] = "Мировой судья Бартилас",
	Magmadar = "Магмадар",
	-- Magmaw = "",
	Magmus = "Магмус",
	Magra = "Магра",
	Magtheridon = "Магтеридон",
	["Maiden of Grief"] = "Дева Скорби",
	["Maiden of Virtue"] = "Благочестивая дева",
	["Majordomo Executus"] = "Мажордом Экзекутус",
	["Mal'Ganis"] = "Мал'Ганис",
	Malacrass = "Малакрасс",
	["Maleki the Pallid"] = "Малекай Бледный",
	-- Maloriak = "",
	Malygos = "Малигос",
	Maraudos = "Мародос",
	["Marduk Blackpool"] = "Мардук Блэкпул",
	["Marisa du'Paige"] = "Мариса дю Пэж",
	Marwyn = "Марвин",
	["Master Engineer Telonicus"] = "Старший инженер Телоникус",
	["Maur Grimtotem"] = "Маур Зловещий Тотем",
	Meathook = "Мясной Крюк",
	["Mechano-Lord Capacitus"] = "Механо-лорд Конденсарон",
	Medivh = "Медив",
	["Mekgineer Steamrigger"] = "Мекжинер Паропуск",
	["Mekgineer Thermaplugg"] = "Мекжинер Термоштепсель",
	["Mennu the Betrayer"] = "Менну Предатель",
	["Meshlok the Harvester"] = "Мешлок Жнец",
	Midnight = "Полночь",
	Mijan = "Миджан",
	Mimiron = "Мимирон",
	-- ["Mindbender Ghur'sha"] = "",
	["Miner Johnson"] = "Шахтер Джонсон",
	["Mistress of Pain"] = "Госпожа Боли",
	Moam = "Моам",
	Mogor = "Могор",
	["Mokra the Skullcrusher"] = "Мокра Дробитель Черепов",
	Moorabi = "Мураби",
	["Mor Grayhoof"] = "Мор Серое Копыто",
	Moragg = "Морагг",
	["Mordresh Fire Eye"] = "Мордреш Огненный Глаз",
	Moroes = "Мороуз",
	["Morogrim Tidewalker"] = "Морогрим Волноступ",
	Morphaz = "Морфаз",
	["Mother Shahraz"] = "Матушка Шахраз",
	["Mother Smolderweb"] = "Мать Дымная Паутина",
	["Mr. Smite"] = "Мистер Каюк",
	["Muradin Bronzebeard"] = "Мурадин Бронзобород",
	["Murkblood Twin"] = "Близнец Темной Крови",
	["Murkblood Twins"] = "Близнецы Темной Крови",
	Murmur = "Бормотун",
	["Murta Grimgut"] = "Мурта Мрачнобрюх",
	Mushgog = "Мушгог",
	["Mutanus the Devourer"] = "Мутанус Пожиратель",
	Nalorakk = "Налоракк",
	Nazan = "Назан",
	Nefarian = "Нефариан",
	["Nekrum Gutchewer"] = "Некрум Кишкожуй",
	["Nerub'enkan"] = "Неруб'энкан",
	["Nethermancer Sepethrea"] = "Пустомант Сепетрея",
	Netherspite = "Гнев Пустоты",
	["Netherstrand Longbow"] = "Длинный лук Края Пустоты",
	["Nexus-Prince Shaffar"] = "Принц Шаффар",
	Nightbane = "Ночная Погибель",
	["Noth the Plaguebringer"] = "Нот Чумной",
	["Novos the Summoner"] = "Новос Призыватель",
	Noxxion = "Ноксион",
	["Obsidian Sentinel"] = "Обсидиановый часовой",
	["Odo the Blindwatcher"] = "Одо Слепой Страж",
	Oggleflint = "Огглфлинт",
	["Ogom the Wretched"] = "Огом Презренный",
	Ohgan = "Охган",
	["Ok'thor the Breaker"] = "Ок'тор Разрушитель",
	Olaf = "Олаф",
	["Old Serra'kis"] = "Старина Серракис",
	["Olm the Summoner"] = "Олм Созывающий",
	["Omor the Unscarred"] = "Омор Неодолимый",
	Onyxia = "Ониксия",
	-- ["Opera Event"] = "",
	["Orgrim's Hammer"] = "Молот Оргрима",
	["Ormorok the Tree-Shaper"] = "Орморок Воспитатель Дерев",
	["Oro Eyegouge"] = "Оро Вырвиглаз",
	["Ossirian the Unscarred"] = "Оссириан Неуязвимый",
	Ouro = "Оуро",
	["Overlord Ramtusk"] = "Властитель Таранный Клык",
	["Overlord Wyrmthalak"] = "Властитель Змейталак",
	["Overmaster Pyron"] = "Подчинитель Пирон",
	["Overseer Tidewrath"] = "Наблюдатель Гнев Прилива",
	-- Ozruk = "",
	-- Ozumat = "",
	Pandemonius = "Пандемоний",
	["Panzor the Invincible"] = "Панцер Непобедимый",
	Patchwerk = "Лоскутик",
	["Pathaleon the Calculator"] = "Паталеон Вычислитель",
	Phalanx = "Фаланкс",
	["Phaseshift Bulwark"] = "Фазовый колет",
	Pimgib = "Пимгиб",
	["Plaguemaw the Rotting"] = "Чумобрюх Гнилой",
	["Plugger Spazzring"] = "Штоппор Наливалс",
	["Postmaster Malown"] = "Почтальон Мэлоун",
	["Priestess Delrissa"] = "Жрица Делрисса",
	["Prince Keleseth"] = "Принц Келесет",
	["Prince Malchezaar"] = "Принц Малчезар",
	["Prince Skaldrenox"] = "Принц Скальдренокс",
	["Prince Taldaram"] = "Принц Талдарам",
	["Prince Tenris Mirkblood"] = "Принц Тенрис Мутная Кровь",
	["Prince Tortheldrin"] = "Принц Тортелдрин",
	["Prince Valanar"] = "Принц Валанар",
	["Princess Huhuran"] = "Принцесса Хухуран",
	["Princess Moira Bronzebeard"] = "Принцесса Мойра Бронзобород",
	["Princess Tempestria"] = "Принцесса Бурилла",
	["Princess Theradras"] = "Принцесса Терадрас",
	["Princess Yauj"] = "Принцесса Яудж",
	["Professor Putricide"] = "Профессор Мерзоцид",
	["Pure Spawn of Hydross"] = "Чистое порождение Гидросса",
	Pusillin = "Пузиллин",
	["Pyroguard Emberseer"] = "Пиростраж Созерцатель Углей",
	["Pyromancer Loregrain"] = "Пироман Зерно Мудрости",
	Quagmirran = "Квагмирран",
	["Quartermaster Zigris"] = "Интендант Зигрис",
	["Rage Winterchill"] = "Лютый Хлад",
	Ragglesnout = "Круглорыл",
	["Raging Spirit"] = "Гневный дух",
	Ragnaros = "Рагнарос",
	-- Rajh = "",
	["Ramstein the Gorger"] = "Рамштайн Ненасытный",
	["Ras Frostwhisper"] = "Рас Ледяной Шепот",
	Rattlegore = "Громоклин",
	["Razorclaw the Butcher"] = "Остроклык Мясник",
	["Razorgore the Untamed"] = "Бритвосмерт Неукротимый",
	Razorlash = "Бритвохлест",
	Razorscale = "Острокрылая",
	["Reliquary of Souls"] = "Гробница душ",
	Renataki = "Ренатаки",
	["Restless Skeleton"] = "Недремлющий скелет",
	Rethilgore = "Ретилгор",
	Revelosh = "Ревелош",
	["Rhahk'Zor"] = "Рак'Зор",
	["Ribbly Screwspigot"] = "Риббли Крутипроб",
	["Right Arm"] = "Правая рука",
	-- ["Ring of Law"] = "",
	Roar = "Хохотун",
	["Rokad the Ravager"] = "Рокад Опустошитель",
	["Rokdar the Sundered Lord"] = "Рокдар Покинутый Лорд",
	["Rokmar the Crackler"] = "Рокмар Трескун",
	-- ["Rom'ogg Bonecrusher"] = "",
	Romulo = "Ромуло",
	["Romulo & Julianne"] = "Ромуло и Джульенна",
	Roogug = "Ругуг",
	Rotface = "Гниломорд",
	Rotgrip = "Гнилопасть",
	["Runemaster Molgeim"] = "Мастер рун Молгейм",
	["Runok Wildmane"] = "Рунок Буйногривый",
	Ruuzlu = "Руузлу",
	["Salramm the Fleshcrafter"] = "Салрамм Плоторез",
	["Sanctum Sentry"] = "Часовой святилища",
	["Sandarr Dunereaver"] = "Сандарр Разоритель Барханов",
	["Sandfury Executioner"] = "Палач из клана Песчаной Бури",
	Sapphiron = "Сапфирон",
	Sara = "Сара",
	["Saronite Animus"] = "Саронитовый враг",
	Sartharion = "Сартарион",
	["Sathrovarr the Corruptor"] = "Сатроварр Осквернитель",
	["Saviana Ragefire"] = "Савиана Огненная Пропасть",
	["Scarlet Commander Mograine"] = "Командир Могрейн из Алого ордена",
	["Scourgelord Tyrannus"] = "Повелитель Плети Тираний",
	["Seeth'rel"] = "Сит'рел",
	["Selin Fireheart"] = "Селин Огненное Сердце",
	["Sergeant Bly"] = "Сержант Блай",
	-- Setesh = "",
	["Shade of Akama"] = "Тень Акамы",
	["Shade of Aran"] = "Тень Арана",
	["Shade of Eranikus"] = "Тень Эраникуса",
	["Shadikith the Glider"] = "Шадикит Скользящий",
	["Shadow Hunter Vosh'gajin"] = "Темная охотница Вос'гаджин",
	["Shadow of Leotheras"] = "Тень Леотераса",
	["Shadowpriest Sezz'ziz"] = "Темный жрец Шезз'зиз",
	Shadron = "Шадрон",
	Shazzrah = "Шаззрах",
	["Shirrak the Dead Watcher"] = "Ширрак Страж Мертвых",
	-- ["Siamat, Lord of South Wind"] = "",
	Sindragosa = "Синдрагоса",
	["Sir Zeliek"] = "Сэр Зелиек",
	["Sjonnir The Ironshaper"] = "Сьоннир Литейщик",
	["Skadi the Ruthless"] = "Скади Безжалостный",
	["Skarr the Unbreakable"] = "Скарр Непреклонный",
	["Skarvald the Constructor"] = "Скальвард Строитель",
	["Skra'gath"] = "Скрагат",
	Skul = "Череп",
	Skum = "Шкам",
	-- Slabhide = "",
	["Slad'ran"] = "Слад'ран",
	Sneed = "Снид",
	["Sneed's Shredder"] = "Крошшер Снида",
	["Solakar Flamewreath"] = "Солакарский огнечервь",
	["Solarium Agent"] = "Агент Солнечной террасы",
	["Solarium Priest"] = "Жрец Солнечной террасы",
	["Spirestone Battle Lord"] = "Полководец из клана Черной Вершины",
	["Spirestone Butcher"] = "Мясник из клана Черной Вершины",
	["Spirestone Lord Magus"] = "Лорд-волхв из клана Черной Вершины",
	["Staff of Disintegration"] = "Посох Распыления",
	Stalagg = "Сталагг",
	Steelbreaker = "Сталелом",
	["Stomper Kreeg"] = "Топотун Криг",
	Stonespine = "Каменный Гребень",
	["Stormcaller Brundir"] = "Буревестник Брундир",
	Strawman = "Балбес",
	["Sulfuron Harbinger"] = "Предвестник Сульфурон",
	Supremus = "Супремус",
	["Svala Sorrowgrave"] = "Свала Вечноскорбящая",
	["Swamplord Musel'ek"] = "Владыка болота Мусел'ек",
	Taerar = "Таэрар",
	["Tainted Spawn of Hydross"] = "Оскверненное порождение Гидросса",
	["Talon King Ikiss"] = "Король воронов Айкисс",
	["Taragaman the Hungerer"] = "Тарагаман Ненасытный",
	["Targorr the Dread"] = "Таргорр Ужасный",
	Tavarok = "Таварок",
	Techbot = "Техбот",
	-- ["Temple Guardian Anhuur"] = "",
	Temporus = "Темпорус",
	["Tendris Warpwood"] = "Тендрис Криводрев",
	Tenebron = "Тенеброн",
	["Terestian Illhoof"] = "Терестиан Больное Копыто",
	["Teron Gorefiend"] = "Терон Кровожад",
	Thaddius = "Таддиус",
	["Thaladred the Darkener"] = "Таладред Затемнитель",
	["Thane Korth'azz"] = "Тан Кортазз",
	["The Beast"] = "Зверь",
	["The Beasts of Northrend"] = "Нордскольские чудовища",
	["The Big Bad Wolf"] = "Злой и страшный серый волк",
	["The Black Knight"] = "Черный рыцарь",
	["The Black Stalker"] = "Черная Охотница",
	["The Blue Brothers"] = "Синие братья",
	["The Bug Family"] = "Семейство насекомых",
	["The Crone"] = "Ведьма",
	["The Curator"] = "Смотритель",
	["The Eredar Twins"] = "Близнецы Эредара",
	["The Four Horsemen"] = "Четыре Всадника",
	["The Illidari Council"] = "Совет Иллидари",
	["The Iron Council"] = "Железное собрание",
	["The Lich King"] = "Король-лич",
	-- ["The Lost Dwarves"] = "",
	["The Lurker Below"] = "Скрытень глубин",
	["The Maker"] = "Мастер",
	["The Prophet Skeram"] = "Пророк Скерам",
	["The Prophet Tharon'ja"] = "Пророк Тарон'джа",
	["The Ravenian"] = "Равениан",
	["The Razza"] = "Разза",
	["The Seven Dwarves"] = "Темные хранители",
	["The Skybreaker"] = "Усмиритель небес",
	["The Tribunal of Ages"] = "Трибунал веков",
	["The Twin Emperors"] = "Близнецы Императора",
	["The Twin Val'kyr"] = "Валь'киры-близнецы",
	["The Unforgiven"] = "Непрощенный",
	["The Windreaver"] = "Ветробой",
	["Theka the Martyr"] = "Тека Мученик",
	Thorim = "Торим",
	["Thorngrin the Tender"] = "Скалезуб Скорбный",
	["Tidewalker Lurker"] = "Приливный крадень",
	["Timmy the Cruel"] = "Тайлер",
	Tinhead = "Медноголовый",
	["Tinkerer Gizlock"] = "Ремонтник Гизлок",
	["Tirion Fordring"] = "Тирион Фордринг",
	Tito = "Тито",
	["Toravon the Ice Watcher"] = "Торавон Страж Льда",
	["Trigore the Lasher"] = "Тригор Хлестун",
	Trollgore = "Кровотролль",
	["Tsu'zee"] = "Цу'зи",
	["Tuten'kash"] = "Тутен'каш",
	["Twilight Lord Kelris"] = "Повелитель сумрака Келрис",
	["Urok Doomhowl"] = "Аррок Смертный Вопль",
	["VX-001"] = "VX-001",
	["Vaelastrasz the Corrupt"] = "Валестраз Порочный",
	["Val'kyr Shadowguard"] = "Валь'кира - страж Тьмы",
	["Valithria Dreamwalker"] = "Валитрия Сноходица",
	["Varian Wrynn"] = "Вариан Ринн",
	["Varos Cloudstrider"] = "Варос Заоблачный Странник",
	Vazruden = "Вазруден",
	["Vazruden the Herald"] = "Геральд Вазруден",
	Vectus = "Вектус",
	Vem = "Вем",
	Veng = "Венг",
	["Veras Darkshadow"] = "Верас Глубокий Мрак",
	["Verdan the Everliving"] = "Вердан Бессмертный",
	Verek = "Верек",
	Vesperon = "Весперон",
	Vexallus = "Вексалиус",
	["Veyzhak the Cannibal"] = "Вейжак Каннибал",
	["Vile'rel"] = "Гнус'рел",
	Viscidus = "Нечистотон",
	["Viscous Fallout"] = "Липкая муть",
	["Void Reaver"] = "Страж Бездны",
	Volkhan = "Волхан",
	["War Master Voone"] = "Воевода Вун",
	["Warbringer O'mrogg"] = "О'мрогг Завоеватель",
	["Warchief Blackhand Piece"] = "Боевой вождь Чернорук",
	["Warchief Kargath Bladefist"] = "Вождь Каргат Острорук",
	["Warchief Rend Blackhand"] = "Вождь Ренд Чернорук",
	["Warden Mellichar"] = "Страж Мелличар",
	["Warder Stilgiss"] = "Тюремщик Стилгисс",
	["Warlord Kalithresh"] = "Полководец Калитреш",
	["Warmaul Champion"] = "Чемпион клана Боевого Молота",
	["Warp Slicer"] = "Искореженная ломтерезка",
	["Warp Splinter"] = "Узлодревень",
	["Watchkeeper Gargolmar"] = "Начальник стражи Гарголмар",
	Weaver = "Ткачик",
	["Witch Doctor Zum'rah"] = "Знахарь Зум'рах",
	["Wolf Master Nandos"] = "Командир воргенов Нандос",
	["Wrath-Scryer Soccothrates"] = "Провидец Гнева Соккорат",
	Wushoolay = "Вушулай",
	["XT-002 Deconstructor"] = "Разрушитель XT-002",
	Xevozz = "Ксевозз",
	["Yogg-Saron"] = "Йогг-Сарон",
	Yor = "Йор",
	Ysondre = "Исондра",
	Zekkis = "Зеккис",
	["Zelemar the Wrathful"] = "Зелемар Гневный",
	["Zereketh the Unbound"] = "Зерекет Бездонный",
	Zerillis = "Зериллис",
	["Zevrim Thornhoof"] = "Зеврим Терновое Копыто",
	Zolo = "Золо",
	["Zul'Farrak Dead Hero"] = "Мертвый Герой Зул'Фаррака",
	["Zul'Lor"] = "Зул'Лор",
	["Zul'jin"] = "Зул'джин",
	["Zul'tore"] = "Зул'тор",
	["Zuramat the Obliterator"] = "Зурамат Уничтожитель",
}
elseif GAME_LOCALE == "zhCN" then
	lib:SetCurrentTranslations {
	Acidmaw = "酸喉",
	Aeonus = "埃欧努斯",
	["Aerial Command Unit"] = "空中指挥单位",
	["Agathelos the Raging"] = "暴怒的阿迦赛罗斯",
	["Aggem Thorncurse"] = "阿格姆",
	Ahune = "埃霍恩",
	["Akil'zon"] = "埃基尔松",
	["Aku'mai"] = "阿库麦尔",
	["Al'ar"] = "奥",
	["Algalon the Observer"] = "观察者奥尔加隆",
	Altarius = "Altarius", -- Needs review
	["Alzzin the Wildshaper"] = "奥兹恩",
	Amanitar = "埃曼尼塔",
	["Ambassador Flamelash"] = "弗莱拉斯大使",
	["Ambassador Hellmaw"] = "赫尔默大使",
	Ammunae = "Ammunae", -- Needs review
	["Amnennar the Coldbringer"] = "寒冰之王亚门纳尔",
	["Ancient Stone Keeper"] = "古代的石头看守者",
	Anetheron = "安纳塞隆",
	["Anger'rel"] = "安格雷尔",
	Anomalus = "阿诺玛鲁斯",
	Anraphet = "Anraphet", -- Needs review
	["Antu'sul"] = "安图苏尔",
	["Anub'Rekhan"] = "阿努布雷坎",
	["Anub'arak"] = "阿努巴拉克",
	["Anub'shiah"] = "阿努希尔",
	["Anubisath Defender"] = "阿努比萨斯防御者",
	["Anubisath Guardian"] = "阿努比萨斯守卫者",
	Anzu = "安苏",
	["Arcane Watchman"] = "奥术看守",
	["Arcanist Doan"] = "奥法师杜安",
	Archaedas = "阿扎达斯",
	["Archavon the Stone Watcher"] = "岩石看守者阿尔卡冯",
	Archimonde = "阿克蒙德",
	["Archivist Galford"] = "档案管理员加尔福特",
	["Archmage Arugal"] = "大法师阿鲁高",
	["Argent Confessor Paletress"] = "银色神官帕尔崔丝",
	["Arugal's Voidwalker"] = "阿鲁高的虚空行者",
	["Ascendant Lord Obsidius"] = "Ascendant Lord Obsidius", -- Needs review
	["Asim al Akir"] = "Asim al Akir", -- Needs review
	["Assault Bot"] = "突击机器人",
	["Assembly of Iron"] = "钢铁议会",
	["Atal'alarion"] = "阿塔拉利恩",
	-- Atramedes = "",
	["Attumen the Huntsman"] = "猎手阿图门",
	Auriaya = "欧尔莉亚",
	Avalanchion = "阿瓦兰奇奥",
	["Avatar of Hakkar"] = "哈卡的化身",
	["Ayamiss the Hunter"] = "狩猎者阿亚米斯",
	Azgalor = "阿兹加洛",
	["Azshir the Sleepless"] = "永醒的艾希尔",
	Azuregos = "艾索雷葛斯",
	["Bael'Gar"] = "贝尔加",
	Baelog = "巴尔洛戈",
	Balnazzar = "巴纳扎尔",
	["Baltharus the Warborn"] = "战争之子巴尔萨鲁斯",
	["Bannok Grimaxe"] = "班诺克·巨斧",
	["Baron Aquanis"] = "阿奎尼斯男爵",
	["Baron Charr"] = "火焰男爵查尔",
	["Baron Geddon"] = "迦顿男爵",
	["Baron Kazum"] = "卡苏姆男爵",
	["Baron Rivendare"] = "瑞文戴尔男爵",
	["Baron Silverlaine"] = "席瓦莱恩男爵",
	["Baroness Anastari"] = "安娜丝塔丽男爵夫人",
	["Battleguard Sartura"] = "沙尔图拉",
	["Bazil Thredd"] = "巴基尔·斯瑞德",
	Bazzalan = "巴扎兰",
	Beauty = "Beauty", -- Needs review
	["Black Guard Swordsmith"] = "黑衣守卫铸剑师",
	["Blackheart the Inciter"] = "煽动者布莱卡特",
	["Blind Hunter"] = "盲眼猎手",
	["Blindeye the Seer"] = "盲眼先知",
	["Blood Guard Porung"] = "血卫士伯鲁恩",
	["Blood Prince Council"] = "鲜血王子议会",
	["Blood Princes"] = "鲜血王子",
	["Blood Steward of Kirtonos"] = "基尔图诺斯的卫士",
	["Blood-Queen Lana'thel"] = "鲜血女王兰娜瑟尔",
	["Bloodlord Mandokir"] = "血领主曼多基尔",
	["Bloodmage Thalnos"] = "血法师萨尔诺斯",
	Boahn = "博艾恩",
	["Bomb Bot"] = "炸弹机器人",
	["Brain of Yogg-Saron"] = "尤格-萨隆的大脑",
	["Brainwashed Noble"] = "被洗脑的贵族",
	Broggok = "布洛戈克",
	Brokentoe = "断蹄",
	Bronjahm = "布隆亚姆",
	["Broodlord Lashlayer"] = "勒什雷尔",
	["Bruegal Ironknuckle"] = "布鲁高·铁拳",
	Brutallus = "布鲁塔卢斯",
	["Burning Felguard"] = "燃烧地狱卫士",
	["Buru the Gorger"] = "吞咽者布鲁",
	["C'Thun"] = "克苏恩",
	["Cache of the Firelord"] = "火焰之王的宝箱",
	["Cache of the Legion"] = "军团宝箱",
	["Cannon Master Willey"] = "炮手威利",
	["Captain Greenskin"] = "绿皮队长",
	["Captain Kromcrush"] = "克罗卡斯",
	["Captain Skarloc"] = "斯卡洛克上尉",
	["Celebras the Cursed"] = "被诅咒的塞雷布拉斯",
	["Charlga Razorflank"] = "卡尔加·刺肋",
	["Chess Event"] = "国际象棋",
	["Chest of The Seven"] = "七贤之箱",
	["Chief Ukorz Sandscalp"] = "乌克兹·沙顶",
	-- Chimaeron = "",
	["Cho'Rush the Observer"] = "观察者克鲁什",
	Chromaggus = "克洛玛古斯",
	["Chrono Lord Deja"] = "时空领主德亚",
	["Chrono-Lord Epoch"] = "时光领主埃博克",
	Claw = "克劳恩",
	["Coilfang Elite"] = "盘牙精英",
	["Coilfang Strider"] = "盘牙巡逻者",
	["Commander Kolurg"] = "指挥官库鲁尔格",
	["Commander Sarannis"] = "指挥官萨拉妮丝",
	["Commander Springvale"] = "指挥官斯普林瓦尔",
	["Commander Stoutbeard"] = "指挥官斯托比德",
	["Commander Ulthok"] = "Commander Ulthok", -- Needs review
	["Constructor & Controller"] = "建筑师与控制者",
	Cookie = "曲奇",
	Corborus = "Corborus", -- Needs review
	["Coren Direbrew"] = "科林·烈酒",
	["Corla, Herald of Twilight"] = "Corla, Herald of Twilight", -- Needs review
	["Cosmic Infuser"] = "宇宙灌注者",
	["Crimson Hammersmith"] = "红衣铸锤师",
	["Crowd Pummeler 9-60"] = "群体打击者9-60",
	["Crystal Fang"] = "水晶之牙",
	Cyanigosa = "塞安妮苟萨",
	["Dalliah the Doomsayer"] = "末日预言者达尔莉安",
	["Dalronn the Controller"] = "控制者达尔隆",
	["Dark Coffer"] = "黑暗宝箱",
	["Dark Iron Ambassador"] = "黑铁大师",
	["Darkmaster Gandling"] = "黑暗院长加丁",
	["Darkweaver Syth"] = "黑暗编织者塞斯",
	["Death Knight Darkreaver"] = "死亡骑士达克雷尔",
	["Death Knight Understudy"] = "见习死亡骑士",
	["Death Speaker Jargba"] = "亡语者贾格巴",
	["Deathbound Ward"] = "缚亡守卫",
	["Deathbringer Saurfang"] = "死亡使者萨鲁法尔",
	["Deathspeaker High Priest"] = "亡语高阶祭司",
	["Deathstalker Visceri"] = "死亡猎手维瑟里",
	["Deathsworn Captain"] = "死亡之誓",
	Devastation = "毁坏",
	["Deviate Faerie Dragon"] = "变异精灵龙",
	["Devourer of Souls"] = "噬魂者",
	["Dextren Ward"] = "迪克斯特·瓦德",
	["Digmaster Shovelphlange"] = "挖掘专家舒尔弗拉格",
	["Doctor Theolen Krastinov"] = "瑟尔林·卡斯迪诺夫教授",
	["Doom Lord Kazzak"] = "末日领主卡扎克",
	["Doom'rel"] = "杜姆雷尔",
	Doomwalker = "末日行者",
	["Dope'rel"] = "多普雷尔",
	Dorothee = "多萝茜",
	["Drahga Shodowburner"] = "Drahga Shodowburner", -- Needs review
	["Drakkari Colossus"] = "达卡莱巨像",
	["Drakos the Interrogator"] = "审讯者达库斯",
	Dreadscale = "恐鳞",
	Dreamscythe = "德姆塞卡尔",
	["Dust Covered Chest"] = "灰尘覆盖的箱子",
	Dustwraith = "灰尘怨灵",
	["Eadric the Pure"] = "纯洁者耶德瑞克",
	["Earthcaller Halmgar"] = "唤地者哈穆加",
	["Earthrager Ptah"] = "Earthrager Ptah", -- Needs review
	Ebonroc = "埃博诺克",
	["Eck the Ferocious"] = "凶残的伊克",
	["Edwin VanCleef"] = "艾德温·范克里夫",
	["Elder Brightleaf"] = "明叶长老",
	["Elder Ironbranch"] = "铁枝长老",
	["Elder Nadox"] = "纳多克斯长老",
	["Elder Stonebark"] = "石树长老",
	["Electrocutioner 6000"] = "电刑器6000型",
	["Emalon the Storm Watcher"] = "风暴看守者埃玛尔隆",
	Emeriss = "艾莫莉丝",
	["Emperor Dagran Thaurissan"] = "达格兰·索瑞森大帝",
	["Emperor Vek'lor"] = "维克洛尔大帝",
	["Emperor Vek'nilash"] = "维克尼拉斯大帝",
	Entropius = "熵魔",
	["Eonar's Gift"] = "艾欧娜尔的礼物",
	["Epoch Hunter"] = "时空猎手",
	Erekem = "埃雷克姆",
	["Eressea Dawnsinger"] = "伊蕾希亚·晨歌",
	["Eric \"The Swift\""] = "埃瑞克",
	Erudax = "Erudax", -- Needs review
	["Erunak Stonespeaker"] = "Erunak Stonespeaker", -- Needs review
	["Essence of Anger"] = "愤怒精华",
	["Essence of Desire"] = "欲望精华",
	["Essence of Suffering"] = "苦痛精华",
	Eviscerator = "剜眼者",
	["Exarch Maladaar"] = "大主教玛拉达尔",
	["Expedition Commander"] = "远征军指挥官",
	["Eydis Darkbane"] = "黑暗邪使艾蒂丝",
	["Eye of C'Thun"] = "克苏恩之眼",
	["Faction Champions"] = "阵营冠军",
	["Fallen Champion"] = "死灵勇士",
	Falric = "法瑞克",
	["Falric and Marwyn"] = "法瑞克和玛维恩",
	["Fankriss the Unyielding"] = "顽强的范克瑞斯",
	["Fathom-Lord Karathress"] = "深水领主卡拉瑟雷斯",
	Felmyst = "菲米丝",
	["Fenrus the Devourer"] = "吞噬者芬鲁斯",
	["Feral Defender"] = "野性防御者",
	Festergut = "烂肠",
	Feugen = "费尔根",
	["Fineous Darkvire"] = "弗诺斯·达克维尔",
	Firemaw = "费尔默",
	["Fjola Lightbane"] = "光明邪使菲奥拉",
	["Flame Leviathan"] = "烈焰巨兽",
	Flamegor = "弗莱格尔",
	["Foreman Thistlenettle"] = "工头希斯耐特",
	["Forgemaster Garfrost"] = "熔炉之主加弗斯特",
	-- ["Forgemaster Throngus"] = "",
	["Four Horsemen Chest"] = "四骑士之箱",
	["Fras Siabi"] = "弗拉斯·希亚比",
	Freya = "弗蕾亚",
	["Gahz'ranka"] = "加兹兰卡",
	["Gahz'rilla"] = "加兹瑞拉",
	["Gal'darah"] = "迦尔达拉",
	["Galgann Firehammer"] = "加加恩·火锤",
	Garr = "加尔",
	["Garrosh Hellscream"] = "加尔鲁什·地狱咆哮",
	Gasher = "加什尔",
	["Gatewatcher Gyro-Kill"] = "看守者盖罗基尔",
	["Gatewatcher Iron-Hand"] = "看守者埃隆汉",
	["Gathios the Shatterer"] = "击碎者加西奥斯",
	Gehennas = "基赫纳斯",
	Gelihast = "格里哈斯特",
	Gelk = "吉尔克",
	["General Angerforge"] = "安格弗将军",
	["General Bjarngrim"] = "比亚格里将军",
	["General Drakkisath"] = "达基萨斯将军",
	["General Husam"] = "General Husam", -- Needs review
	["General Rajaxx"] = "拉贾克斯将军",
	["General Umbriss"] = "General Umbriss", -- Needs review
	["General Vezax"] = "维扎克斯将军",
	["General Zarithrian"] = "萨瑞瑟里安将军",
	["Ghamoo-ra"] = "加摩拉",
	["Ghaz'an"] = "加兹安",
	["Ghok Bashguud"] = "霍克·巴什古德",
	Gilnid = "基尔尼格",
	["Gizrul the Slavener"] = "奴役者基兹鲁尔",
	["Gloom'rel"] = "格鲁雷尔",
	Gluth = "格拉斯",
	Glutton = "暴食者",
	["Golem Lord Argelmach"] = "傀儡统帅阿格曼奇",
	["Golemagg the Incinerator"] = "焚化者古雷曼格",
	["Goraluk Anvilcrack"] = "古拉鲁克",
	["Gormok the Impaler"] = "穿刺者哥莫克",
	["Gorosh the Dervish"] = "修行者高罗什",
	["Gortok Palehoof"] = "戈托克·苍蹄",
	["Gothik the Harvester"] = "收割者戈提克",
	["Grand Astromancer Capernian"] = "星术师卡波妮娅",
	["Grand Champions"] = "总冠军",
	["Grand Crusader Dathrohan"] = "大十字军战士达索汉",
	["Grand Magus Telestra"] = "大魔导师泰蕾丝塔",
	["Grand Vizier Ertan"] = "Grand Vizier Ertan", -- Needs review
	["Grand Warlock Alythess"] = "高阶术士奥蕾塞丝",
	["Grand Warlock Nethekurse"] = "高阶术士奈瑟库斯",
	["Grand Widow Faerlina"] = "黑女巫法琳娜",
	["Grandmaster Vorpil"] = "沃匹尔大师",
	Grandmother = "老奶奶",
	["Grethok the Controller"] = "黑翼控制者",
	["Gri'lek"] = "格里雷克",
	Grimlok = "格瑞姆洛克",
	Grizzle = "格里兹尔",
	Grobbulus = "格罗布鲁斯",
	Grubbis = "格鲁比斯",
	["Gruul the Dragonkiller"] = "屠龙者格鲁尔",
	["Guard Fengus"] = "卫兵芬古斯",
	["Guard Mol'dar"] = "卫兵摩尔达",
	["Guard Slip'kik"] = "卫兵斯里基克",
	["Guardian of Yogg-Saron"] = "尤格-萨隆的卫士",
	["Gurtogg Bloodboil"] = "古尔图格·血沸",
	Gyth = "盖斯",
	Hadronox = "哈多诺克斯",
	Hakkar = "哈卡",
	Halazzi = "哈尔拉兹",
	Halion = "海里昂",
	Halycon = "哈雷肯",
	Hamhock = "哈姆霍克",
	["Harbinger Skyriss"] = "预言者斯克瑞斯",
	["Hate'rel"] = "黑特雷尔",
	["Hazza'rah"] = "哈扎拉尔",
	Hazzas = "哈扎斯",
	["Headless Horseman"] = "无头骑士",
	["Hearthsinger Forresten"] = "弗雷斯特恩",
	["Hedrum the Creeper"] = "爬行者赫杜姆",
	["Heigan the Unclean"] = "肮脏的希尔盖",
	["Hellfire Channeler"] = "地狱火导魔者",
	["Henry Stern"] = "亨利·斯特恩",
	["Herald Volazj"] = "传令官沃拉兹",
	Herod = "赫洛德",
	["Hex Lord Malacrass"] = "妖术领主玛拉卡斯",
	["High Astromancer Solarian"] = "大星术师索兰莉安",
	["High Botanist Freywinn"] = "高级植物学家弗雷温",
	["High Inquisitor Fairbanks"] = "大检察官法尔班克斯",
	["High Inquisitor Whitemane"] = "大检察官怀特迈恩",
	["High Interrogator Gerstahn"] = "审讯官格斯塔恩",
	["High King Maulgar"] = "莫加尔大王",
	["High Marshal Whirlaxis"] = "大元帅维拉希斯",
	["High Nethermancer Zerevor"] = "高阶灵术师塞勒沃尔",
	["High Overlord Saurfang"] = "萨鲁法尔大王",
	["High Priest Thekal"] = "高阶祭司塞卡尔",
	["High Priest Venoxis"] = "高阶祭司温诺希斯",
	["High Priestess Arlokk"] = "高阶祭司娅尔罗",
	["High Priestess Azil"] = "High Priestess Azil", -- Needs review
	["High Priestess Jeklik"] = "高阶祭司耶克里克",
	["High Priestess Mar'li"] = "高阶祭司玛尔里",
	["High Priestess of Thaurissan"] = "索瑞森高阶女祭司",
	["High Prophet Barim"] = "High Prophet Barim", -- Needs review
	["High Warlord Naj'entus"] = "高阶督军纳因图斯",
	["Highlord Mograine"] = "大领主莫格莱尼",
	["Highlord Omokk"] = "欧莫克大王",
	Hodir = "霍迪尔",
	["Houndmaster Grebmar"] = "驯犬者格雷布玛尔",
	["Houndmaster Loksey"] = "驯犬者洛克希",
	Hukku = "胡库",
	Hungarfen = "霍加尔芬",
	["Hurley Blackbreath"] = "霍尔雷·黑须",
	["Hyakiss the Lurker"] = "潜伏者希亚其斯",
	["Hydromancer Thespia"] = "水术师瑟丝比娅",
	["Hydromancer Velratha"] = "水占师维蕾萨",
	Hydrospawn = "海多斯博恩",
	["Hydross the Unstable"] = "不稳定的海度斯",
	["Ice Sphere"] = "Ice Sphere", -- Needs review
	["Icecrown Gunship Battle"] = "冰冠炮舰战斗",
	Icehowl = "冰吼",
	Ichoron = "艾库隆",
	Ick = "伊克",
	["Ignis the Furnace Master"] = "掌炉者伊格尼斯",
	["Illidan Stormrage"] = "伊利丹·怒风",
	["Illidari Council"] = "伊利达雷议会",
	["Illyanna Ravenoak"] = "伊琳娜·暗木",
	["Immol'thar"] = "伊莫塔尔",
	["Infinite Corruptor"] = "永恒腐蚀者",
	["Infinity Blades"] = "无尽之刃",
	["Ingvar the Plunderer"] = "劫掠者因格瓦尔",
	["Instructor Malicia"] = "讲师玛丽希亚",
	["Instructor Razuvious"] = "教官拉苏维奥斯",
	["Interrogator Vishas"] = "审讯员韦沙斯",
	Ionar = "艾欧纳尔",
	Ironaya = "艾隆纳亚",
	Ironspine = "铁脊死灵",
	Isalien = "伊萨利恩",
	Isiset = "Isiset", -- Needs review
	Jade = "玉龙",
	["Jammal'an the Prophet"] = "预言者迦玛兰",
	["Jan'alai"] = "加亚莱",
	["Jandice Barov"] = "詹迪斯·巴罗夫",
	["Jed Runewatcher"] = "杰德",
	["Jedoga Shadowseeker"] = "耶戈达·觅影者",
	["Jergosh the Invoker"] = "祈求者耶戈什",
	["Jin'do the Hexxer"] = "妖术师金度",
	["Jormungar Behemoth"] = "蛰猛巨虫",
	Jormungars = "冰虫",
	Julianne = "朱丽叶",
	["Junk Bot"] = "废物机器人",
	["Kael'thas Sunstrider"] = "凯尔萨斯·逐日者",
	Kalecgos = "卡雷苟斯",
	["Kam Deepfury"] = "卡姆·深怒",
	["Karsh Steelbender"] = "Karsh Steelbender", -- Needs review
	["Kaz'rogal"] = "卡兹洛加",
	["Kazkaz the Unholy"] = "邪恶的卡萨卡兹",
	["Kel'Thuzad"] = "克尔苏加德",
	["Keli'dan the Breaker"] = "击碎者克里丹",
	Keristrasza = "克莉斯塔萨",
	["Kiggler the Crazed"] = "疯狂的基戈尔",
	["Kil'jaeden"] = "基尔加丹",
	["Kil'rek"] = "基尔里克",
	["King Dred"] = "暴龙之王爵德",
	["King Gordok"] = "戈多克大王",
	["King Llane Piece"] = "莱恩国王",
	["King Ymiron"] = "伊米隆国王",
	["Kirtonos the Herald"] = "传令官基尔图诺斯",
	["Knot Thimblejack's Cache"] = "诺特·希姆加克的储物箱",
	Kolk = "考尔克",
	Kologarn = "科隆加恩",
	["Koralon the Flame Watcher"] = "火焰看守者科拉隆",
	Kormok = "库尔莫克",
	Kresh = "克雷什",
	Krick = "科里克",
	["Krick and Ick"] = "科里克和伊克",
	["Krik'thir the Gatewatcher"] = "看门者克里克希尔",
	["Krosh Firehand"] = "克洛什·火拳",
	Krystallus = "克莱斯塔卢斯",
	Kurinnaxx = "库林纳克斯",
	["Lady Anacondra"] = "安娜科德拉",
	["Lady Blaumeux"] = "女公爵布劳缪克丝",
	["Lady Deathwhisper"] = "亡语者女士",
	["Lady Illucia Barov"] = "伊露希亚·巴罗夫",
	["Lady Malande"] = "女公爵玛兰德",
	["Lady Naz'jar"] = "Lady Naz'jar", -- Needs review
	["Lady Sacrolash"] = "萨洛拉丝女王",
	["Lady Sarevess"] = "萨利维丝",
	["Lady Vashj"] = "瓦丝琪",
	Laj = "拉伊",
	Landslide = "兰斯利德",
	Lavanthor = "拉文索尔",
	["Left Arm"] = "左臂",
	["Leotheras the Blind"] = "盲眼者莱欧瑟拉斯",
	Lethon = "莱索恩",
	Lethtendris = "蕾瑟塔蒂丝",
	["Leviathan Mk II"] = "巨兽二型",
	["Ley-Guardian Eregos"] = "魔网守护者埃雷苟斯",
	["Lieutenant Drake"] = "德拉克中尉",
	["Lieutenant General Andorov"] = "安多洛夫中将",
	Loatheb = "洛欧塞布",
	Lockmaw = "Lockmaw", -- Needs review
	Loken = "洛肯",
	["Lord Alexei Barov"] = "阿雷克斯·巴罗夫",
	["Lord Cobrahn"] = "考布莱恩",
	["Lord Hel'nurath"] = "赫尔努拉斯",
	["Lord Incendius"] = "伊森迪奥斯",
	["Lord Jaraxxus"] = "加拉克苏斯大王",
	["Lord Kazzak"] = "末日领主卡扎克",
	["Lord Kri"] = "克里勋爵",
	["Lord Marrowgar"] = "玛洛加尔领主",
	["Lord Pythas"] = "皮萨斯",
	["Lord Roccor"] = "洛考尔",
	["Lord Sanguinar"] = "萨古纳尔男爵",
	["Lord Serpentis"] = "瑟芬迪斯",
	["Lord Skwol"] = "斯古恩男爵",
	["Lord Valthalak"] = "瓦塔拉克公爵",
	["Lord Victor Nefarius"] = "维克多·奈法里奥斯",
	["Lord Vyletongue"] = "维利塔恩",
	["Lorekeeper Polkelt"] = "博学者普克尔特",
	["Lorgus Jett"] = "洛古斯·杰特",
	Loro = "洛若尔",
	Lucifron = "鲁西弗隆",
	["M'uru"] = "穆鲁",
	["Mad Magglish"] = "疯狂的马格利什",
	Maexxna = "迈克斯纳",
	["Mage-Lord Urom"] = "法师领主伊洛姆",
	["Magister Kalendris"] = "卡雷迪斯镇长",
	["Magistrate Barthilas"] = "巴瑟拉斯镇长",
	Magmadar = "玛格曼达",
	-- Magmaw = "",
	Magmus = "玛格姆斯",
	Magra = "玛格拉",
	Magtheridon = "玛瑟里顿",
	["Maiden of Grief"] = "悲伤圣女",
	["Maiden of Virtue"] = "贞节圣女",
	["Majordomo Executus"] = "管理者埃克索图斯",
	["Mal'Ganis"] = "玛尔加尼斯",
	Malacrass = "玛拉卡斯",
	["Maleki the Pallid"] = "苍白的玛勒基",
	-- Maloriak = "",
	Malygos = "玛里苟斯",
	Maraudos = "玛拉多斯",
	["Marduk Blackpool"] = "马杜克·布莱克波尔",
	["Marisa du'Paige"] = "玛里莎·杜派格",
	Marwyn = "玛维恩",
	["Master Engineer Telonicus"] = "首席技师塔隆尼库斯",
	["Maur Grimtotem"] = "玛尔·恐怖图腾",
	Meathook = "肉钩",
	["Mechano-Lord Capacitus"] = "机械领主卡帕西图斯",
	Medivh = "麦迪文",
	["Mekgineer Steamrigger"] = "机械师斯蒂里格",
	["Mekgineer Thermaplugg"] = "麦克尼尔·瑟玛普拉格",
	["Mennu the Betrayer"] = "背叛者门努",
	["Meshlok the Harvester"] = "收割者麦什洛克",
	Midnight = "午夜",
	Mijan = "米杉",
	Mimiron = "米米尔隆",
	["Mindbender Ghur'sha"] = "Mindbender Ghur'sha", -- Needs review
	["Miner Johnson"] = "矿工约翰森",
	["Mistress of Pain"] = "痛苦女王",
	Moam = "莫阿姆",
	Mogor = "穆戈尔",
	["Mokra the Skullcrusher"] = "碎颅者莫克拉",
	Moorabi = "莫拉比",
	["Mor Grayhoof"] = "莫尔·灰蹄",
	Moragg = "摩拉格",
	["Mordresh Fire Eye"] = "火眼莫德雷斯",
	Moroes = "莫罗斯",
	["Morogrim Tidewalker"] = "莫洛格里·踏潮者",
	Morphaz = "摩弗拉斯",
	["Mother Shahraz"] = "莎赫拉丝主母",
	["Mother Smolderweb"] = "烟网蛛后",
	["Mr. Smite"] = "重拳先生",
	["Muradin Bronzebeard"] = "穆拉丁·铜须",
	["Murkblood Twin"] = "暗血双子",
	["Murkblood Twins"] = "暗血双子",
	Murmur = "摩摩尔",
	["Murta Grimgut"] = "穆尔塔",
	Mushgog = "姆斯高格",
	["Mutanus the Devourer"] = "吞噬者穆坦努斯",
	Nalorakk = "纳洛拉克",
	Nazan = "纳杉",
	Nefarian = "奈法利安",
	["Nekrum Gutchewer"] = "耐克鲁姆",
	["Nerub'enkan"] = "奈鲁布恩坎",
	["Nethermancer Sepethrea"] = "灵术师塞比瑟蕾",
	Netherspite = "虚空幽龙",
	["Netherstrand Longbow"] = "灵弦长弓",
	["Nexus-Prince Shaffar"] = "节点亲王沙法尔",
	Nightbane = "夜之魇",
	["Noth the Plaguebringer"] = "药剂师诺斯",
	["Novos the Summoner"] = "召唤者诺沃斯",
	Noxxion = "诺克赛恩",
	["Obsidian Sentinel"] = "黑曜石哨兵",
	["Odo the Blindwatcher"] = "盲眼守卫奥杜",
	Oggleflint = "奥格弗林特",
	["Ogom the Wretched"] = "可悲的奥戈姆",
	Ohgan = "奥根",
	["Ok'thor the Breaker"] = "破坏者奥科索尔",
	Olaf = "奥拉夫",
	["Old Serra'kis"] = "瑟拉吉斯",
	["Olm the Summoner"] = "召唤者沃尔姆",
	["Omor the Unscarred"] = "无疤者奥摩尔",
	Onyxia = "奥妮克希亚",
	["Opera Event"] = "歌剧院",
	["Orgrim's Hammer"] = "奥格瑞姆之锤",
	["Ormorok the Tree-Shaper"] = "塑树者奥莫洛克",
	["Oro Eyegouge"] = "欧罗·血眼",
	["Ossirian the Unscarred"] = "无疤者奥斯里安",
	Ouro = "奥罗",
	["Overlord Ramtusk"] = "主宰拉姆塔斯",
	["Overlord Wyrmthalak"] = "维姆萨拉克",
	["Overmaster Pyron"] = "征服者派隆",
	["Overseer Tidewrath"] = "工头泰德瓦斯",
	Ozruk = "Ozruk", -- Needs review
	Ozumat = "Ozumat", -- Needs review
	Pandemonius = "潘德莫努斯",
	["Panzor the Invincible"] = "无敌的潘佐尔",
	Patchwerk = "帕奇维克",
	["Pathaleon the Calculator"] = "计算者帕萨雷恩",
	Phalanx = "方阵",
	["Phaseshift Bulwark"] = "相位壁垒",
	Pimgib = "匹姆吉布",
	["Plaguemaw the Rotting"] = "腐烂的普雷莫尔",
	["Plugger Spazzring"] = "普拉格",
	["Postmaster Malown"] = "邮差马龙",
	["Priestess Delrissa"] = "女祭司德莉希亚",
	["Prince Keleseth"] = "凯雷塞斯王子",
	["Prince Malchezaar"] = "玛克扎尔王子",
	["Prince Skaldrenox"] = "斯卡德诺克斯王子",
	["Prince Taldaram"] = "塔达拉姆王子",
	["Prince Tenris Mirkblood"] = "特里斯·黯血王子",
	["Prince Tortheldrin"] = "托塞德林王子",
	["Prince Valanar"] = "瓦拉纳王子",
	["Princess Huhuran"] = "哈霍兰公主",
	["Princess Moira Bronzebeard"] = "铁炉堡公主茉艾拉·铜须",
	["Princess Tempestria"] = "泰比斯蒂亚公主",
	["Princess Theradras"] = "瑟莱德丝公主",
	["Princess Yauj"] = "亚尔基公主",
	["Professor Putricide"] = "普崔塞德教授",
	["Pure Spawn of Hydross"] = "纯净的海度斯爪牙",
	Pusillin = "普希林",
	["Pyroguard Emberseer"] = "烈焰卫士艾博希尔",
	["Pyromancer Loregrain"] = "控火师罗格雷恩",
	Quagmirran = "夸格米拉",
	["Quartermaster Zigris"] = "军需官兹格雷斯",
	["Rage Winterchill"] = "雷基·冬寒",
	Ragglesnout = "拉戈斯诺特",
	["Raging Spirit"] = "暴怒灵魂", -- Needs review
	Ragnaros = "拉格纳罗斯",
	Rajh = "Rajh", -- Needs review
	["Ramstein the Gorger"] = "吞咽者拉姆斯登",
	["Ras Frostwhisper"] = "莱斯·霜语",
	Rattlegore = "血骨傀儡",
	["Razorclaw the Butcher"] = "屠夫拉佐克劳",
	["Razorgore the Untamed"] = "狂野的拉佐格尔",
	Razorlash = "锐刺鞭笞者",
	Razorscale = "锋鳞",
	["Reliquary of Souls"] = "灵魂之匣",
	Renataki = "雷纳塔基",
	["Restless Skeleton"] = "无法安息的骷髅",
	Rethilgore = "雷希戈尔",
	Revelosh = "鲁维罗什",
	["Rhahk'Zor"] = "拉克佐",
	["Ribbly Screwspigot"] = "雷布里·斯库比格特",
	["Right Arm"] = "右臂",
	["Ring of Law"] = "秩序竞技场",
	Roar = "胆小的狮子",
	["Rokad the Ravager"] = "蹂躏者洛卡德",
	["Rokdar the Sundered Lord"] = "裂石之王洛卡达尔",
	["Rokmar the Crackler"] = "巨钳鲁克玛尔",
	["Rom'ogg Bonecrusher"] = "Rom'ogg Bonecrusher", -- Needs review
	Romulo = "罗密欧",
	["Romulo & Julianne"] = "罗密欧与朱丽叶",
	Roogug = "鲁古",
	Rotface = "腐面",
	Rotgrip = "洛特格里普",
	["Runemaster Molgeim"] = "符文大师莫尔基姆",
	["Runok Wildmane"] = "鲁诺克·蛮鬃",
	Ruuzlu = "卢兹鲁",
	["Salramm the Fleshcrafter"] = "塑血者沙尔拉姆",
	["Sanctum Sentry"] = "圣所警卫",
	["Sandarr Dunereaver"] = "杉达尔·沙掠者",
	["Sandfury Executioner"] = "沙怒刽子手",
	Sapphiron = "萨菲隆",
	Sara = "萨拉",
	["Saronite Animus"] = "萨隆邪铁畸体",
	Sartharion = "萨塔里奥",
	["Sathrovarr the Corruptor"] = "腐蚀者萨索瓦尔",
	["Saviana Ragefire"] = "塞维娅娜·怒火",
	["Scarlet Commander Mograine"] = "血色十字军指挥官莫格莱尼",
	["Scourgelord Tyrannus"] = "天灾领主泰兰努斯",
	["Seeth'rel"] = "西斯雷尔",
	["Selin Fireheart"] = "塞林·火心",
	["Sergeant Bly"] = "布莱中士",
	Setesh = "Setesh", -- Needs review
	["Shade of Akama"] = "阿卡玛之影",
	["Shade of Aran"] = "埃兰之影",
	["Shade of Eranikus"] = "伊兰尼库斯的阴影",
	["Shadikith the Glider"] = "滑翔者沙德基斯",
	["Shadow Hunter Vosh'gajin"] = "暗影猎手沃什加斯",
	["Shadow of Leotheras"] = "莱欧瑟拉斯之影",
	["Shadowpriest Sezz'ziz"] = "暗影祭司塞瑟斯",
	Shadron = "沙德隆",
	Shazzrah = "沙斯拉尔",
	["Shirrak the Dead Watcher"] = "死亡观察者希尔拉克",
	["Siamat, Lord of South Wind"] = "Siamat, Lord of South Wind", -- Needs review
	Sindragosa = "辛达苟萨",
	["Sir Zeliek"] = "瑟里耶克爵士",
	["Sjonnir The Ironshaper"] = "塑铁者斯约尼尔",
	["Skadi the Ruthless"] = "残忍的斯卡迪",
	["Skarr the Unbreakable"] = "无敌的斯卡尔",
	["Skarvald the Constructor"] = "建筑师斯卡瓦尔德",
	["Skra'gath"] = "瑟克拉加斯",
	Skul = "斯库尔",
	Skum = "斯卡姆",
	Slabhide = "Slabhide",
	["Slad'ran"] = "斯拉德兰",
	Sneed = "斯尼德",
	["Sneed's Shredder"] = "斯尼德的伐木机",
	["Solakar Flamewreath"] = "索拉卡·火冠",
	["Solarium Agent"] = "日晷密探",
	["Solarium Priest"] = "日晷祭司",
	["Spirestone Battle Lord"] = "尖石统帅",
	["Spirestone Butcher"] = "尖石屠夫",
	["Spirestone Lord Magus"] = "尖石首席法师",
	["Staff of Disintegration"] = "瓦解法杖",
	Stalagg = "斯塔拉格",
	Steelbreaker = "断钢者",
	["Stomper Kreeg"] = "践踏者克雷格",
	Stonespine = "石脊",
	["Stormcaller Brundir"] = "唤雷者布隆迪尔",
	Strawman = "稻草人",
	["Sulfuron Harbinger"] = "萨弗隆先驱者",
	Supremus = "苏普雷姆斯",
	["Svala Sorrowgrave"] = "席瓦拉·索格蕾",
	["Swamplord Musel'ek"] = "沼地领主穆塞雷克",
	Taerar = "泰拉尔",
	["Tainted Spawn of Hydross"] = "污染的海度斯爪牙",
	["Talon King Ikiss"] = "利爪之王艾吉斯",
	["Taragaman the Hungerer"] = "饥饿者塔拉加曼",
	["Targorr the Dread"] = "可怕的塔格尔",
	Tavarok = "塔瓦洛克",
	Techbot = "尖端机器人",
	["Temple Guardian Anhuur"] = "Temple Guardian Anhuur", -- Needs review
	Temporus = "坦普卢斯",
	["Tendris Warpwood"] = "特迪斯·扭木",
	Tenebron = "塔尼布隆",
	["Terestian Illhoof"] = "特雷斯坦·邪蹄",
	["Teron Gorefiend"] = "塔隆·血魔",
	Thaddius = "塔迪乌斯",
	["Thaladred the Darkener"] = "亵渎者萨拉德雷",
	["Thane Korth'azz"] = "库尔塔兹领主",
	["The Beast"] = "比斯巨兽",
	["The Beasts of Northrend"] = "诺森德猛兽",
	["The Big Bad Wolf"] = "大灰狼",
	["The Black Knight"] = "黑骑士",
	["The Black Stalker"] = "黑色阔步者",
	["The Blue Brothers"] = "蓝色兄弟",
	["The Bug Family"] = "虫子一家",
	["The Crone"] = "巫婆",
	["The Curator"] = "馆长",
	["The Eredar Twins"] = "艾瑞达双子",
	["The Four Horsemen"] = "四骑士",
	["The Illidari Council"] = "伊利达雷议会",
	["The Iron Council"] = "钢铁议会",
	["The Lich King"] = "巫妖王",
	["The Lost Dwarves"] = "失踪的矮人",
	["The Lurker Below"] = "鱼斯拉",
	["The Maker"] = "制造者",
	["The Prophet Skeram"] = "预言者斯克拉姆",
	["The Prophet Tharon'ja"] = "先知萨隆亚",
	["The Ravenian"] = "拉文尼亚",
	["The Razza"] = "拉扎尔",
	["The Seven Dwarves"] = "七贤矮人",
	["The Skybreaker"] = "破天号",
	["The Tribunal of Ages"] = "远古法庭",
	["The Twin Emperors"] = "双子皇帝",
	["The Twin Val'kyr"] = "瓦格里双子",
	["The Unforgiven"] = "不可宽恕者",
	["The Windreaver"] = "烈风掠夺者",
	["Theka the Martyr"] = "殉教者塞卡",
	Thorim = "托里姆",
	["Thorngrin the Tender"] = "看管者索恩格林",
	["Tidewalker Lurker"] = "踏潮潜伏者",
	["Timmy the Cruel"] = "悲惨的提米",
	Tinhead = "铁皮人",
	["Tinkerer Gizlock"] = "工匠吉兹洛克",
	["Tirion Fordring"] = "提里奧·弗丁",
	Tito = "托托",
	["Toravon the Ice Watcher"] = "寒冰看守者图拉旺",
	["Trigore the Lasher"] = "鞭笞者特里高雷",
	Trollgore = "托尔戈",
	["Tsu'zee"] = "苏斯",
	["Tuten'kash"] = "图特卡什",
	["Twilight Lord Kelris"] = "梦游者克尔里斯",
	["Urok Doomhowl"] = "乌洛克",
	["VX-001"] = "VX-001",
	["Vaelastrasz the Corrupt"] = "堕落的瓦拉斯塔兹",
	["Val'kyr Shadowguard"] = "Val'kyr Shadowguard", -- Needs review
	["Valithria Dreamwalker"] = "踏梦者瓦莉瑟瑞娅",
	["Varian Wrynn"] = "瓦里安·乌瑞恩",
	["Varos Cloudstrider"] = "瓦尔洛斯·云击",
	Vazruden = "瓦兹德",
	["Vazruden the Herald"] = "传令官瓦兹德",
	Vectus = "维克图斯",
	Vem = "维姆",
	Veng = "温格",
	["Veras Darkshadow"] = "维尔莱斯·深影",
	["Verdan the Everliving"] = "永生者沃尔丹",
	Verek = "维雷克",
	Vesperon = "维斯匹隆",
	Vexallus = "维萨鲁斯",
	["Veyzhak the Cannibal"] = "食尸者维萨克",
	["Vile'rel"] = "瓦勒雷尔",
	Viscidus = "维希度斯",
	["Viscous Fallout"] = "粘性辐射尘",
	["Void Reaver"] = "空灵机甲",
	Volkhan = "沃尔坎",
	["War Master Voone"] = "指挥官沃恩",
	["Warbringer O'mrogg"] = "战争使者沃姆罗格",
	["Warchief Blackhand Piece"] = "黑手酋长",
	["Warchief Kargath Bladefist"] = "酋长卡加斯·刃拳",
	["Warchief Rend Blackhand"] = "大酋长雷德·黑手",
	["Warden Mellichar"] = "监护者梅里卡尔",
	["Warder Stilgiss"] = "典狱官斯迪尔基斯",
	["Warlord Kalithresh"] = "督军卡利瑟里斯",
	["Warmaul Champion"] = "战槌勇士",
	["Warp Slicer"] = "迁跃切割者",
	["Warp Splinter"] = "迁跃扭木",
	["Watchkeeper Gargolmar"] = "巡视者加戈玛",
	Weaver = "德拉维沃尔",
	["Witch Doctor Zum'rah"] = "巫医祖穆拉恩",
	["Wolf Master Nandos"] = "狼王南杜斯",
	["Wrath-Scryer Soccothrates"] = "天怒预言者苏克拉底",
	Wushoolay = "乌苏雷",
	["XT-002 Deconstructor"] = "XT-002拆解者",
	Xevozz = "谢沃兹",
	["Yogg-Saron"] = "尤格-萨隆",
	Yor = "尤尔",
	Ysondre = "伊森德雷",
	Zekkis = "泽基斯",
	["Zelemar the Wrathful"] = "愤怒者塞雷玛尔",
	["Zereketh the Unbound"] = "自由的瑟雷凯斯",
	Zerillis = "泽雷利斯",
	["Zevrim Thornhoof"] = "瑟雷姆·刺蹄",
	Zolo = "祖罗",
	["Zul'Farrak Dead Hero"] = "祖尔法拉克阵亡英雄",
	["Zul'Lor"] = "祖罗尔",
	["Zul'jin"] = "祖尔金",
	["Zul'tore"] = "祖尔托",
	["Zuramat the Obliterator"] = "湮灭者祖拉玛特",
}
else
	lib:SetCurrentTranslations(true)
end
